/* eslint  no-console:0 */
const AWS = require('aws-sdk');
const axios = require('axios');

async function getRegions() {
    const EC2 = new AWS.EC2({ region: 'us-east-1' });
    const response = await EC2.describeRegions().promise();

    return response.Regions.map(region => region.RegionName);
}

async function generateData(amis) {
    const promises = amis.map(async ami => {
        console.log('Handling region:', ami.region);

        const response = await axios.get(`https://pricing.us-east-1.amazonaws.com/offers/v1.0/aws/AmazonEC2/current/${ami.region}/index.json`);

        const filtered = Object.values(response.data.products).filter(product => product.productFamily === 'Compute Instance'
            && product.attributes.operatingSystem === 'Linux'
            && product.attributes.instanceFamily === 'General purpose'
            && product.attributes.vcpu === '4'
            && product.attributes.memory === '16 GiB'
            && product.attributes.tenancy === 'Shared'
            && product.attributes.storage === 'EBS only'
            && product.attributes.enhancedNetworkingSupported === 'Yes'
            && product.attributes.instanceType === 'm5.xlarge');

        return {
            region: ami.region,
            ami: ami.imageId,
            instanceType: filtered.length === 0 ? 'm4.xlarge' : 'm5.xlarge'
        };
    });

    return Promise.all(promises);
}

async function getLatesAmazonLinuxAmi(region) {
    console.log('Getting images for region:', region);

    const EC2 = new AWS.EC2({ region });

    const response = await EC2.describeImages({
        Filters:
            [
                {
                    Name: 'virtualization-type',
                    Values: ['hvm']
                },
                {
                    Name: 'architecture',
                    Values: ['x86_64']
                },
                {
                    Name: 'state',
                    Values: ['available']
                },
                {
                    Name: 'root-device-type',
                    Values: ['ebs']
                },
                {
                    Name: 'hypervisor',
                    Values: ['xen']
                },
                {
                    Name: 'image-type',
                    Values: ['machine']
                },
                {
                    Name: 'block-device-mapping.volume-type',
                    Values: ['gp2']
                },
                {
                    Name: 'name',
                    Values: ['amzn2-ami-hvm-*']
                },
                {
                    Name: 'owner-id',
                    Values: ['137112412989']
                }
            ],
        Owners: ['amazon']
    })
        .promise();

    console.log(`Found ${response.Images.length} images in region:`, region);

    // 2016-02-10T23:44:07.000Z <- ISO_8601
    const sorted = response.Images.sort((a, b) => Date.parse(b.CreationDate) - Date.parse(a.CreationDate));

    return {
        region,
        imageId: sorted[0].ImageId
    };
}

async function main() {
    console.log('Starting...');

    console.log('Getting all AWS regions...');
    const regions = await getRegions();
    console.log('Got regions:', regions);

    const promises = regions.map(region => getLatesAmazonLinuxAmi(region));
    const latestAmazonLinuxAMIs = await Promise.all(promises);
    console.log('Got latest AMIs:', latestAmazonLinuxAMIs);

    const data = await generateData(latestAmazonLinuxAMIs);
    const regionMap = {};

    data.forEach(x => {
        regionMap[x.region] = {
            ami: x.ami,
            InstanceType: x.instanceType
        };
    });

    console.log(JSON.stringify(regionMap));
}

main();
