/* eslint  no-console:0 */
const AWS = require('aws-sdk');
const {
    createWriteStream, createReadStream, remove, writeFile, readFile
} = require('fs-extra');
const { join } = require('path');
const { v4 } = require('uuid');
const yargs = require('yargs'); // eslint-disable-line import/no-extraneous-dependencies
const archiver = require('archiver'); // eslint-disable-line import/no-extraneous-dependencies
const gitrev = require('git-rev-sync'); // eslint-disable-line import/no-extraneous-dependencies

const NODE_VERSION = '12.2.0';

const { argv } = yargs
    .usage('Usage: $0 --bucket --accessKeyId --secretAccessKey --environment')
    .demandOption(['bucket', 'accessKeyId', 'secretAccessKey', 'environment'])
    .array('bundleCacheRefresh')
    .option('label');

async function getVersion() {
    const rootPackage = JSON.parse(await readFile('./package.json'));
    return `${rootPackage.version}.${gitrev.count()}-${gitrev.short()}-${gitrev.branch()}`;
}

async function zip(directory) {
    return new Promise((resolve, reject) => {
        const out = `./${v4()}.zip`;
        const output = createWriteStream(out);
        const archive = archiver('zip');
        output.on('close', () => resolve(out));
        archive.on('error', err => reject(err));
        archive.pipe(output);

        archive
            .directory(join(__dirname, directory), false)
            .finalize();
    });
}

async function upload(file, bucket, accessKeyId, secretAccessKey, dataBrokerEnvironment, versionToUpdate, key, label) {
    const s3 = new AWS.S3({ accessKeyId, secretAccessKey });

    const tagSet = [
        {
            Key: 'buildTime',
            Value: Date.now().toString()
        },
        {
            Key: 'environment',
            Value: dataBrokerEnvironment
        },
        {
            Key: 'version',
            Value: versionToUpdate
        }
    ];


    if (label) {
        tagSet.push({
            Key: 'label',
            Value: label
        });
    }

    const response = await new AWS.S3.ManagedUpload({
        params: {
            Bucket: bucket,
            Key: key,
            Body: createReadStream(file),
            ACL: 'public-read'
        },
        tags: tagSet,
        service: s3
    }).promise();

    return response.Location;
}

async function main() {
    try {
        const versionToUpdate = await getVersion();
        const { environment: dataBrokerEnvironment, label } = argv;
        const key = (label) ? `repo/${dataBrokerEnvironment}/phoenix/data-broker_${versionToUpdate}_${label}.zip`
            : `repo/${dataBrokerEnvironment}/phoenix/data-broker_${versionToUpdate}.zip`;
        // add about.json
        await writeFile(join(__dirname, '../build/about.json'), JSON.stringify({
            version: versionToUpdate,
            nodeVersion: NODE_VERSION
        }, null, 4));
        // archive build directory
        const out = await zip('../build');

        // upload and make public
        await upload(out, argv.bucket, argv.accessKeyId, argv.secretAccessKey, dataBrokerEnvironment, versionToUpdate, key, argv.label);
        console.log('Bundle uploaded');
        // remove the ZIP file
        await remove(out);
    } catch (e) {
        console.error('Error occurred during create bundle ', e);
        throw e;
    }
}

main();
