const https = require('https');
const { stringify } = require('querystring');
const Promise = require('bluebird');
const { homedir } = require('os');
const { join } = require('path');
const chokidar = require('chokidar');
const AWS = require('aws-sdk');
const proxy = require('proxy-agent');
const mimeType = require('mime-types');
const config = require('config');
const bytes = require('bytes');
const {
    isNil, compact, last, reduce, keys, sortBy, tail, dropRight, zip, range, first
} = require('lodash');
const {
    TYPE_FILE, TYPE_DIRECTORY, TYPE_SYMBOLIC, SQS_DEFAULT_REGION
} = require('../consts');
const getLogger = require('../logger/logger');

const logger = getLogger();

const RETRIABLE_ERRORS = ['MalformedXML', 'InvalidPart', 'NoSuchUpload', 'MethodNotAllowed'];

chokidar.watch(join(homedir(), '.aws', 'credentials'), { usePolling: { interval: 3000 } })
    .on('change', path => {
        try {
            logger.info(`Shared credentials file ${path} changed, clearing cache...`);
            AWS.util.iniLoader.clearCachedFiles();
            logger.info('Shared credentials file cache cleared');
        } catch (err) {
            logger.error('Failed to clear shared credentials file cache', err);
        }
    });

AWS.events
    .on('retry', response => {
        if (RETRIABLE_ERRORS.includes(response.error.code)) {
            logger.warn('Retrying AWS S3 error:', response.error.code);
            response.error.retryable = true;
        }
    })
    .on('build', request => { request.httpRequest.headers['User-Agent'] = 'APN/1.0 NetApp/1.0 CloudSync/1.0'; });

function getS3(region) {
    const options = { region, correctClockSkew: true, useDualstack: config.get('protocols.s3.useDualstack') };

    const endpoint = config.get('protocols.s3.endpoint');
    if (endpoint) {
        options.endpoint = endpoint;
    }

    const useProxy = config.get('protocols.s3.use-proxy');
    const rejectInvalidCertificates = config.get('protocols.s3.reject-invalid-certificates');
    if (process.env.http_proxy && useProxy) {
        options.httpOptions = {
            agent: proxy(process.env.http_proxy, { rejectUnauthorized: rejectInvalidCertificates })
        };
    } else {
        options.httpOptions = {
            agent: new https.Agent({ rejectUnauthorized: rejectInvalidCertificates })
        };
    }

    return new AWS.S3(options);
}

async function getBucketLocation(bucket) {
    const s3 = getS3();
    const response = await s3.getBucketLocation({ Bucket: bucket }).promise();
    return response.LocationConstraint === '' ? SQS_DEFAULT_REGION : response.LocationConstraint;
}

async function getBucketTagging(region, bucket) {
    const s3 = getS3(region);
    try {
        const response = await s3.getBucketTagging({ Bucket: bucket }).promise();
        return response.TagSet;
    } catch (err) {
        if (err.code === 'NoSuchTagSet') {
            return [];
        }
        throw err;
    }
}

async function listBuckets() {
    const s3 = getS3();
    const response = await s3.listBuckets().promise();

    const promises = response.Buckets.map(async bucket => {
        try {
            const location = await getBucketLocation(bucket.Name);
            const tags = await getBucketTagging(location, bucket.Name);
            return { name: bucket.Name, region: location, tags };
        } catch (err) {
            logger.warn(`Failed to get bucket location/tagging, bucket name ${bucket.Name}`, err);
        }
    });

    const result = compact(await Promise.all(promises));
    return sortBy(result.filter(bucket => !bucket.name.startsWith('fabric-pool-')), 'name');
}

async function headBucket(bucket) {
    const s3 = getS3(SQS_DEFAULT_REGION);
    return new Promise((resolve, reject) => {
        s3.headBucket({ Bucket: bucket })
            .on('httpHeaders', (statusCode, headers) => {
                if (headers['x-amz-bucket-region']) {
                    resolve(headers['x-amz-bucket-region']);
                } else {
                    reject(new Error(`Got unexpected status code ${statusCode} - for headBucket operation on bucket ${bucket}`));
                }
            })
            .on('error', error => reject(error))
            .on('httpError', error => reject(error))
            .send();
    });
}

async function getObjectTagging(region, bucket, key) {
    const s3 = getS3(region);
    try {
        const tags = await s3.getObjectTagging({ Bucket: bucket, Key: key }).promise();
        return tags.TagSet;
    } catch (err) {
        if (err.code === 'NoSuchKey') { // in case of a folder without tags returns NoSuchKey
            return [];
        }
        if (err.code === 'MethodNotAllowed') { // temporary fix for eric
            logger.error(`Got MethodNotAllowed on getObjectTags on ${key} inside the bucket ${bucket}`);
            return [];
        }
        throw err;
    }
}

async function getObjectsAndTags(region, bucket, key, size, objectType, mode, lastModified) {
    const s3Tags = await getObjectTagging(region, bucket, key);
    const s3TagsObject = reduce(s3Tags, (result, current) => (Object.assign(result, { [current.Key]: current.Value })), {});
    const modifiedTime = lastModified === undefined ? Date.now() : lastModified.getTime();
    if (!key.endsWith('/') || objectType !== TYPE_FILE) {
        return {
            path: (key === '/') ? '/' : last(compact(key.split('/'))),
            type: (s3TagsObject['netapp-cs-slink'] === 'True') ? TYPE_SYMBOLIC : objectType,
            size,
            mtimeSec: parseInt(s3TagsObject['netapp-cs-mtimeSec']) || parseInt(modifiedTime / 1000),
            mode: parseInt(s3TagsObject['netapp-cs-mode']) || mode,
            uid: parseInt(s3TagsObject['netapp-cs-uid']) || 0,
            gid: parseInt(s3TagsObject['netapp-cs-gid']) || 0
        };
    }
    return null;
}

async function listBucketFolders(region, bucket, prefix, continuationToken, dirs = []) {
    const s3 = getS3(region);

    let fixedPrefix = isNil(prefix) || prefix === '' ? undefined : prefix.replace(/^\//, '');
    if (fixedPrefix) {
        fixedPrefix = fixedPrefix.endsWith('/') ? fixedPrefix : `${fixedPrefix}/`;
    }
    const objects = await s3.listObjectsV2({
        Bucket: bucket, Prefix: fixedPrefix, Delimiter: '/', ContinuationToken: continuationToken
    }).promise();

    // remove trailing slash if exists, and the starting prefix
    const updatedDirs = compact(dirs.concat(compact(objects.CommonPrefixes.map(commonPrefix => commonPrefix.Prefix.replace(/\/$/, '').replace(fixedPrefix, '')))));
    if (objects.IsTruncated) {
        return listBucketFolders(region, bucket, prefix, objects.NextContinuationToken, updatedDirs);
    }
    return updatedDirs;
}

async function listObjects(region, bucket, prefix, continuationToken, dirsAndFiles = []) {
    const s3 = getS3(region);

    let fixedPrefix = isNil(prefix) || prefix === '' ? undefined : prefix;
    if (fixedPrefix) {
        fixedPrefix = fixedPrefix.endsWith('/') ? fixedPrefix : `${fixedPrefix}/`;
    }
    const objects = await s3.listObjectsV2({
        Bucket: bucket, Prefix: fixedPrefix, Delimiter: '/', ContinuationToken: continuationToken
    }).promise();

    const directoriesPromises = objects.CommonPrefixes.map(async resultPrefix => getObjectsAndTags(region, bucket, resultPrefix.Prefix, 0, TYPE_DIRECTORY, 16822));
    const filePromises = objects.Contents.map(async obj => getObjectsAndTags(region, bucket, obj.Key, obj.Size, TYPE_FILE, 33206, obj.LastModified));
    const result = await Promise.all(filePromises.concat(compact(directoriesPromises)));

    const updatedDirsAndFiles = compact(dirsAndFiles.concat(compact(result)));
    if (objects.IsTruncated) {
        return listObjects(region, bucket, prefix, objects.NextContinuationToken, updatedDirsAndFiles);
    }
    return updatedDirsAndFiles;
}

async function deleteObject(region, bucket, key) {
    const s3 = getS3(region);
    return s3.deleteObject({ Bucket: bucket, Key: key }).promise();
}

function getS3ReadStream(region, bucket, key) {
    const s3 = getS3(region);
    return s3.getObject({ Bucket: bucket, Key: key }).createReadStream();
}

async function getS3Object(region, bucket, key) {
    const s3 = getS3(region);
    const data = await s3.getObject({ Bucket: bucket, Key: key }).promise();
    return data.Body;
}

function upload(region, bucket, bodyStream, key, tags, kmsKeyId, storageClass) {
    const contentType = mimeType.lookup(key) || 'application/octet-stream';
    const s3 = getS3(region);
    const params = {
        Bucket: bucket,
        Key: key,
        Body: bodyStream,
        ContentType: contentType,
        ACL: 'bucket-owner-full-control',
        StorageClass: storageClass || 'STANDARD'
    };

    if (kmsKeyId) {
        params.SSEKMSKeyId = kmsKeyId;
        params.ServerSideEncryption = 'aws:kms';
    }

    return new AWS.S3.ManagedUpload({
        partSize: bytes(config.get('protocols.s3.partSize')),
        queueSize: config.get('protocols.s3.queueSize'),
        params,
        leavePartsOnError: false,
        tags,
        service: s3
    });
}

function calculateObjectStoragePath(dir, path) {
    return dir ? `${dir}/${path}` : path;
}

function createTagsArray(requestTags, entry) {
    const convertTag = {
        'netapp-cs-uid': 'uid', 'netapp-cs-gid': 'gid', 'netapp-cs-mode': 'mode', 'netapp-cs-mtimeSec': 'mtimeSec'
    };

    const propertiesToTags = keys(convertTag).map(property => ({ Key: property, Value: `${entry[convertTag[property]]}` }));
    if (entry.type === TYPE_SYMBOLIC) {
        propertiesToTags.push({ Key: 'netapp-cs-slink', Value: 'True' });
    }

    return requestTags ? requestTags.concat(propertiesToTags) : propertiesToTags;
}

function convertTagsArrayToString(tags) {
    const tagsObj = tags.reduce((accumulator, item) => {
        accumulator[item.Key] = item.Value;
        return accumulator;
    }, {});
    return stringify(tagsObj);
}

async function createMultipartUpload(region, targetBucket, targetKey, encryptionKey, tagsToAdd, storageClass) {
    const s3 = getS3(region);
    const multipartUpload = await s3.createMultipartUpload({
        Bucket: targetBucket,
        Key: targetKey,
        StorageClass: storageClass || 'STANDARD',
        ACL: 'bucket-owner-full-control',
        SSEKMSKeyId: encryptionKey,
        ServerSideEncryption: (encryptionKey) ? 'aws:kms' : undefined,
        Tagging: tagsToAdd
    }).promise();

    return multipartUpload.UploadId;
}

async function completeMultipartUpload(region, targetBucket, targetKey, copyPartResults, uploadId) {
    const s3 = getS3(region);
    return s3.completeMultipartUpload(
        {
            Bucket: targetBucket,
            Key: targetKey,
            MultipartUpload: {
                Parts: copyPartResults
            },
            UploadId: uploadId
        }
    ).promise();
}

async function copyPart(region, targetBucket, targetKey, copySource, partNum, uploadId, startBytes, endBytes) {
    const s3 = getS3(region);
    return s3.uploadPartCopy({
        Bucket: targetBucket,
        CopySource: copySource,
        Key: targetKey,
        PartNumber: partNum,
        UploadId: uploadId,
        CopySourceRange: `bytes=${startBytes}-${endBytes}`
    }).promise();
}


async function copyMultipartObject(region, targetBucket, targetKey, copySource, sourceObjectSize, uploadId) {
    // calculate size
    const partSize = bytes(config.get('protocols.s3.part-size-s3-to-s3'));
    const rangeArr = range(0, sourceObjectSize, partSize);
    rangeArr.push(sourceObjectSize);
    // create array of objects - contains the start and end of each range and the partNumber
    const copyParts = zip(dropRight(rangeArr, 1), tail(rangeArr)).map((bytesPair, index) => ({
        bytesPair,
        partNum: index + 1
    }));

    return Promise.map(copyParts, (async partDetails => {
        const nextStartBytesRange = first(partDetails.bytesPair);
        const partNumber = partDetails.partNum;
        const nextEndBytesRange = last(partDetails.bytesPair) - 1; // because the last in the range in inclusive
        try {
            const result = await copyPart(region, targetBucket, targetKey, copySource, partNumber, uploadId, nextStartBytesRange, nextEndBytesRange);
            return {
                ETag: result.CopyPartResult.ETag,
                PartNumber: partNumber
            };
        } catch (err) {
            logger.error(`Failed upload part number ${partNumber} for target from ${copySource} to ${targetKey} with error ${err}`);
            throw err;
        }
    }), { concurrency: 10 }); // when there are a lot of parts - receive reduce your rate
}

async function copyObject(region, destinationBucket, copySource, destinationKey, storageClass, tagsToAdd, kmsKeyId, entrySize) {
    const s3 = getS3(region);
    const tags = convertTagsArrayToString(tagsToAdd);
    if (entrySize < bytes(config.get('protocols.s3.part-size-s3-to-s3'))) {
        const params = {
            Bucket: destinationBucket,
            CopySource: copySource,
            Key: destinationKey,
            StorageClass: storageClass || 'STANDARD',
            SSEKMSKeyId: kmsKeyId,
            ServerSideEncryption: (kmsKeyId) ? 'aws:kms' : undefined,
            ACL: 'bucket-owner-full-control',
            TaggingDirective: 'REPLACE',
            Tagging: tags
        };
        await s3.copyObject(params).promise();
    } else {
        // create multipart upload
        const uploadId = await createMultipartUpload(region, destinationBucket, destinationKey, kmsKeyId, tags, storageClass);
        const copyPartResults = await copyMultipartObject(region, destinationBucket, destinationKey, copySource, entrySize, uploadId);
        // send complete multipart upload request
        await completeMultipartUpload(region, destinationBucket, destinationKey, copyPartResults, uploadId);
    }
}


async function copyDirectory(region, destinationBucket, copySource, destinationKey, storageClass, tagsToAdd, kmsKeyId) {
    const tags = convertTagsArrayToString(tagsToAdd);
    const s3 = getS3(region);
    try {
        const params = {
            Bucket: destinationBucket,
            CopySource: copySource,
            Key: destinationKey,
            StorageClass: storageClass || 'STANDARD',
            SSEKMSKeyId: kmsKeyId,
            ServerSideEncryption: kmsKeyId ? 'aws:kms' : undefined,
            ACL: 'bucket-owner-full-control',
            TaggingDirective: 'REPLACE',
            Tagging: tags
        };
        await s3.copyObject(params).promise();
    } catch (err) {
        if (err.code === 'NoSuchKey') { // in case we didn't create the directory
            logger.warn(`Received NoSuchKey error on key ${copySource} on Directory, the user created the folder`);
        } else {
            throw err;
        }
    }
}

module.exports = {
    listBuckets,
    headBucket,
    listObjects,
    listBucketFolders,
    deleteObject,
    getS3ReadStream,
    getS3Object,
    upload,
    calculateObjectStoragePath,
    createTagsArray,
    getBucketTagging,
    getBucketLocation,
    getObjectTagging,
    copyObject,
    copyDirectory
};
