// http://azure.github.io/azure-storage-node/BlobService.html
const azure = require('azure-storage');
const mime = require('mime-types');
const ms = require('ms');
const config = require('config');
const {
    last, compact, map, sortBy
} = require('lodash');
const {
    TYPE_FILE, TYPE_DIRECTORY, TYPE_SYMBOLIC, MOUNT_TARGET_SUFFIX, MOUNT_SOURCE_SUFFIX, ABORT_ERROR
} = require('../consts');
const { getMetadata } = require('../metadata-store.js');

const getLogger = require('../logger/logger');

const logger = getLogger();

async function getBlobService(relationshipId, enumDirection) {
    const metadataConnectionString = await getMetadata(`${relationshipId}.${enumDirection}.azure`);
    if (metadataConnectionString) {
        return azure.createBlobService(metadataConnectionString);
    }
    throw new Error('Failed to find connection-string');
}

async function getContainerProperties(blobService, container) {
    return new Promise((resolve, reject) => {
        blobService.getContainerProperties(container, (error, result) => {
            if (error) {
                reject(error);
            } else {
                resolve(result);
            }
        });
    });
}

async function listContainersSegmented(blobService) {
    return new Promise((resolve, reject) => {
        blobService.listContainersSegmented(undefined, (error, results) => {
            if (error) {
                reject(error);
            } else {
                resolve(results.entries);
            }
        });
    });
}

async function listContainers(connectionString) {
    const blobService = azure.createBlobService(connectionString);
    const containers = await listContainersSegmented(blobService);

    const promises = containers.map(async container => {
        const containerProperties = await getContainerProperties(blobService, container.name);
        return {
            name: containerProperties.name,
            tags: map(containerProperties.metadata, (value, key) => ({ Key: key, Value: value }))
        };
    });

    const result = await Promise.all(promises);
    return sortBy(result, 'name');
}

async function listBlobDirectories(container, prefix, connectionString, currentToken = undefined, accumulator = []) {
    const blobService = azure.createBlobService(connectionString);

    return new Promise((resolve, reject) => {
        blobService.listBlobDirectoriesSegmentedWithPrefix(container, prefix ? `${prefix}/` : undefined, currentToken, async (error, result) => {
            if (error) {
                reject(error);
            } else {
                const updatedAccumulator = accumulator.concat(result.entries);

                if (result.continuationToken) {
                    try {
                        resolve(await listBlobDirectories(container, prefix, connectionString, result.continuationToken, updatedAccumulator));
                    } catch (err) {
                        reject(err);
                    }
                }

                resolve(updatedAccumulator);
            }
        });
    });
}

async function listBlobDirectoriesSegmentedWithPrefix(container, prefix, relationshipId, srcOrtrg, currentToken = undefined, accumulator = []) {
    const blobService = await getBlobService(relationshipId, srcOrtrg);
    return new Promise((resolve, reject) => {
        blobService.listBlobDirectoriesSegmentedWithPrefix(container, prefix ? `${prefix}/` : undefined, currentToken, async (error, result) => {
            if (error) {
                reject(error);
            } else {
                const updatedAccumulator = accumulator.concat(result.entries);

                if (result.continuationToken) {
                    try {
                        resolve(await listBlobDirectoriesSegmentedWithPrefix(container, prefix, relationshipId, srcOrtrg, result.continuationToken, updatedAccumulator));
                    } catch (err) {
                        reject(err);
                    }
                }

                resolve(updatedAccumulator);
            }
        });
    });
}

async function listBlobsSegmentedWithPrefix(container, prefix, relationshipId, srcOrtrg, currentToken = undefined, accumulator = []) {
    const blobService = await getBlobService(relationshipId, srcOrtrg);
    return new Promise((resolve, reject) => {
        blobService.listBlobsSegmentedWithPrefix(container, prefix ? `${prefix}/` : undefined, currentToken, { delimiter: '/', include: azure.BlobUtilities.BlobListingDetails.METADATA }, async (error, result) => {
            if (error) {
                reject(error);
            } else {
                const updatedAccumulator = accumulator.concat(result.entries);

                if (result.continuationToken) {
                    try {
                        resolve(await listBlobsSegmentedWithPrefix(container, prefix, relationshipId, srcOrtrg, result.continuationToken, updatedAccumulator));
                    } catch (err) {
                        reject(err);
                    }
                }

                resolve(updatedAccumulator);
            }
        });
    });
}

async function getBlobMetadata(relationshipId, srcOrtrg, container, blob) {
    const blobService = await getBlobService(relationshipId, srcOrtrg);
    return new Promise((resolve, reject) => {
        blobService.getBlobMetadata(container, blob, (error, result) => {
            if (error) {
                reject(error);
            } else {
                resolve(result);
            }
        });
    });
}

async function listBlobs(container, prefix, relationshipId, srcOrtrg) {
    const [directories, files] = await Promise.all([listBlobDirectoriesSegmentedWithPrefix(container, prefix, relationshipId, srcOrtrg),
        listBlobsSegmentedWithPrefix(container, prefix, relationshipId, srcOrtrg)]);

    const filesMetadata = compact(files.map(file => {
        const path = last(file.name.split('/'));
        // blob might be without name - displayed as "<no name>" in azure console
        if (path) {
            return {
                path,
                type: file.metadata.netappcsslink === 'True' ? TYPE_SYMBOLIC : TYPE_FILE,
                size: parseInt(file.contentLength),
                mtimeSec: parseInt(file.metadata.netappcsmtimesec) || parseInt(new Date(file.lastModified) / 1000),
                mode: parseInt(file.metadata.netappcsmode) || 33206,
                uid: parseInt(file.metadata.netappcsuid) || 0,
                gid: parseInt(file.metadata.netappcsgid) || 0
            };
        }
        return undefined;
    }));


    const directoriesMetadataPromises = directories.map(async directory => {
        let directoryMetadata;
        try {
            directoryMetadata = await getBlobMetadata(relationshipId, srcOrtrg, container, directory.name);
        } catch (err) {
            if (err.code !== 'NotFound') {
                throw err;
            }
        }

        return {
            path: last(compact(directory.name.split('/'))),
            type: TYPE_DIRECTORY,
            size: 0,
            mtimeSec: directoryMetadata ? parseInt(directoryMetadata.metadata.netappcsmtimesec) || parseInt(new Date(directoryMetadata.lastModified) / 1000) : 0,
            mode: directoryMetadata ? parseInt(directoryMetadata.metadata.netappcsmode) : 16822,
            uid: directoryMetadata ? parseInt(directoryMetadata.metadata.netappcsuid) : 0,
            gid: directoryMetadata ? parseInt(directoryMetadata.metadata.netappcsgid) : 0
        };
    });

    const directoriesMetadata = await Promise.all(directoriesMetadataPromises);

    return filesMetadata.concat(directoriesMetadata);
}

async function deleteBlob(container, blob, relationshipId) {
    const blobService = await getBlobService(relationshipId, MOUNT_TARGET_SUFFIX);
    return new Promise((resolve, reject) => {
        blobService.deleteBlob(container, blob, (error, result) => {
            if (error) {
                reject(error);
            } else {
                resolve(result);
            }
        });
    });
}

async function getBlobToText(container, blob, relationshipId) {
    const blobService = await getBlobService(relationshipId, MOUNT_SOURCE_SUFFIX);
    return new Promise((resolve, reject) => {
        blobService.getBlobToText(container, blob, (error, result) => {
            if (error) {
                reject(error);
            } else {
                resolve(result);
            }
        });
    });
}

async function createBlockBlobFromText(container, blob, text, metadata, relationshipId) {
    const blobService = await getBlobService(relationshipId, MOUNT_TARGET_SUFFIX);
    return new Promise((resolve, reject) => {
        blobService.createBlockBlobFromText(container, blob, text, { metadata, contentSettings: { contentType: 'text/plain' } }, (error, result) => {
            if (error) {
                reject(error);
            } else {
                resolve(result);
            }
        });
    });
}

async function createBlobReadStream(container, blob, relationshipId) {
    const blobService = await getBlobService(relationshipId, MOUNT_SOURCE_SUFFIX);
    return blobService.createReadStream(container, blob);
}

async function createWriteStreamToBlockBlob(container, blob, metadata, relationshipId) {
    const contentType = mime.lookup(blob) || 'application/octet-stream';
    const blobService = await getBlobService(relationshipId, MOUNT_TARGET_SUFFIX);
    const writeStream = blobService.createWriteStreamToBlockBlob(container, blob, { metadata, contentSettings: { contentType } });
    // we are using pump which uses end-of-stream in transferrer
    // azure stream is a custom stream without _writableState object -> end-of-stream issues "premature close" error in such case
    // https://github.com/mafintosh/end-of-stream/issues/11
    writeStream._writableState = { ended: true }; // eslint-disable-line no-underscore-dangle

    return writeStream;
}

async function setBlobTier(container, blob, relationshipId, blobTier) {
    if (blobTier !== 'HOT') {
        const blobService = await getBlobService(relationshipId, MOUNT_TARGET_SUFFIX);
        return new Promise((resolve, reject) => {
            blobService.setBlobTier(container, blob, blobTier, (error, response) => {
                if (error) {
                    reject(error);
                } else {
                    resolve(response);
                }
            });
        });
    }
}

function createMetadata(requestMetadata, entry) {
    const metadata = {
        netappcsuid: entry.uid,
        netappcsgid: entry.gid,
        netappcsmode: entry.mode,
        netappcsmtimesec: entry.mtimeSec
    };

    if (entry.type === TYPE_SYMBOLIC) {
        metadata.netappcsslink = 'True';
    }

    if (requestMetadata) {
        requestMetadata.forEach(rm => { metadata[rm.Key] = rm.Value; });
    }

    return metadata;
}

function createSasUrl(storageConnectionString, targetContainerName, targetBlobName) {
    const sharedAccessSignaturePattern = /SharedAccessSignature=(.*?)(;|$)/;
    const blobEndPointPattern = /BlobEndpoint=(.*?)(;|$)/;
    const sas = sharedAccessSignaturePattern.exec(storageConnectionString) ? sharedAccessSignaturePattern.exec(storageConnectionString)[1] : undefined;
    if (!sas) {
        throw new Error('source connection-string should contain be  SAS Connection-string and not regular connection-string');
    }
    const blobEndPoint = blobEndPointPattern.exec(storageConnectionString) ? blobEndPointPattern.exec(storageConnectionString)[1] : undefined;

    if (!blobEndPoint) {
        throw new Error('An error occurred- received invalid connection string, Expected a connection-string with BlobEndPoint');
    }
    return `${blobEndPoint}${targetContainerName}/${targetBlobName}?${sas}`;
}

function timeout(milliseconds) {
    return new Promise(resolve => setTimeout(resolve, milliseconds));
}

async function getBlobProperties(relationshipId, srcOrTrg, containerName, blobName) {
    const blobService = await getBlobService(relationshipId, srcOrTrg);
    return new Promise((resolve, reject) => {
        blobService.getBlobProperties(containerName, blobName, async (error, result) => {
            if (error) {
                reject(error);
            } else {
                resolve(result);
            }
        });
    });
}

async function waitForCopy(relationshipId, containerName, blobName) {
    const timeoutMilliSeconds = ms(config.get('protocols.azure.wait-timeout'));
    await timeout(timeoutMilliSeconds);
    const result = await getBlobProperties(relationshipId, MOUNT_TARGET_SUFFIX, containerName, blobName);
    const copyStatus = result.copy ? result.copy.status : 'failed';
    if (copyStatus === 'success') {
        return copyStatus;
    } if (copyStatus === 'failed') {
        logger.info(`[${relationshipId}] failed to copy blob ${blobName} received result ${result}`);
        throw new Error(`Failed to copy, Blob copy status is ${copyStatus}`);
    } else if (copyStatus === 'pending') {
        return waitForCopy(relationshipId, containerName, blobName);
    } else if (copyStatus === 'aborted') {
        const error = new Error(`item: ${blobName} is aborted`);
        error.code = ABORT_ERROR;
        throw error;
    } else {
        throw new Error(`Received unknown status ${copyStatus}`);
    }
}

async function startBlobCopy(relationshipId, sourceContainerName, targetContainerName, sourceBlobName, targetBlobName) {
    const blobService = await getBlobService(relationshipId, MOUNT_TARGET_SUFFIX);
    const sourceStorageConnectionString = await getMetadata(`${relationshipId}.${MOUNT_SOURCE_SUFFIX}.azure`);
    const startCopyBlobResult = await new Promise((resolve, reject) => {
        const sourceSasUrl = createSasUrl(sourceStorageConnectionString, sourceContainerName, sourceBlobName);
        blobService.startCopyBlob(sourceSasUrl, targetContainerName, targetBlobName, async (error, result) => {
            if (error) {
                reject(error);
            } else {
                resolve(result);
            }
        });
    });
    const { copy: { id: copyId } } = startCopyBlobResult;
    return copyId;
}


async function setBlobMetadata(relationshipId, srgOrTrg, containerName, blobName, metadata) {
    const blobService = await getBlobService(relationshipId, srgOrTrg);
    return new Promise((resolve, reject) => {
        blobService.setBlobMetadata(containerName, blobName, metadata, async (error, result) => {
            if (error) {
                reject(error);
            } else {
                resolve(result);
            }
        });
    });
}

async function abortCopyBlob(relationshipId, containerName, blobName, copyId) {
    const blobService = await getBlobService(relationshipId, MOUNT_TARGET_SUFFIX);
    return new Promise((resolve, reject) => {
        blobService.abortCopyBlob(containerName, blobName, copyId, async (error, result) => {
            if (error) {
                reject(error);
            } else {
                resolve(result);
            }
        });
    });
}

module.exports = {
    listContainers,
    listBlobDirectories,
    listBlobs,
    deleteBlob,
    createBlobReadStream,
    getBlobToText,
    createWriteStreamToBlockBlob,
    createMetadata,
    createBlockBlobFromText,
    startBlobCopy,
    setBlobMetadata,
    getBlobMetadata,
    waitForCopy,
    abortCopyBlob,
    setBlobTier
};
