const {
    readdir, lstat, unlink, mkdirs, chmod, chown
} = require('fs-extra');
const { join } = require('path');
const { compact } = require('lodash');
const Promise = require('bluebird');
const execute = require('../executor');
const getLogger = require('../logger/logger');

const logger = getLogger();

const {
    TYPE_FILE, TYPE_DIRECTORY, TYPE_SYMBOLIC, TYPE_OTHER
} = require('../consts');

function gracefulJoin(...paths) {
    return join(...compact(paths));
}

function getType(stats) {
    let type = TYPE_OTHER;

    if (stats.isFile()) {
        type = TYPE_FILE;
    } else if (stats.isDirectory()) {
        type = TYPE_DIRECTORY;
    } else if (stats.isSymbolicLink()) {
        type = TYPE_SYMBOLIC;
    }

    return type;
}

async function copyNfsAcl(sourcePath, targetPath) {
    const command = `nfs4_getfacl ${sourcePath}| nfs4_setfacl -S - ${targetPath}`;
    await execute(command);
}

async function mkdirsChmodAndChown(path, entry) {
    const targetPath = gracefulJoin(path, entry.path);
    await mkdirs(targetPath);
    await Promise.all([chmod(targetPath, entry.mode), chown(targetPath, entry.uid, entry.gid)]);
}

async function chownIfNeeded(path, uid, gid) {
    if (process.getuid() !== uid || process.getgid() !== gid) {
        try {
            await chown(path, uid, gid);
        } catch (err) {
            if (err.code === 'EINVAL') {
                // squashed NFS mount
                logger.warn('Unable to chown due to EINVAL', { path, uid, gid });
            } else {
                throw err;
            }
        }
    }
}

async function stat(path, entry) {
    const stats = await lstat(join(path, entry));
    const type = getType(stats);

    return {
        path: entry,
        type,
        size: (type === TYPE_DIRECTORY) ? 0 : stats.size,
        mode: stats.mode,
        uid: stats.uid,
        gid: stats.gid,
        mtimeSec: parseInt(stats.mtimeMs / 1000)
    };
}

async function list(path) {
    const entries = await readdir(path);
    return Promise.map(entries, entry => stat(path, entry), { concurrency: 100 });
}

async function removeCoreFiles() {
    logger.info('Looking for core files...');
    try {
        const entries = await list('.');
        const coreFiles = entries.filter(entry => entry.type === TYPE_FILE && /^core\.\d+$/.test(entry.path));
        logger.info(`${coreFiles.length} core files found`);
        const promises = coreFiles.map(coreFile => unlink(coreFile.path));
        await Promise.all(promises);
    } catch (err) {
        logger.warn('Failed to remove core files', err);
    }
}

module.exports = {
    list, gracefulJoin, removeCoreFiles, mkdirsChmodAndChown, chownIfNeeded, copyNfsAcl
};
