const { Storage } = require('@google-cloud/storage');
const mimeType = require('mime-types');
const {
    keys, last, compact, isNil
} = require('lodash');
const {
    TYPE_FILE, TYPE_DIRECTORY, TYPE_SYMBOLIC
} = require('../consts');
const getLogger = require('../logger/logger');
const { getMetadata } = require('../metadata-store.js');

const logger = getLogger();

async function getGcpCredentials(relationshipId, enumDirection) {
    const credentials = await getMetadata(`${relationshipId}.${enumDirection}.gcp`);
    const { projectId, clientEmail, privateKey } = credentials;

    return { projectId, clientEmail, privateKey };
}

function getStorage(credentials) {
    return new Storage({
        projectId: credentials.projectId,
        credentials: {
            client_email: credentials.clientEmail,
            private_key: credentials.privateKey
        }
    });
}

async function getBucketLocationAndTags(credentials, bucketName) {
    const storage = getStorage(credentials);
    const bucket = storage.bucket(bucketName);
    const metadata = await bucket.getMetadata();
    const { location, labels: tags } = metadata[0];
    const tagsArr = tags ? keys(tags).map(key => ({ Key: key, Value: tags[key] })) : [];

    return { location, tags: tagsArr };
}

async function listGcpBuckets(credentials) {
    const storage = getStorage(credentials);
    const buckets = await storage.getBuckets();
    const promises = buckets[0].map(async bucket => {
        try {
            const { location, tags } = await getBucketLocationAndTags(credentials, bucket.name);
            return { name: bucket.name, region: location, tags };
        } catch (err) {
            logger.warn(`Failed to get bucket location/tagging, bucket name ${bucket.name}`, err);
        }
    });

    return Promise.all(promises);
}

async function deleteGcpObject(credentials, bucket, file) {
    const storage = getStorage(credentials);
    return storage.bucket(bucket).file(file).delete();
}

function getGcpReadStream(credentials, bucket, file) {
    const storage = getStorage(credentials);
    return storage.bucket(bucket).file(file).createReadStream({ validation: false });
}

function getGcpWriteStream(credentials, bucket, file, tags) {
    const contentType = mimeType.lookup(file) || 'application/octet-stream';
    const storage = getStorage(credentials);
    const options = {
        resumable: false,
        metadata: {
            contentType,
            metadata: tags
        }
    };
    return storage.bucket(bucket).file(file).createWriteStream(options);
}
async function isObjectExists(credentials, bucketName, fileName) {
    const storage = getStorage(credentials);
    const objectExist = await storage.bucket(bucketName).file(fileName).exists();
    return objectExist[0];
}

async function getGcpObjectTagging(credentials, bucketName, fileName, objectType) {
    const storage = getStorage(credentials);
    let noNameFileExists = false;
    const noNameFile = '<no name>';

    // the purpose of the if is to avoid redundant exists() calls in case of TYPE_FILE/SYMBOLIC
    if (objectType === TYPE_DIRECTORY) { // if the object is a directory - we need to check if there's a <no name> file that contains the directory's permissions
        noNameFileExists = await isObjectExists(credentials, bucketName, `${fileName}/${noNameFile}`);
    }
    const file = noNameFileExists ? storage.bucket(bucketName).file(`${fileName}/${noNameFile}`) : storage.bucket(bucketName).file(fileName);

    try {
        const metadata = await file.getMetadata();
        const tags = metadata[0].metadata;
        return tags ? keys(tags).map(key => ({ Key: key, Value: tags[key] })) : [];
    } catch (err) {
        if (err.code === 404) {
            return [];
        }
        throw err;
    }
}

async function createGcpFolder(credentials, bucketName, folderName, tags) {
    const contentType = mimeType.lookup(folderName) || 'application/octet-stream';
    const storage = getStorage(credentials);
    const metadata = {
        contentType,
        metadata: tags
    };

    await storage.bucket(bucketName).file(`${folderName}/`).save('');
    const noNameFile = await storage.bucket(bucketName).file(`${folderName}/<no name>`);
    await noNameFile.save('');
    return noNameFile.setMetadata(metadata);
}

async function createGcpSymlink(credentials, bucketName, fileName, fileContent, tags) {
    const contentType = mimeType.lookup(fileName) || 'application/octet-stream';
    const storage = getStorage(credentials);
    const options = {
        metadata: {
            contentType,
            metadata: tags
        }
    };
    return storage.bucket(bucketName).file(fileName).save('', options);
}

async function getGcpObjectMetadata(credentials, bucketName, fileName) {
    const storage = getStorage(credentials);
    const noNameFileExists = await isObjectExists(credentials, bucketName, `${fileName}/<no name>`);
    const file = noNameFileExists ? storage.bucket(bucketName).file(`${fileName}/<no name>`) : storage.bucket(bucketName).file(fileName);
    const result = await file.getMetadata();
    return result[0];
}

function createGcpMetadata(requestMetadata, entry, symlinkContent) {
    const metadata = {
        'netapp-cs-uid': entry.uid,
        'netapp-cs-gid': entry.gid,
        'netapp-cs-mode': entry.mode,
        'netapp-cs-mtimeSec': entry.mtimeSec
    };

    if (entry.type === TYPE_SYMBOLIC) {
        metadata['netapp-cs-slink'] = 'True';
        metadata.symlink = symlinkContent;
    }

    if (requestMetadata) {
        requestMetadata.forEach(rm => { metadata[rm.Key] = rm.Value; });
    }

    return metadata;
}

async function getGcpObjectsAndTags(credentials, bucket, key, size, objectType, mode, lastModified) {
    const tagsArray = await getGcpObjectTagging(credentials, bucket, key, objectType);
    const tagsObject = tagsArray.reduce((accumulator, currentValue) => (Object.assign(accumulator, { [currentValue.Key]: currentValue.Value })), {});
    const modifiedTime = lastModified === undefined ? Date.now() : lastModified.getTime();

    if (!key.endsWith('/') || objectType !== TYPE_FILE) {
        return {
            path: key === '/' ? '/' : last(compact(key.split('/'))),
            type: tagsObject['netapp-cs-slink'] === 'True' ? TYPE_SYMBOLIC : objectType,
            size: parseInt(size),
            mtimeSec: parseInt(tagsObject['netapp-cs-mtimeSec']) || parseInt(modifiedTime / 1000),
            mode: parseInt(tagsObject['netapp-cs-mode']) || mode,
            uid: parseInt(tagsObject['netapp-cs-uid']) || 0,
            gid: parseInt(tagsObject['netapp-cs-gid']) || 0
        };
    }

    return null;
}

async function listGcpBucketFolders(credentials, bucketName, prefix, nextQuery, dirs = []) {
    const storage = getStorage(credentials);
    const bucket = storage.bucket(bucketName);
    let fixedPrefix = isNil(prefix) || prefix === '' ? undefined : prefix.replace(/^\//, '');
    if (fixedPrefix) {
        fixedPrefix = fixedPrefix.endsWith('/') ? fixedPrefix : `${fixedPrefix}/`;
    }

    const response = await bucket.getFiles({
        prefix: fixedPrefix, autoPaginate: false, pageToken: nextQuery, delimiter: '/'
    });
    const responseDirs = response[2].prefixes ? response[2].prefixes : [];
    const updatedDirs = compact(dirs.concat(compact(responseDirs.map(dir => dir.replace(fixedPrefix, '').replace(/\/$/, '')))));
    if (response[2].nextPageToken) {
        return listGcpBucketFolders(credentials, bucket, fixedPrefix, response[2].nextPageToken, updatedDirs);
    }

    return updatedDirs;
}

async function listGcpObjects(credentials, bucketName, prefix, nextQuery, dirsAndFiles = []) {
    const storage = getStorage(credentials);
    const bucket = storage.bucket(bucketName);
    let fixedPrefix = isNil(prefix) || prefix === '' ? undefined : prefix;
    if (fixedPrefix) {
        fixedPrefix = fixedPrefix.endsWith('/') ? fixedPrefix : `${fixedPrefix}/`;
    }

    const response = await bucket.getFiles({
        prefix: fixedPrefix, autoPaginate: false, pageToken: nextQuery, delimiter: '/'
    });
    const responseFiles = compact(response[0].map(file => file.name.replace(fixedPrefix, '').replace(/\/$/, '')).filter(file => file !== '<no name>'));
    let responseDirs = response[2].prefixes ? response[2].prefixes : [];
    responseDirs = compact(responseDirs.map(dir => dir.replace(fixedPrefix, '').replace(/\/$/, '')));

    const filesPromises = responseFiles.map(async file => {
        const filePath = fixedPrefix ? fixedPrefix + file : file;
        const metadata = await getGcpObjectMetadata(credentials, bucketName, filePath);
        let mtimeSec = metadata.metadata ? new Date(metadata.metadata['netapp-cs-mtimeSec']) : undefined;
        if (mtimeSec === undefined) {
            mtimeSec = new Date(metadata.updated);
        }
        return getGcpObjectsAndTags(credentials, bucketName, filePath, metadata.size, TYPE_FILE,
            33206, mtimeSec);
    });

    const directoriesPromises = responseDirs.map(async dir => {
        const dirPath = fixedPrefix ? fixedPrefix + dir : dir;
        return getGcpObjectsAndTags(credentials, bucketName, dirPath, 0, TYPE_DIRECTORY, 16822);
    });

    const result = await Promise.all(filesPromises.concat(compact(directoriesPromises)));
    const updatedDirsAndFiles = compact(dirsAndFiles.concat(compact(result)));
    if (response[1] !== null && 'pageToken' in response[1]) {
        return listGcpObjects(credentials, bucketName, prefix, response[1].pageToken, updatedDirsAndFiles);
    }

    return updatedDirsAndFiles;
}

async function isGcpBucketExists(credentials, bucketName) {
    const storage = getStorage(credentials);
    const bucket = storage.bucket(bucketName);
    const result = await bucket.exists();
    if (result[0] === false) {
        throw new Error(`Bucket ${bucketName} doesn't exist`);
    }
    return true;
}

module.exports = {
    getGcpCredentials,
    listGcpBuckets,
    deleteGcpObject,
    getGcpReadStream,
    getGcpWriteStream,
    createGcpSymlink,
    createGcpFolder,
    createGcpMetadata,
    getGcpObjectTagging,
    listGcpBucketFolders,
    listGcpObjects,
    isGcpBucketExists,
    getBucketLocationAndTags,
    getGcpObjectMetadata,
    isObjectExists
};
