const util = require('util');
const { readFileSync } = require('fs');
const log4js = require('log4js');
const cluster = require('cluster');
const config = require('config');
const chokidar = require('chokidar');
const {
    set, isPlainObject, isArray, isString, cloneDeep, isError
} = require('lodash');
const { format } = require('util');
const { DATA_BROKER_RUNTIME_ABOUT, DATA_BROKER_RUNTIME_CONFIG } = require('../consts');

// override default break length which is 80
util.inspect.defaultOptions.breakLength = Infinity;

const LOG4JS_CONFIG_FILE = config.get('log4js-config-file');

const SECRET_KEYS = ['password', 'privateKey', 'storageAccountConnectionString', 'secretKey', 'PASSWD', 'certificate'];

function hideSecretsHelper(obj) {
    if (isArray(obj)) {
        obj.forEach(arrayObj => hideSecretsHelper(arrayObj));
    } else if (isPlainObject(obj)) {
        Object.keys(obj).forEach(key => {
            if (isPlainObject(obj[key]) || isArray(obj[key])) {
                hideSecretsHelper(obj[key]);
            } else if (SECRET_KEYS.includes(key)) {
                set(obj, `${key}`, '******');
            }
        });
    }
}

// logEventData - can be anything, string, object or array
function hideSecrets(logEventData) {
    try {
        const jsonObj = JSON.parse(logEventData);
        hideSecretsHelper(jsonObj);
        return JSON.stringify(jsonObj);
    } catch (err) {
        if (!isString(logEventData)) {
            hideSecretsHelper(logEventData);
        }
        return logEventData;
    }
}

/*
* the function on message value used to return hideSecret(obj) only.
* in some cases, while we were debugging stuff and used prints with logger, that changed objects in memory, because
* it uses set function and changed the value of essential values (certificate, passwords etc).
* now every time we print an object, we clone it and then hide secret keys.
* */
function configureLayoutWithToken() {
    const configFileContent = JSON.parse(readFileSync(LOG4JS_CONFIG_FILE));
    Object.keys(configFileContent.appenders).forEach(appenderType => {
        const { layout } = configFileContent.appenders[appenderType];
        if (layout) {
            const { pattern } = layout;
            if (pattern) {
                layout.tokens = {
                    message: logEvent => format(...logEvent.data.map(obj => {
                        // used to return hideSecrets(obj) without cloneDeep. explanation above.
                        if (isError(obj)) { // added isError because cloneDeep of Error is {} for error we shouldn't clone
                            return hideSecrets(obj);
                        }
                        const objClone = cloneDeep(obj);
                        return hideSecrets(objClone);
                    }))
                };
            }
        }
    });

    log4js.configure(configFileContent);
}

configureLayoutWithToken();

function getLogger(category) {
    let workerType = 'UNKNOWN';
    if (cluster.isMaster) {
        workerType = 'MASTER';
    } else if (cluster.worker.process.argv.find(arg => arg.includes('transferrer.js'))) {
        workerType = 'TRANSFERRER';
    } else if (cluster.worker.process.argv.find(arg => arg.includes('scanner.js'))) {
        workerType = 'SCANNER';
    } else if (cluster.worker.process.argv.find(arg => arg.includes('server.js'))) {
        workerType = 'SERVER';
    }

    const logger = log4js.getLogger(category);

    logger.addContext('worker-type', workerType);
    logger.addContext('worker-id', cluster.isMaster ? 0 : cluster.worker.id);
    logger.addContext('process-id', process.pid);
    logger.addContext('data-broker-version', DATA_BROKER_RUNTIME_ABOUT.version);
    logger.addContext('data-broker-id', DATA_BROKER_RUNTIME_CONFIG['data-broker-id']);
    logger.addContext('environment', DATA_BROKER_RUNTIME_CONFIG.environment);
    logger.addContext('deployment', DATA_BROKER_RUNTIME_CONFIG.type);
    logger.addContext('category', category || 'data-broker');

    return logger;
}

chokidar.watch(LOG4JS_CONFIG_FILE, { usePolling: { interval: 3000 } })
    .on('change', () => {
        try {
            getLogger().fatal('Got LOG4JS configuration change identified, LOG4JS is shutting down...');
            log4js.shutdown(() => {
                getLogger().fatal('Reloading LOG4JS configuration');
                configureLayoutWithToken();
                getLogger().fatal('LOG4JS configuration reloaded!');
            });
        } catch (err) {
            getLogger().fatal('Failed to change LOG4JS configuration', err);
        }
    });

module.exports = getLogger;
