const {
    set, get, isEmpty, merge
} = require('lodash');
const { writeFile, readFileSync, pathExistsSync } = require('fs-extra');
const getLogger = require('./logger/logger');

const logger = getLogger();
let META_DATA = {};
const META_DATA_FILE = './metadata.json';

// sync
async function syncMetadata() {
    if (pathExistsSync(META_DATA_FILE)) {
        try {
            META_DATA = JSON.parse(readFileSync(META_DATA_FILE));
        } catch (err) {
            logger.error('could not parse metadata.json', err);
        }
    } else {
        logger.error('metadata.json does not exist');
    }
}

// read
async function getMetadata(key) {
    if (!get(META_DATA, key)) { await syncMetadata(); }
    return get(META_DATA, key);
}

// write
async function updateMetadata(path, data) {
    // When booting, the META_DATA variable is empty and should sync the data from the metadata.json file
    if (isEmpty(META_DATA)) {
        await syncMetadata();
    }
    set(META_DATA, path, data);
    await writeFile(META_DATA_FILE, JSON.stringify(META_DATA, null, 4));
}

async function registerMetadata(source, target, sourcePassword, targetPassword, relationshipId) {
    if (source.azure) {
        const { storageAccountName, storageAccountConnectionString } = source.azure;
        if (storageAccountName && storageAccountConnectionString) {
            await updateMetadata(`${relationshipId}.src.azure`, storageAccountConnectionString);
        }
    }

    if (target.azure) {
        const { storageAccountName, storageAccountConnectionString } = target.azure;
        if (storageAccountName && storageAccountConnectionString) {
            await updateMetadata(`${relationshipId}.trg.azure`, storageAccountConnectionString);
        }
    }
    if (source.cifs) {
        const {
            username, workgroup
        } = source.cifs;
        if (username && sourcePassword) {
            await updateMetadata(`${relationshipId}.src.cifs`, { username, password: sourcePassword, workgroup });
        }
    }

    if (target.cifs) {
        const {
            username, workgroup
        } = target.cifs;
        if (username && targetPassword) {
            await updateMetadata(`${relationshipId}.trg.cifs`, { username, password: targetPassword, workgroup });
        }
    }

    if (source.sgws) {
        const {
            accessKey, secretKey
        } = source.sgws;
        if (accessKey && secretKey) {
            await updateMetadata(`${relationshipId}.src.sgws`, { accessKey, secretKey });
        }
    }

    if (target.sgws) {
        const {
            accessKey, secretKey
        } = target.sgws;
        if (accessKey && secretKey) {
            await updateMetadata(`${relationshipId}.trg.sgws`, { accessKey, secretKey });
        }
    }

    if (source.gcp) {
        const {
            projectId, clientEmail, privateKey
        } = source.gcp;
        if (projectId && clientEmail && privateKey) {
            await updateMetadata(`${relationshipId}.src.gcp`, { projectId, clientEmail, privateKey });
        }
    }

    if (target.gcp) {
        const {
            projectId, clientEmail, privateKey
        } = target.gcp;
        if (projectId && clientEmail && privateKey) {
            await updateMetadata(`${relationshipId}.trg.gcp`, { projectId, clientEmail, privateKey });
        }
    }
}

async function exportMetadata() {
    await syncMetadata();
    return META_DATA;
}

async function importMetadata(metadata) {
    await syncMetadata();
    // merge metadata
    META_DATA = merge(META_DATA, metadata);
    await writeFile(META_DATA_FILE, JSON.stringify(META_DATA, null, 4));
}

module.exports = {
    getMetadata,
    registerMetadata,
    exportMetadata,
    importMetadata
};
