const download = require('download');
const config = require('config');
const { readFile } = require('fs-extra');
const getLogger = require('../logger/logger');
const execute = require('../executor');

const logger = getLogger();

async function updateNodeVersion() {
    logger.info('Installed node version', process.versions.node);
    const aboutFile = await readFile(config.get('about-file'));
    const about = JSON.parse(aboutFile);
    logger.info('ABOUT file is', about);

    if (about.nodeVersion && about.nodeVersion !== process.versions.node) {
        logger.info('Installing node version', about.nodeVersion);
        try {
            await execute(`N_PREFIX=/ n ${about.nodeVersion}`);
            logger.info('Removing all versions except the current version...');
            await execute('N_PREFIX=/ n prune');
        } catch (err) {
            logger.error('Failed to update node', err);
        }
    }
}

async function performUpdate(bundleVersion, bundleUrl) {
    try {
        logger.info('Downloading and extracting...', { bundleVersion, bundleUrl });

        const options = process.env.http_proxy ? { extract: true, proxy: process.env.http_proxy } : { extract: true };

        await download(bundleUrl, '/opt/netapp/databroker', options);
    } catch (err) {
        logger.error(`Failed to download and extract ${bundleUrl}`, err);
    }

    try {
        await updateNodeVersion();
    } catch (err) {
        logger.error('Failed to update node');
    }

    logger.info('Installing production modules...');
    await execute('npm i --production');

    try {
        logger.info('Install latest PM2 version...');
        await execute('npm i pm2 -g');
    } catch (err) {
        logger.error('Failed to install latest PM2 version', err);
    }

    logger.info('Restarting...');
    await execute('pm2 restart all');
}

module.exports = performUpdate;
