const AWS = require('aws-sdk');
const proxy = require('proxy-agent');
const https = require('https');
const mimeType = require('mime-types');
const bytes = require('bytes');
const config = require('config');
const {
    isNil, compact, last
} = require('lodash');
const {
    TYPE_FILE, TYPE_DIRECTORY, TYPE_SYMBOLIC, MOUNT_SOURCE_SUFFIX, MOUNT_TARGET_SUFFIX
} = require('../consts');

const { getMetadata } = require('../metadata-store.js');

async function getSgws(host, port, accessKey, secretKey, isS3EndPoint, srcOrtrg, relationshipId) {
    let httpOptions;
    let accessKeyId = accessKey;
    let secretAccessKey = secretKey;
    const useProxy = config.get('protocols.sgws.use-proxy');
    const rejectInvalidCertificates = config.get('protocols.sgws.reject-invalid-certificates');
    if (useProxy) {
        httpOptions = {
            agent: proxy(process.env.http_proxy, { rejectUnauthorized: rejectInvalidCertificates })
        };
    } else {
        httpOptions = {
            agent: new https.Agent({ rejectUnauthorized: rejectInvalidCertificates })
        };
    }

    if (!accessKeyId || !secretAccessKey) {
        const credentials = await getMetadata(`${relationshipId}.${srcOrtrg}.sgws`);
        ({ accessKey: accessKeyId, secretKey: secretAccessKey } = credentials);
    }

    return new AWS.S3({
        endpoint: `${host}:${port}`,
        accessKeyId,
        secretAccessKey,
        httpOptions,
        s3BucketEndpoint: isS3EndPoint,
        s3ForcePathStyle: true,
        correctClockSkew: true
    });
}

async function listSgwsBuckets(host, port, accessKeyId, secretAccessKey) {
    const sgws = await getSgws(host, port, accessKeyId, secretAccessKey);
    const response = await sgws.listBuckets().promise();
    return response.Buckets.map(bucket => ({ name: bucket.Name }));
}

async function listSgwsBucketFolders(host, port, accessKeyId, secretAccessKey, bucket, prefix, continuationToken, dirs = []) {
    const sgws = await getSgws(host, port, accessKeyId, secretAccessKey, true);

    let fixedPrefix = isNil(prefix) || prefix === '' ? undefined : prefix.replace(/^\//, '');
    if (fixedPrefix) {
        fixedPrefix = fixedPrefix.endsWith('/') ? fixedPrefix : `${fixedPrefix}/`;
    }
    const objects = await sgws.listObjects({
        Bucket: bucket, Prefix: fixedPrefix, Delimiter: '/', Marker: continuationToken
    }).promise();

    // remove trailing slash if exists, and the starting prefix
    const updatedDirs = compact(dirs.concat(compact(objects.CommonPrefixes.map(commonPrefix => commonPrefix.Prefix.replace(/\/$/, '').replace(fixedPrefix, '')))));
    if (objects.IsTruncated) {
        return listSgwsBucketFolders(host, port, accessKeyId, secretAccessKey, bucket, prefix, objects.NextMarker, updatedDirs);
    }
    return updatedDirs;
}

async function headObject(sgws, bucket, key) {
    try {
        const metadata = await sgws.headObject({ Bucket: bucket, Key: key }).promise();
        return metadata.Metadata;
    } catch (e) {
        if (e.code === 'NotFound') { // in case of a folder without tags returns NoSuchKey
            return {};
        }
        throw e;
    }
}

async function getObjectsMetaData(sgws, bucket, key, size, objectType, mode, lastModified) {
    const metadata = await headObject(sgws, bucket, key);

    const modifiedTime = lastModified === undefined ? Date.now() : lastModified.getTime();
    if (!key.endsWith('/') || objectType !== TYPE_FILE) {
        return {
            path: (key === '/') ? '/' : last(compact(key.split('/'))),
            type: (metadata['netapp-cs-slink'] === 'True') ? TYPE_SYMBOLIC : objectType,
            size,
            mtimeSec: parseInt(metadata['netapp-cs-mtimesec']) || parseInt(modifiedTime / 1000),
            mode: parseInt(metadata['netapp-cs-mode']) || mode,
            uid: parseInt(metadata['netapp-cs-uid']) || 0,
            gid: parseInt(metadata['netapp-cs-gid']) || 0
        };
    }
    return null;
}

async function listSgwsObjects(host, port, accessKeyId, secretAccessKey, bucket, prefix, srcOrtrg, relationshipId, continuationToken, dirsAndFiles = []) {
    const sgws = await getSgws(host, port, accessKeyId, secretAccessKey, true, srcOrtrg, relationshipId);
    let fixedPrefix = isNil(prefix) || prefix === '' ? undefined : prefix;
    if (fixedPrefix) {
        fixedPrefix = fixedPrefix.endsWith('/') ? fixedPrefix : `${fixedPrefix}/`;
    }
    const objects = await sgws.listObjects({
        Bucket: bucket, Prefix: fixedPrefix, Delimiter: '/', Marker: continuationToken
    }).promise();

    const directoriesPromises = objects.CommonPrefixes.map(async resultPrefix => getObjectsMetaData(sgws, bucket, resultPrefix.Prefix, 0, TYPE_DIRECTORY, 16822));
    const filePromises = objects.Contents.map(async obj => getObjectsMetaData(sgws, bucket, obj.Key, obj.Size, TYPE_FILE, 33206, obj.LastModified));
    const result = await Promise.all(filePromises.concat(compact(directoriesPromises)));

    const updatedDirsAndFiles = compact(dirsAndFiles.concat(compact(result)));
    if (objects.IsTruncated) {
        return listSgwsObjects(host, port, accessKeyId, secretAccessKey, bucket, prefix, srcOrtrg, relationshipId, objects.NextMarker, updatedDirsAndFiles);
    }
    return updatedDirsAndFiles;
}

async function deleteSgwsObject(host, port, accessKeyId, secretAccessKey, bucket, object, srcOrtrg, relationshipId) {
    const sgws = await getSgws(host, port, accessKeyId, secretAccessKey, true, srcOrtrg, relationshipId);
    return sgws.deleteObject({ Bucket: bucket, Key: object }).promise();
}

async function uploadSgws(host, port, accessKeyId, secretAccessKey, bucket, bodyStream, key, metadata, relationshipId) { // add metadata
    const contentType = mimeType.lookup(key) || 'application/octet-stream';
    const sgws = await getSgws(host, port, accessKeyId, secretAccessKey, true, MOUNT_TARGET_SUFFIX, relationshipId);
    return new AWS.S3.ManagedUpload({
        partSize: bytes(config.get('protocols.s3.partSize')),
        queueSize: config.get('protocols.s3.queueSize'),
        params: {
            Bucket: bucket,
            Key: key,
            Body: bodyStream,
            Metadata: metadata,
            ContentType: contentType
        },
        leavePartsOnError: false,
        service: sgws
    });
}

async function getSgwsObject(host, port, accessKeyId, secretAccessKey, bucket, key, srcOrtrg, relationshipId) {
    const sgws = await getSgws(host, port, accessKeyId, secretAccessKey, true, srcOrtrg, relationshipId);
    const data = await sgws.getObject({ Bucket: bucket, Key: key }).promise();
    return data.Body;
}

async function getSgwsReadStream(host, port, accessKeyId, secretAccessKey, bucket, key, relationshipId) {
    const sgws = await getSgws(host, port, accessKeyId, secretAccessKey, true, MOUNT_SOURCE_SUFFIX, relationshipId);
    return sgws.getObject({ Bucket: bucket, Key: key }).createReadStream();
}

function createSgwsMetadata(entry, tags) {
    const metadata = {
        'netapp-cs-uid': `${entry.uid}`,
        'netapp-cs-gid': `${entry.gid}`,
        'netapp-cs-mode': `${entry.mode}`,
        'netapp-cs-mtimesec': `${entry.mtimeSec}`
    };

    if (entry.type === TYPE_SYMBOLIC) {
        metadata['netapp-cs-slink'] = 'True';
    }

    tags.forEach(tag => {
        const { Key, Value } = tag;
        metadata[Key] = Value;
    });

    return metadata;
}

module.exports = {
    listSgwsBuckets, listSgwsBucketFolders, listSgwsObjects, deleteSgwsObject, uploadSgws, getSgwsObject, getSgwsReadStream, createSgwsMetadata
};
