const Joi = require('joi');
const endPointSchema = require('../schemas/endPoint-schema');
const metadataSchema = require('../schemas/metadata-schema');
const { NFS_VERSIONS, CIFS_VERSIONS } = require('../../lib/consts');

const masterCopySchema = Joi.object().keys({
    relationshipId: Joi.string().required(),
    source: endPointSchema.required(),
    target: endPointSchema.required(),
    scannerQueue: Joi.string().required(),
    transferrerQueue: Joi.string().required(),
    grace: Joi.number().min(0).required(),
    deleteOnTarget: Joi.boolean().required()
});

const masterRelationshipsListSchema = Joi.object().keys({
    relationships: Joi.array().items(Joi.object({
        relationshipId: Joi.string(),
        scannerQueue: Joi.string().required(),
        transferrerQueue: Joi.string().required(),
        source: endPointSchema.required(),
        target: endPointSchema.required()
    })),
    metadata: metadataSchema
});

const masterListBucketFoldersSchema = Joi.object().keys({
    bucket: Joi.string().required(),
    prefix: Joi.string().allow('')
});

const masterListAzureBlobStorageContainersSchema = Joi.object().keys({
    storageAccountName: Joi.string().required(),
    storageAccountConnectionString: Joi.string()
});

const masterListAzureBlobStorageContainerFoldersSchema = Joi.object().keys({
    container: Joi.string().required(),
    prefix: Joi.string(),
    storageAccountName: Joi.string().required(),
    storageAccountConnectionString: Joi.string()
});

const masterListExportsSchema = Joi.object().keys({
    host: Joi.string().required()
});

const masterListExportFoldersSchema = Joi.object().keys({
    host: Joi.string().required(),
    path: Joi.string().allow(''),
    export: Joi.string().required(),
    version: Joi.number().valid(NFS_VERSIONS)
});

const doneRemoveAbortSchema = Joi.object().keys({
    relationshipId: Joi.string().required(),
    source: endPointSchema.required(),
    target: endPointSchema.required()
});

const updateSchema = Joi.object().keys({
    bundleUrl: Joi.string().uri({ scheme: ['https'] }).required(),
    bundleVersion: Joi.string().required(),
    relationships: Joi.array().items(Joi.object({
        relationshipId: Joi.string(),
        scannerQueue: Joi.string().required(),
        transferrerQueue: Joi.string().required(),
        source: endPointSchema.required(),
        target: endPointSchema.required()
    }))
});

const masterSharesSchema = Joi.object().keys({
    host: Joi.string().required(),
    username: Joi.string().required(),
    password: Joi.string().required(),
    workgroup: Joi.string(),
    version: Joi.number().valid(CIFS_VERSIONS)
});

const masterListShareFoldersSchema = Joi.object().keys({
    host: Joi.string().required(),
    share: Joi.string().required(),
    path: Joi.string(),
    username: Joi.string().required(),
    password: Joi.string().required(),
    workgroup: Joi.string(),
    version: Joi.number().valid(CIFS_VERSIONS)
});

const masterListSgwsBucketsSchema = Joi.object().keys({
    host: Joi.string().required(),
    port: Joi.string().required(),
    accessKey: Joi.string().required(),
    secretKey: Joi.string().required()
});

const masterListGcpBucketsFoldersSchema = Joi.object().keys({
    bucket: Joi.string().required(),
    prefix: Joi.string().allow(''),
    projectId: Joi.string().required(),
    clientEmail: Joi.string().required(),
    privateKey: Joi.string().required()
});

const masterListGcpBucketsSchema = Joi.object().keys({
    projectId: Joi.string().required(),
    clientEmail: Joi.string().required(),
    privateKey: Joi.string().required()
});


const masterListSgwsBucketsFoldersSchema = Joi.object().keys({
    bucket: Joi.string().required(),
    prefix: Joi.string().allow(''),
    host: Joi.string().required(),
    port: Joi.string().required(),
    accessKey: Joi.string(),
    secretKey: Joi.string().required()
});

const masterShutdownSchema = Joi.object().keys({});

const masterMetadataRequestSchema = Joi.object().keys({
    targetDataBrokerId: Joi.string().required()
});


const sqsMasterIncomingMessagesSchemes = {
    startup: masterRelationshipsListSchema,
    copy: masterCopySchema,
    buckets: Joi.object(),
    'bucket-folders': masterListBucketFoldersSchema,
    'sgws-buckets': masterListSgwsBucketsSchema,
    'gcp-buckets': masterListGcpBucketsSchema,
    'gcp-bucket-folders': masterListGcpBucketsFoldersSchema,
    'sgws-buckets-folders': masterListSgwsBucketsFoldersSchema,
    exports: masterListExportsSchema,
    'export-folders': masterListExportFoldersSchema,
    'azure-blob-storage-containers': masterListAzureBlobStorageContainersSchema,
    'azure-blob-storage-container-folders': masterListAzureBlobStorageContainerFoldersSchema,
    add: masterRelationshipsListSchema,
    done: doneRemoveAbortSchema,
    remove: doneRemoveAbortSchema,
    abort: doneRemoveAbortSchema,
    update: updateSchema,
    shares: masterSharesSchema,
    'share-folders': masterListShareFoldersSchema,
    shutdown: masterShutdownSchema,
    restart: masterRelationshipsListSchema,
    metadata: masterMetadataRequestSchema
};

module.exports = sqsMasterIncomingMessagesSchemes;
