const {cpus} = require('os');
const cluster = require('cluster');
const {join} = require('path');
const config = require('config');
const {
    mkdirs, remove
} = require('fs-extra');
const {readFileSync} = require('fs');
const prettyBytes = require('bytes');
const {
    update, head, keys, merge, trim, get, isEmpty
} = require('lodash');
const si = require('systeminformation');
const getLogger = require('../lib/logger/logger');
const convertMetadataJson = require('../lib/convert-metadata');
const {receiveMessages, sendMessage, deleteMessage} = require('../lib/aws/sqs');
const {listBuckets, listBucketFolders, headBucket} = require('../lib/aws/s3');
const getAwsAccountId = require('../lib/aws/sts');
const {listContainers, listBlobDirectories} = require('../lib/azure/blob');
const {registerMetadata, exportMetadata, importMetadata} = require('../lib/metadata-store.js');
const {isGcpBucketExists, getGcpCredentials} = require('../lib/gcp/storage');
const getAwsMetadata = require('../lib/aws/metadata');
const getGcpMetadata = require('../lib/gcp/metadata');
const getAzureMetadata = require('../lib/azure/metadata');
const getOsMetadata = require('../lib/os/metadata');
const sqsMasterIncomingMessagesSchemes = require('../messages/master/incoming-messages');
const sqsMasterOutgoingMessagesSchemes = require('../messages/master/outgoing-messages');
const {
    WORKER_TYPE_SCANNER, WORKER_TYPE_TRANSFERRER, WORKER_TYPE_SERVER, DATA_BROKER_RUNTIME_CONFIG, CERT_AND_KEY_FOLDER,
    DATA_BROKER_RUNTIME_ABOUT, STATUS_FAILED, DEPLOYMENT_TYPES, INITIALIZE_STEP, TYPE_DIRECTORY, STATUS_SUCCEEDED,
    LIST_STEP, TRANSFER_STEP, ACTION_DELETE, ACTION_DELETE_FAILED, ACTION_COPY_FAILED, CLEAN_RELATIONSHIP_RESOURCES_TIMEOUT,
    MOUNT_SOURCE_SUFFIX, MOUNT_TARGET_SUFFIX, MOUNT_EXPORTS_FOLDER,
    MOUNT_SHARES_FOLDER
} = require('../lib/consts');
const execute = require('../lib/executor');
const {failedSchema} = require('../messages/schemas/statuses-outgoing-messages');
const sendAccessListRequest = require('../lib/access-list');

const {statusesQueue} = DATA_BROKER_RUNTIME_CONFIG;
const {showmount, listShares} = require('../lib/fs/mount');
const {mountIfNeeded, umountAndDeleteRelationshipFolder, umountAndDelete} = require('../lib/fs/mount_auxiliary');
const {list, gracefulJoin, removeCoreFiles} = require('../lib/fs/fs');
const validateMessage = require('../messages/schemas/message-validator');
const performUpdate = require('../lib/phoenix/update');
const {startHealthLogInterval} = require('../lib/health');
const {listSgwsBuckets, listSgwsBucketFolders} = require('../lib/sgws/sgws');
const {listGcpBuckets, listGcpBucketFolders, getBucketLocationAndTags} = require('../lib/gcp/storage');
const {secureMkdirs} = require('../lib/snfs/snfs');

const logger = getLogger();
startHealthLogInterval();
const relationshipsLogger = getLogger('relationships');

const META_DATA = {};
const RELATIONSHIPS_STATUS = {};
const TOTALS = {};

function forkProcess(file, type) {
    logger.info('Forking file:', file);

    let execArgv = [];
    if (type === WORKER_TYPE_SCANNER) {
        const heapLimitInMb = prettyBytes(config.get('workers.scanner.heap-limit')) / 1024 / 1024;
        execArgv = [`--max_old_space_size=${heapLimitInMb}`];
    }

    cluster.setupMaster({
        exec: file,
        silent: true,
        execArgv
    });

    const childProcess = cluster.fork({UV_THREADPOOL_SIZE: config.get('uv-thread-pool-size')});
    childProcess.workerType = type;
    return childProcess;
}

function startTransferrers() {
    logger.info('Starting transferrers...');
    const processesCpuRatio = config.get('workers.transferrer.processes-cpu-ratio');
    const processesLimitNumber = config.get('workers.transferrer.processes-limit');
    const numberOfTransferrers = Math.min(cpus().length * processesCpuRatio, processesLimitNumber);
    logger.info('Process to CPU ratio', processesCpuRatio);
    logger.info('The max number of processes allowed', processesLimitNumber);
    logger.info('Number of transferrers', numberOfTransferrers);
    for (let i = 0; i < numberOfTransferrers; i++) {
        forkProcess(join(__dirname, 'transferrer.js'), WORKER_TYPE_TRANSFERRER);
    }
}

function startServer() {
    logger.info('Starting server...');
    const processesCpuRatio = config.get('workers.server.processes-cpu-ratio');
    const processesLimitNumber = config.get('workers.server.processes-limit');
    logger.info('Process to CPU ratio', processesCpuRatio);
    logger.info('The max number of processes allowed', processesLimitNumber);
    logger.info('Number of servers', processesLimitNumber);
    forkProcess(join(__dirname, 'server.js'), WORKER_TYPE_SERVER);
}

function startScanners() {
    logger.info('Starting scanners...');
    const processesCpuRatio = config.get('workers.scanner.processes-cpu-ratio');
    const processesLimitNumber = config.get('workers.scanner.processes-limit');
    const numberOfScanners = Math.min(cpus().length * processesCpuRatio, processesLimitNumber);
    logger.info('Process to CPU ratio', processesCpuRatio);
    logger.info('The max number of processes allowed', processesLimitNumber);
    logger.info('Number of scanners', numberOfScanners);
    for (let i = 0; i < numberOfScanners; i++) {
        forkProcess(join(__dirname, 'scanner.js'), WORKER_TYPE_SCANNER);
    }
}

function getEndPointString(endPoint) {
    if (endPoint.nfs || endPoint.cifs) {
        return 'local';
    }
    return head(keys(endPoint));
}

async function mkdirsIfNeeded(relationshipId, requestId, endPoint, mountPoint = undefined) {
    if (endPoint.local) {
        const {local: {path}} = endPoint;
        await mkdirs(path);
    } else if (endPoint.nfs) {
        const {nfs: {path}} = endPoint;
        await mkdirs(gracefulJoin(mountPoint, path));
    } else if (endPoint.cifs) {
        const {cifs: {path}} = endPoint;
        await mkdirs(gracefulJoin(mountPoint, path));
    } else if (endPoint.snfs) {
        const {
            snfs: {
                serverIP, port, certificate, serverName, path
            }
        } = endPoint;
        await secureMkdirs(relationshipId, serverIP, port, certificate, serverName, requestId, path);
    }
}

async function prepareBucketIfNeeded(endPoint, relationshipId, srcOrtrg) {
    // done to make sure the bucket exists, otherwise a fatally failed error will be sent from the master
    if (endPoint) {
        if (endPoint.s3) {
            const {s3: {bucket}} = endPoint;
            return headBucket(bucket);
        }
        if (endPoint.gcp) {
            const {gcp: {bucket}} = endPoint;
            const credentials = await getGcpCredentials(relationshipId, srcOrtrg);
            return isGcpBucketExists(credentials, bucket);
        }
    }
}

async function sendStatistics(relationshipId) {
    const statistics = RELATIONSHIPS_STATUS[relationshipId];
    if (statistics) {
        // sent statistics should be deleted synchronously since new statistics message are comming through IPC and we don't want to send the same counters more then once
        delete RELATIONSHIPS_STATUS[relationshipId];
        statistics.relationshipId = relationshipId;
        const {attributes} = statistics;
        delete statistics.attributes;
        if (validateMessage(statistics, 'Outgoing-SQS', sqsMasterOutgoingMessagesSchemes.statistics)) {
            await sendMessage(statusesQueue, statistics, attributes);
            logger.debug(`Sent statistics for relationship ${relationshipId} statistics: ${JSON.stringify(statistics)} and attributes ${JSON.stringify(attributes)}`);
        } else {
            logger.error(`Failed to send statistics ${JSON.stringify(statistics)} for relationship ${relationshipId} due to validation error`);
        }
    }
}

async function cleanRelationshipResources(relationshipId, sourceEndpoint, targetEndpoint) {
    logger.info(`Cleaning resources for rid: ${relationshipId}`);
    await sendStatistics(relationshipId);
    const totals = TOTALS[relationshipId];
    if (totals) {
        logger.info(`before cleaning relationship ${relationshipId}, current totals ${JSON.stringify(totals)}`);
    }
    // delete relationship's queues
    delete META_DATA[relationshipId];
    delete RELATIONSHIPS_STATUS[relationshipId];
    delete TOTALS[relationshipId];

    // unmount if mount point exists
    await umountAndDeleteRelationshipFolder(relationshipId, sourceEndpoint, targetEndpoint);
}

function getWorkersIdsByType(type) {
    return Object.keys(cluster.workers).filter(id => cluster.workers[id].workerType === type);
}

function sendWorkersMessage(message) {
    Object.keys(cluster.workers).forEach(id => cluster.workers[id].send(message));
}

function initializeScannerAndTrnasferrer(relationshipId, requestId, scannerQueue, transferrerQueue, serverName, certificate) {
    getWorkersIdsByType(WORKER_TYPE_SCANNER).forEach(id => cluster.workers[id].send({
        relationshipId, requestId, scannerQueue, transferrerQueue, serverName, certificate, type: 'copy'
    }));

    getWorkersIdsByType(WORKER_TYPE_TRANSFERRER).forEach(id => cluster.workers[id].send({
        relationshipId, requestId, transferrerQueue, serverName, certificate, type: 'copy'
    }));
}

async function handleStartUpAdd(message) {
    const {body: {relationships, metadata}, attributes: {requestId}} = message;
    await convertMetadataJson(relationships); // TODO-EG To delete in the beginning of June
    if (!isEmpty(metadata)) {
        await importMetadata(metadata);
    }
    const promises = relationships.map(async relationship => {
        const {
            relationshipId, scannerQueue, transferrerQueue, source, target
        } = relationship;
        try {
            relationshipsLogger.info(`[${relationshipId}] unmount source and target if needed`);
            await umountAndDeleteRelationshipFolder(relationshipId, source, target);
            relationshipsLogger.info(`[${relationshipId}] mount source and target if needed`);
            await mountIfNeeded(relationshipId, requestId, MOUNT_SOURCE_SUFFIX, source);
            await mountIfNeeded(relationshipId, requestId, MOUNT_TARGET_SUFFIX, target);
            META_DATA[relationshipId] = {scannerQueue, transferrerQueue, requestId};
            if (source.snfs) {
                initializeScannerAndTrnasferrer(relationshipId, requestId, scannerQueue, transferrerQueue,
                    source.snfs.serverName, source.snfs.certificate);
            } else if (target.snfs) {
                initializeScannerAndTrnasferrer(relationshipId, requestId, scannerQueue, transferrerQueue,
                    target.snfs.serverName, target.snfs.certificate);
            } else {
                initializeScannerAndTrnasferrer(relationshipId, requestId, scannerQueue, transferrerQueue);
            }
        } catch (err) {
            const errorMessage = merge({
                step: INITIALIZE_STEP,
                relationshipId,
                status: STATUS_FAILED,
                reason: err.message || err
            }, message.body);
            if (validateMessage(errorMessage, 'SQS', failedSchema)) {
                sendMessage(statusesQueue, errorMessage, message.attributes);
            }
        }
    });
    await Promise.all(promises);
}

async function handleCopyMessage(copyMessage) {
    await convertMetadataJson([copyMessage.body]); // TODO-EG To delete in the beginning of June
    const sourcePassword = get(copyMessage, `body.source.${Object.keys(copyMessage.body.source)[0]}.password`);
    const targetPassword = get(copyMessage, `body.target.${Object.keys(copyMessage.body.target)[0]}.password`);
    const {
        body: {
            relationshipId, scannerQueue, transferrerQueue, source, target
        },
        attributes: {requestId}
    } = copyMessage;

    if (source.snfs || target.snfs) {
        try {
            const listenerIp = config.get('snfs.listener-ip');
            logger.info(`********* Got data broker listener ip from config file *********`);
            logger.info(`********* data broker listener ip inserted by user: ${listenerIp}, connection to the listener data broker will be established using this ip *********`);
        } catch (err) {
            logger.info('********* Couldnt locate listener ip from config file:', err, '*********');
            logger.info('********* relationship will use private ip located in the relationship message *********');
        }
    }

    META_DATA[relationshipId] = {scannerQueue, transferrerQueue, requestId};

    await registerMetadata(source, target, sourcePassword, targetPassword, relationshipId);

    // prepare source
    await mountIfNeeded(relationshipId, requestId, MOUNT_SOURCE_SUFFIX, source, sourcePassword);
    await prepareBucketIfNeeded(source, relationshipId, MOUNT_SOURCE_SUFFIX);

    // prepare target
    const targetMountPoint = await mountIfNeeded(relationshipId, requestId, MOUNT_TARGET_SUFFIX, target, targetPassword);
    await mkdirsIfNeeded(relationshipId, requestId, target, targetMountPoint);
    await prepareBucketIfNeeded(target, relationshipId, MOUNT_TARGET_SUFFIX);

    if (source.snfs) {
        initializeScannerAndTrnasferrer(relationshipId, requestId, scannerQueue, transferrerQueue,
            source.snfs.serverName, source.snfs.certificate);
    } else if (target.snfs) {
        initializeScannerAndTrnasferrer(relationshipId, requestId, scannerQueue, transferrerQueue,
            target.snfs.serverName, target.snfs.certificate);
    } else {
        initializeScannerAndTrnasferrer(relationshipId, requestId, scannerQueue, transferrerQueue);
    }
}

async function handleExportFoldersMessage(message) {
    const {
        body: {
            host, path, export: exportName, version
        }, attributes: {requestId}
    } = message;
    const mountPoint = await mountIfNeeded(MOUNT_EXPORTS_FOLDER, requestId, '', {
        nfs: {
            host,
            export: exportName,
            version
        }
    });

    const content = await list(gracefulJoin(mountPoint, path));
    const folders = content.filter(entry => entry.type === TYPE_DIRECTORY).map(directory => `/${directory.path}`);
    const response = {
        host, path, export: exportName, folders, status: STATUS_SUCCEEDED
    };

    if (validateMessage(response, 'Outgoing-SQS', sqsMasterOutgoingMessagesSchemes['export-folders'])) {
        await sendMessage(statusesQueue, response, message.attributes);
    }
    await umountAndDelete(mountPoint);
}

async function handleDoneRemoveAbortMessage(message) {
    const {
        body: {
            relationshipId, source, target, scannerQueue, transferrerQueue
        }, attributes: {type}
    } = message;
    const {scannerQueue: relationshipScannerQueue, transferrerQueue: relationshipTransferrerQueue} = META_DATA[relationshipId] || {};
    if (relationshipScannerQueue === scannerQueue && relationshipTransferrerQueue === transferrerQueue) {
        logger.info(`Handling ${type} message for relationshipId: ${relationshipId} with scanner queue ${scannerQueue} and transferrer queue ${transferrerQueue}`);
        sendWorkersMessage({
            source: getEndPointString(source), target: getEndPointString(target), relationshipId, type
        });
        // TODO - clean relationship from realtime when all items Done aborting (For now "X" sec - need fixing).
        setTimeout(async () => {
            try {
                await cleanRelationshipResources(relationshipId, source, target);
            } catch (e) {
                logger.error(`failed cleaning relationship: ${relationshipId} resources with scanner queue ${scannerQueue} and transferrer queue ${transferrerQueue}`, e);
            }
        }, CLEAN_RELATIONSHIP_RESOURCES_TIMEOUT);

        if (type === 'done' && head(keys(source)) === 'cifs' && head(keys(target)) === 'cifs') {
            const {cifs: {aclServer}} = source;
            logger.info('trying to connect to access-list-server');
            if (aclServer) {
                const response = await sendAccessListRequest(source, target, aclServer, relationshipId);
                logger.info('done copying metadata', response.data);
            }
        }
    } else if (scannerQueue && transferrerQueue) { // on remove there are no queues on the message no need to log this case
        logger.warn(`received ${type} message with scannerQueue ${scannerQueue} instead of ${relationshipScannerQueue} 
        and transferrerQueue ${transferrerQueue} instead of ${relationshipTransferrerQueue} for rid ${relationshipId}`);
    }
}

async function handleUpdateMessage(message, queueUrl) {
    if (config.get('update-new-version')) {
        const {body: {bundleUrl, bundleVersion, relationships}} = message;

        logger.info('Cleaning all resources');
        relationships.forEach(relationship => {
            const {source, target, relationshipId} = relationship;
            sendWorkersMessage({
                source: getEndPointString(source), target: getEndPointString(target), relationshipId, type: 'update'
            });
        });
        // TODO - clean relationship from realtime when all items Done aborting (For now "X" sec - need fixing).
        setTimeout(async () => {
            try {
                const promises = relationships.map(async relationship => {
                    const {relationshipId, source, target} = relationship;
                    try {
                        await cleanRelationshipResources(relationshipId, source, target);
                    } catch (err) {
                        logger.error(`failed cleaning relationship: ${relationshipId} resources`, err);
                    }
                });
                await Promise.all(promises);
            } finally {
                // deleting message before restart
                logger.info('done cleaning resources..., deleting update message & updating & restart');
                await deleteMessage(queueUrl, message.receiptHandle);
                await performUpdate(bundleVersion, bundleUrl);
            }
        }, CLEAN_RELATIONSHIP_RESOURCES_TIMEOUT);
    } else {
        logger.info('Version update is disabled, ignoring and deleting update message');
        await deleteMessage(queueUrl, message.receiptHandle);
    }
}

async function handleExportsMessage(message) {
    const {body: {host}} = message;
    const exports = await showmount(host);
    const response = {host, exports, status: STATUS_SUCCEEDED};
    if (validateMessage(response, 'Outgoing-SQS', sqsMasterOutgoingMessagesSchemes.exports)) {
        await sendMessage(statusesQueue, response, message.attributes);
    }
}

async function handleBucketsMessage(message) {
    const buckets = await listBuckets();
    const response = {buckets, status: STATUS_SUCCEEDED};
    if (validateMessage(response, 'Outgoing-SQS', sqsMasterOutgoingMessagesSchemes.buckets)) {
        await sendMessage(statusesQueue, response, message.attributes);
    }
}


async function handleGcpBucketsMessage(message) {
    const {body: {projectId, clientEmail, privateKey}} = message;
    const buckets = await listGcpBuckets({projectId, clientEmail, privateKey});
    const response = {projectId, buckets, status: STATUS_SUCCEEDED};
    if (validateMessage(response, 'Outgoing-SQS', sqsMasterOutgoingMessagesSchemes['gcp-buckets'])) {
        await sendMessage(statusesQueue, response, message.attributes);
    }
}

async function handleGcpBucketFoldersMessage(message) {
    const {
        body: {
            bucket, prefix, projectId, clientEmail, privateKey
        }
    } = message;
    const {location} = await getBucketLocationAndTags({projectId, clientEmail, privateKey}, bucket);
    const folders = await listGcpBucketFolders({projectId, clientEmail, privateKey}, bucket, prefix);
    const response = {
        projectId, bucket: {name: bucket, region: location}, prefix, folders, status: STATUS_SUCCEEDED
    };
    if (validateMessage(response, 'Outgoing-SQS', sqsMasterOutgoingMessagesSchemes['gcp-bucket-folders'])) {
        await sendMessage(statusesQueue, response, message.attributes);
    }
}


async function handleSgwsBucketsMessage(message) {
    const {
        body: {
            host, port, accessKey, secretKey
        }
    } = message;
    const buckets = await listSgwsBuckets(host, port, accessKey, secretKey);
    const response = {
        host, port, buckets, status: STATUS_SUCCEEDED
    };
    if (validateMessage(response, 'Outgoing-SQS', sqsMasterOutgoingMessagesSchemes['sgws-buckets'])) {
        await sendMessage(statusesQueue, response, message.attributes);
    }
}

async function handleSgwsBucketFoldersMessage(message) {
    const {
        body: {
            bucket, prefix, host, port, accessKey, secretKey
        }
    } = message;
    const folders = await listSgwsBucketFolders(host, port, accessKey, secretKey, bucket, prefix);
    const response = {
        bucket, prefix, folders, host, port, status: STATUS_SUCCEEDED
    };
    if (validateMessage(response, 'Outgoing-SQS', sqsMasterOutgoingMessagesSchemes['sgws-buckets-folders'])) {
        await sendMessage(statusesQueue, response, message.attributes);
    }
}


async function handleBucketFoldersMessage(message) {
    const {body: {bucket, prefix}} = message;
    const region = await headBucket(bucket);
    const folders = await listBucketFolders(region, bucket, prefix);
    const response = {
        bucket: {name: bucket, region}, prefix, folders, status: STATUS_SUCCEEDED
    };
    if (validateMessage(response, 'Outgoing-SQS', sqsMasterOutgoingMessagesSchemes['bucket-folders'])) {
        await sendMessage(statusesQueue, response, message.attributes);
    }
}

async function handleAzureBlobStorageContainersMessage(message) {
    const {body: {storageAccountName, storageAccountConnectionString}} = message;
    const containers = await listContainers(storageAccountConnectionString);
    const response = {storageAccountName, containers, status: STATUS_SUCCEEDED};
    if (validateMessage(response, 'Outgoing-SQS', sqsMasterOutgoingMessagesSchemes['azure-blob-storage-containers'])) {
        await sendMessage(statusesQueue, response, message.attributes);
    }
}

async function handleAzureBlobStorageContainerFoldersMessage(message) {
    const {
        body: {
            container, prefix, storageAccountName, storageAccountConnectionString
        }
    } = message;

    const directories = await listBlobDirectories(container, prefix, storageAccountConnectionString);
    const folders = directories.map(directory => trim(directory.name.slice(prefix ? prefix.length : 0), '/'));

    const response = {
        storageAccountName, container, prefix, folders, status: STATUS_SUCCEEDED
    };

    if (validateMessage(response, 'Outgoing-SQS', sqsMasterOutgoingMessagesSchemes['azure-blob-storage-container-folders'])) {
        await sendMessage(statusesQueue, response, message.attributes);
    }
}

async function handleSharesMessage(message) {
    const password = get(message, 'body.password');
    const {
        body: {
            host, version, username, workgroup
        }
    } = message;

    const shares = await listShares(host, version, username, password, workgroup);
    const response = {shares, host, status: STATUS_SUCCEEDED};

    if (validateMessage(response, 'Outgoing-SQS', sqsMasterOutgoingMessagesSchemes.shares)) {
        await sendMessage(statusesQueue, response, message.attributes);
    }
}

async function handleShareFoldersMessage(message) {
    const password = get(message, 'body.password');
    const {
        body: {
            host, path, share, username, workgroup, version
        }, attributes: {requestId}
    } = message;

    const mountPoint = await mountIfNeeded(MOUNT_SHARES_FOLDER, requestId, '', {
            cifs: {
                host, share, username, workgroup, version
            }
        },
        password);

    const content = await list(gracefulJoin(mountPoint, path));
    const folders = content.filter(entry => entry.type === TYPE_DIRECTORY).map(directory => `/${directory.path}`);
    const response = {
        host, path, share, folders, status: STATUS_SUCCEEDED
    };

    if (validateMessage(response, 'Outgoing-SQS', sqsMasterOutgoingMessagesSchemes['share-folders'])) {
        await sendMessage(statusesQueue, response, message.attributes);
    }
    await umountAndDelete(mountPoint);
}

async function handleShutdownMessage() {
    logger.info('Shutting down...');
    await execute('pm2 delete all');
}

async function handleRestartMessage(message, queueUrl) {
    logger.info('Restarting  ...');
    const {body: {relationships}} = message;

    logger.info('Cleaning all resources');
    relationships.forEach(relationship => {
        const {source, target, relationshipId} = relationship;
        sendWorkersMessage({
            source: getEndPointString(source), target: getEndPointString(target), relationshipId, type: 'update'
        });
    });
    // TODO - clean relationship from realtime when all items Done aborting (For now "X" sec - need fixing).
    setTimeout(async () => {
        try {
            const promises = relationships.map(async relationship => {
                const {relationshipId, source, target} = relationship;
                try {
                    await cleanRelationshipResources(relationshipId, source, target);
                } catch (e) {
                    logger.error(`failed cleaning relationship: ${relationshipId} resources`, e);
                }
            });
            await Promise.all(promises);
        } finally {
            // deleting message before restart
            logger.info('done cleaning resources..., deleting update message & updating & restart');
            await deleteMessage(queueUrl, message.receiptHandle);
            await execute('pm2 restart all');
        }
    }, CLEAN_RELATIONSHIP_RESOURCES_TIMEOUT);
}

async function handleMetadataMessage(message) {
    logger.info('handle metadataRequestMessage', message);
    const {body: {targetDataBrokerId}} = message;
    // create metadata response
    const metadata = await exportMetadata();
    const response = {targetDataBrokerId, metadata, status: STATUS_SUCCEEDED};
    if (validateMessage(response, 'Outgoing-SQS', sqsMasterOutgoingMessagesSchemes.metadata)) {
        await sendMessage(statusesQueue, response, message.attributes);
    }
}

async function handleMessage(message, queueUrl) {
    try {
        if (validateMessage(message.body, 'SQS', sqsMasterIncomingMessagesSchemes[message.attributes.type])) {
            switch (message.attributes.type) {
                case 'copy':
                    await handleCopyMessage(message);
                    break;
                case 'done':
                case 'remove':
                case 'abort':
                    await handleDoneRemoveAbortMessage(message);
                    break;
                case 'exports':
                    await handleExportsMessage(message);
                    break;
                case 'export-folders':
                    await handleExportFoldersMessage(message);
                    break;
                case 'buckets':
                    await handleBucketsMessage(message);
                    break;
                case 'gcp-buckets':
                    await handleGcpBucketsMessage(message);
                    break;
                case 'gcp-bucket-folders':
                    await handleGcpBucketFoldersMessage(message);
                    break;
                case 'sgws-buckets':
                    await handleSgwsBucketsMessage(message);
                    break;
                case 'sgws-buckets-folders':
                    await handleSgwsBucketFoldersMessage(message);
                    break;
                case 'bucket-folders':
                    await handleBucketFoldersMessage(message);
                    break;
                case 'startup':
                case 'add':
                    await handleStartUpAdd(message);
                    break;
                case 'update':
                    await handleUpdateMessage(message, queueUrl);
                    break;
                case 'azure-blob-storage-containers':
                    await handleAzureBlobStorageContainersMessage(message);
                    break;
                case 'azure-blob-storage-container-folders':
                    await handleAzureBlobStorageContainerFoldersMessage(message);
                    break;
                case 'shares':
                    await handleSharesMessage(message);
                    break;
                case 'share-folders':
                    await handleShareFoldersMessage(message);
                    break;
                case 'shutdown':
                    await handleShutdownMessage();
                    break;
                case 'restart':
                    await handleRestartMessage(message, queueUrl);
                    break;
                case 'metadata':
                    await handleMetadataMessage(message);
                    break;
                default:
                    logger.error(`Got unknown command type ${message.attributes.type}:`, message);
            }
        }
    } catch (e) {
        const {body: {relationshipId}} = message;
        const errorMessage = merge({
            step: INITIALIZE_STEP,
            relationshipId,
            status: STATUS_FAILED,
            reason: e.message || e
        }, message.body);
        if (validateMessage(errorMessage, 'SQS', failedSchema)) {
            if (relationshipId) {
                relationshipsLogger.error(`[${relationshipId}] fatally failed due to error:`, e);
            } else {
                logger.error(`[${message.attributes.type}]  fatally failed due to error:`, e);
            }
            sendMessage(statusesQueue, errorMessage, message.attributes);
        }
    } finally {
        await deleteMessage(queueUrl, message.receiptHandle);
    }
}

async function poll(queueUrl) {
    logger.debug('Polling queue', queueUrl);
    try {
        const messages = await receiveMessages(queueUrl);
        messages.forEach(msg => handleMessage(msg, queueUrl));
        setImmediate(() => poll(queueUrl));
    } catch (err) {
        if (err.code === 'AWS.SimpleQueueService.NonExistentQueue') {
            logger.warn(`'${queueUrl}' queue does not exist -> stop polling`);
        } else {
            logger.warn(`Delaying polling for SQS queue '${queueUrl}' due to error`, err);
            setTimeout(() => poll(queueUrl), 5000);
        }
    }
}

async function sendPlacement() {
    const {type} = DATA_BROKER_RUNTIME_CONFIG;
    const awsAccountId = await getAwsAccountId();
    const osMetadata = await getOsMetadata();
    const {version} = DATA_BROKER_RUNTIME_ABOUT;
    const proxy = process.env.http_proxy;

    let message = Object.assign(osMetadata, {version, awsAccountId, proxy});

    if (DATA_BROKER_RUNTIME_CONFIG.port) {
        const {'data-broker-id': databrokerId} = DATA_BROKER_RUNTIME_CONFIG;
        const certPath = `${CERT_AND_KEY_FOLDER}/${databrokerId}.crt`;
        message.certificate = readFileSync(certPath, 'utf8');
    }

    try {
        let messageType;
        switch (type) {
            case DEPLOYMENT_TYPES.AWS: {
                const awsMetadata = await getAwsMetadata();
                message = Object.assign(message, awsMetadata);
                messageType = sqsMasterOutgoingMessagesSchemes['placement-aws'];
                break;
            }
            case DEPLOYMENT_TYPES.ONPREM: {
                messageType = sqsMasterOutgoingMessagesSchemes['placement-on-prem'];
                break;
            }
            case DEPLOYMENT_TYPES.AZURE: {
                const azureMetadata = await getAzureMetadata();
                message = Object.assign(message, azureMetadata);
                messageType = sqsMasterOutgoingMessagesSchemes['placement-azure'];
                break;
            }
            case DEPLOYMENT_TYPES.GCP: {
                let gcpMetadata;
                try {
                    gcpMetadata = await getGcpMetadata();
                } catch (err) {
                    logger.warn('Couldn\'t get GCP metadata. ', err);
                }
                if (gcpMetadata) {
                    message = Object.assign(message, gcpMetadata);
                    messageType = sqsMasterOutgoingMessagesSchemes['placement-gcp'];
                } else {
                    messageType = sqsMasterOutgoingMessagesSchemes['placement-on-prem'];
                }
                break;
            }
            default:
                logger.error(`Failed to send placement - unknown deployement type: ${type}`);
        }

        if (validateMessage(message, 'Outgoing-SQS', messageType)) {
            await sendMessage(statusesQueue, message, {type: 'placement'});
        }
    } catch (err) {
        logger.error('Failed to send placement', err);
    }
}

async function getTransferRate() {
    try {
        const networkStats = await si.networkStats('*');

        return networkStats
            .filter(networkStat => networkStat.iface !== 'lo')
            .filter(networkStat => networkStat.tx_sec !== -1)
            .map(networkStat => networkStat.tx_sec)
            .reduce((accumulator, value) => accumulator + value, 0);
    } catch (err) {
        logger.warn('Unable to get transfer rate', err);
        return 0;
    }
}

async function sendHealth() {
    const message = {
        transferrers: getWorkersIdsByType(WORKER_TYPE_TRANSFERRER).length,
        scanners: getWorkersIdsByType(WORKER_TYPE_SCANNER).length,
        memory: process.memoryUsage(),
        cpuUsage: process.cpuUsage(),
        transferRate: await getTransferRate()
    };

    if (validateMessage(message, 'Outgoing-SQS', sqsMasterOutgoingMessagesSchemes.health)) {
        sendMessage(statusesQueue, message, {type: 'health'});
    }
}

function findTargetRequiredAction(status, step, targetRequiredAction) {
    if (status === STATUS_FAILED && step === LIST_STEP) {
        return STATUS_FAILED;
    }
    if (status === STATUS_FAILED && step === TRANSFER_STEP) {
        if (targetRequiredAction === ACTION_DELETE) {
            return ACTION_DELETE_FAILED;
        }
        return ACTION_COPY_FAILED;
    }

    return targetRequiredAction;
}

async function boot() {
    logger.info('Booting, number of CPUs', cpus().length);

    cluster
        .on('exit', async (worker, code, signal) => {
            const {workerType} = worker;
            logger.warn(`Worker #${worker.id}, PID ${worker.process.pid} died ${(signal || code)}.`);
            if (signal === 'SIGABRT') {
                await remove(`core.${worker.process.pid}`);
            }
            if (workerType === WORKER_TYPE_SCANNER) {
                logger.info(`Restarting worker of type ${workerType}.`);
                const scannerProcess = forkProcess(join(__dirname, 'scanner.js'), workerType);
                Object.keys(META_DATA).forEach(async relationshipId => {
                    scannerProcess.send({
                        relationshipId,
                        scannerQueue: META_DATA[relationshipId].scannerQueue,
                        transferrerQueue: META_DATA[relationshipId].transferrerQueue,
                        requestId: META_DATA[relationshipId].requestId,
                        type: 'copy'
                    });
                });
            } else if (workerType === WORKER_TYPE_TRANSFERRER) {
                logger.info(`Restarting worker of type ${workerType}.`);
                const transferrerProcess = forkProcess(join(__dirname, 'transferrer.js'), workerType);
                Object.keys(META_DATA).forEach(async relationshipId => {
                    transferrerProcess.send({
                        relationshipId,
                        transferrerQueue: META_DATA[relationshipId].transferrerQueue,
                        requestId: META_DATA[relationshipId].requestId,
                        type: 'copy'
                    });
                });
            } else if (workerType === WORKER_TYPE_SERVER && code !== 8) {
                logger.info(`Restarting worker of type ${workerType}.`);
                forkProcess(join(__dirname, 'server.js'), workerType);
            }
        })
        .on('message', (worker, message) => {
            if (message.topic !== 'log4js:message') {
                const {
                    status, relationshipId, objectType, step, bytes, items
                } = message;

                const targetRequiredAction = findTargetRequiredAction(status, step, message.targetRequiredAction);

                update(RELATIONSHIPS_STATUS, `${relationshipId}.${step}.${targetRequiredAction}.${objectType}`, obj => (obj ? {
                    bytes: obj.bytes + bytes,
                    items: obj.items + items
                } : {bytes, items}));
                RELATIONSHIPS_STATUS[relationshipId].attributes = message.attributes;

                update(TOTALS, `${relationshipId}.${step}.${targetRequiredAction}.${objectType}`, obj => (obj ? {
                    bytes: obj.bytes + bytes,
                    items: obj.items + items
                } : {bytes, items}));
            }
        });
    if (prettyBytes(config.get('protocols.s3.max-object-size')) < prettyBytes('1TB')) {
        logger.warn('consider increasing protocols.s3.max-object-size in order to avoid batching problems');
    }
    startTransferrers();
    startScanners();
    if (DATA_BROKER_RUNTIME_CONFIG.port) {
        startServer();
    }
    removeCoreFiles();
    setInterval(sendHealth, config.get('statuses.health.send-interval'));
    setInterval(() => Object.keys(META_DATA).forEach(relationshipId => sendStatistics(relationshipId)), config.get('statuses.statistics.interval'));

    await poll(DATA_BROKER_RUNTIME_CONFIG.commandsQueue);
    await sendPlacement();
}

module.exports = boot;
