const cluster = require('cluster');
const https = require('https');
const http = require('http');
const config = require('config');
const bytes = require('bytes');
const randomize = require('randomatic');
const { queue } = require('async');
const {
    _, shuffle, head, partition, keyBy, mapValues, keys, compact, isEqual, isEqualWith, get, includes, sumBy, chunk, groupBy
} = require('lodash');
const ms = require('ms');
const Promise = require('bluebird');
const retry = require('async-retry');
const getLogger = require('../lib/logger/logger');
const { list, gracefulJoin } = require('../lib/fs/fs');
const {
    receiveMessages, sendMessage, sendMessageBatch, deleteMessage, changeMessageVisibilityBatch
} = require('../lib/aws/sqs');
const {
    TYPE_FILE, TYPE_DIRECTORY, TYPE_OTHER, TYPE_SYMBOLIC, SCAN_MODE_SOURCE_ONLY, SCAN_MODE_TARGET_ONLY, SCAN_MODE_SOURCE_AND_TARGET,
    ACTION_DELETE, ACTION_COPY, ACTION_IGNORE, ACTION_LIST, STATUS_SUCCEEDED, STATUS_FAILED,
    STATUS_IGNORED, SCAN_STEP, LIST_STEP, IPC_TYPE_DONE, IPC_TYPE_UPDATE, IPC_TYPE_ABORT, IPC_TYPE_REMOVE, IPC_TYPE_COPY,
    NETAPP_ENTRY_SUFFIX, NETAPP_ENTRY_RANDOM_SIZE, NETAPP_ENTRY_RANDOM_PATTERN,
    MOUNT_SOURCE_SUFFIX, MOUNT_TARGET_SUFFIX, SQS_DEFAULT_WAIT_TIMEOUT, IGNORED_DIRECTORIES, IGNORED_FILES, SQS_MESSAGE_UNCHANGED_VISIBILITY_TIMEOUT_LIMIT, SQS_CHANGE_MESSAGE_VISIBILITY_TIMEOUT_INTERVAL
} = require('../lib/consts');
const { listObjects, calculateObjectStoragePath } = require('../lib/aws/s3');
const { listBlobs } = require('../lib/azure/blob');
const { listGcpObjects, getGcpCredentials } = require('../lib/gcp/storage');
const scannerIncomingSQSSchema = require('../messages/scanner/sqs-incoming-messages');
const scannerIncomingIPCMessagesSchemes = require('../messages/scanner/ipc-incoming-messages');
const { copyStatisticsSchema } = require('../messages/schemas/statuses-outgoing-messages');
const validateMessage = require('../messages/schemas/message-validator');
const { getNewEndPointIfNeeded } = require('../lib/fs/mount_auxiliary');
const { startHealthLogInterval } = require('../lib/health');
const { listSgwsObjects } = require('../lib/sgws/sgws');


const { listSecureNFS } = require('../lib/snfs/snfs');

const logger = getLogger();
const relationshipsLogger = getLogger('relationships');
const failedRelationshipLogger = getLogger('failed-relationships');

http.globalAgent.maxSockets = 500;
https.globalAgent.maxSockets = 500;

startHealthLogInterval();

process.on('uncaughtException', err => logger.error('Got uncaught exception', err));
process.on('unhandledRejection', event => logger.error('Got unhandled rejection', event));

const META_DATA = {}; // relationship id to map metadata
let INTERNAL_DISPATCHER_QUEUE;

function compareFiles(sourceFiles, targetFiles, targetProtocol) {
    const sourceFilesObj = keyBy(sourceFiles, 'path');
    const targetFilesObj = keyBy(targetFiles, 'path');
    const sizeLimitProtocols = ['s3', 'gcp', 'azure'];
    const result = {
        TASK_TYPE_COPY: [],
        TASK_TYPE_DELETE: [],
        TASK_TYPE_IGNORE: []
    };

    function ignoreEntry(e, reason) {
        e.reason = reason;
        result.TASK_TYPE_IGNORE.push(e);
    }

    keys(sourceFilesObj).forEach(path => {
        const src = sourceFilesObj[path];
        const trg = targetFilesObj[path];
        if (src.type === TYPE_SYMBOLIC && config.get('workers.scanner.ignore.symbolic-links')) {
            ignoreEntry(src, 'ignoring symbolic links');
        } else if (sizeLimitProtocols.includes(targetProtocol) && src.size > bytes(config.get(`protocols.${targetProtocol}.max-object-size`))) {
            ignoreEntry(src, 'ignoring exceeding size limit files');
        } else if (includes(IGNORED_FILES, path)) {
            ignoreEntry(src, 'ignored files');
        } else if (src.type === TYPE_OTHER) {
            ignoreEntry(src, 'special file');
        } else if (!trg || !isEqual(src, trg)) { // In source but Not in target or not equal ---> copy
            result.TASK_TYPE_COPY.push(src);
        }
    });

    keys(targetFilesObj).forEach(path => {
        const src = sourceFilesObj[path];
        const trg = targetFilesObj[path];
        if (trg.type === TYPE_SYMBOLIC && config.get('workers.scanner.ignore.symbolic-links')) {
            ignoreEntry(trg, 'ignoring symbolic links');
        } else if (includes(IGNORED_FILES, path)) {
            ignoreEntry(trg, 'ignored files');
        } else if (trg.type === TYPE_OTHER) {
            ignoreEntry(trg, 'special file');
        } else if (!src || src.type === TYPE_OTHER) { // In target but Not in source (or special file in source) ---> delete
            result.TASK_TYPE_DELETE.push(trg);
        }
    });
    return result;
}

function compareDirs(sourceDirs, targetDirs) {
    const sourceDirsObj = keyBy(sourceDirs, 'path');
    const targetDirsObj = keyBy(targetDirs, 'path');
    const result = {
        TASK_TYPE_EQUAL: [],
        TASK_TYPE_SOURCE_ONLY: [],
        TASK_TYPE_TARGET_ONLY: [],
        TASK_TYPE_IGNORE: []
    };

    function ignoreEntry(e, reason) {
        e.reason = reason;
        result.TASK_TYPE_IGNORE.push(e);
    }

    keys(sourceDirsObj).forEach(path => {
        const src = sourceDirsObj[path];
        const trg = targetDirsObj[path];
        if (includes(IGNORED_DIRECTORIES, path)) {
            ignoreEntry(src, 'ignored directory');
        } else if (!trg || !isEqualWith(src, trg, (val1, val2, key) => (key === 'mtimeSec' ? true : undefined))) { // In source but Not in target
            result.TASK_TYPE_SOURCE_ONLY.push(src);
        } else {
            result.TASK_TYPE_EQUAL.push(src);
        }
    });

    keys(targetDirsObj).forEach(path => {
        const src = sourceDirsObj[path];
        const trg = targetDirsObj[path];
        if (includes(IGNORED_DIRECTORIES, path)) {
            ignoreEntry(trg, 'ignored directory');
        } else if (!src) { // In target but Not in source
            result.TASK_TYPE_TARGET_ONLY.push(trg);
        }
    });
    return result;
}

function transformPath(endPoint, directory) {
    const protocol = head(keys(endPoint));
    switch (protocol) {
        case 'cifs':
        case 'nfs':
        case 'local':
        case 'snfs':
            return mapValues(endPoint, value => Object.assign({}, value, { path: gracefulJoin(value.path, directory.path) }));
        case 's3':
        case 'sgws':
        case 'azure':
        case 'gcp':
            return mapValues(endPoint, value => Object.assign({}, value, { path: `${calculateObjectStoragePath(value.path, directory.path)}` }));
        default:
            throw new Error(`transformPath, got wrong protocol for endPoint ${endPoint}`);
    }
}

async function sendScannerMessages(message, type, directories) {
    const {
        body: {
            source, target, relationshipId, grace, deleteOnTarget
        }
    } = message;
    const queues = META_DATA[relationshipId] ? META_DATA[relationshipId].queues : undefined;
    if (queues) {
        return Promise.map(directories, async dir => {
            const scannerMessage = {
                relationshipId,
                source: transformPath(source, dir),
                target: transformPath(target, dir),
                grace,
                deleteOnTarget,
                type
            };
            logger.debug(`Send message to scanner got type: ${type}, message: ${scannerMessage}`);
            await sendMessage(queues.scannerQueue, scannerMessage, message.attributes);
        }, { concurrency: 50 });
    }
    return Promise.resolve();
}

function sendStatisticsMessage(statisticsMessage) {
    const isValid = validateMessage(statisticsMessage, 'SQS', copyStatisticsSchema);
    if (isValid) {
        process.send(statisticsMessage);
    }
}

function getEndPointProtocol(endPoint) {
    return _(endPoint).keys().head();
}

async function sendTransferrerMessages(message, type, entries, source, target) {
    const protocolsForMultipart = ['local', 'nfs', 'cifs'];
    const { body: { relationshipId } } = message;
    const queues = META_DATA[relationshipId] ? META_DATA[relationshipId].queues : undefined;

    if (queues) {
        const transferrerMessages = entries.map(entry => {
            const statisticsMessage = {
                step: SCAN_STEP,
                relationshipId,
                targetRequiredAction: type,
                bytes: entry.size,
                items: 1,
                objectType: entry.type,
                path: entry.path,
                status: STATUS_SUCCEEDED,
                attributes: message.attributes
            };
            sendStatisticsMessage(statisticsMessage);
            if (protocolsForMultipart.includes(getEndPointProtocol(source)) && protocolsForMultipart.includes(getEndPointProtocol(target))) {
                const partSize = bytes(config.get('protocols.nfs.max-entry-part-size'));
                if (entry.size > partSize) {
                    logger.info(`entry that can be copied as multipart source ${JSON.stringify(source)} target ${JSON.stringify(target)}, ${JSON.stringify(entry)}`);
                    entry.partsDetails = {
                        partNumber: 1,
                        start: 0,
                        end: Math.min(partSize, entry.size),
                        partSize
                    };
                }
            }
            entry.tempEntryName = (protocolsForMultipart.includes(getEndPointProtocol(target))) ? `.${entry.path}_${randomize(NETAPP_ENTRY_RANDOM_PATTERN, NETAPP_ENTRY_RANDOM_SIZE)}${NETAPP_ENTRY_SUFFIX}` : undefined;
            return {
                type,
                source: type === ACTION_COPY ? source : undefined,
                target,
                entry,
                relationshipId
            };
        });
        const [smalls, bigs] = partition(transferrerMessages, transferrerMessage => transferrerMessage.entry.size <= bytes(config.get('workers.scanner.batching.max-entry-size')));
        await Promise.all([
            sendMessageBatch(queues.transferrerQueue, bigs, message.attributes),
            sendMessageBatch(queues.transferrerQueue, chunk(smalls, config.get('workers.scanner.batching.max-size')), message.attributes)
        ]);
    }
}

async function ignoreEntries(message, entries) {
    const { body: { relationshipId } } = message;
    entries.forEach(entry => {
        relationshipsLogger.info(`[${relationshipId}] IGNORING`, entry);
        const statisticsMessage = {
            step: SCAN_STEP,
            relationshipId,
            targetRequiredAction: ACTION_IGNORE,
            bytes: entry.size,
            items: 1,
            objectType: entry.type,
            path: entry.path,
            status: STATUS_IGNORED,
            attributes: message.attributes
        };
        sendStatisticsMessage(statisticsMessage);
    });
}

function calculateListStatistics(sourceFiles, relationshipId, attributes) {
    const createStatsMessage = (itemsBytes, items, objectType) => ({
        step: LIST_STEP,
        relationshipId,
        targetRequiredAction: ACTION_LIST,
        bytes: itemsBytes,
        items,
        objectType,
        status: STATUS_SUCCEEDED,
        attributes
    });
    // directories
    const dirStatisticsMessage = createStatsMessage(0, 1, TYPE_DIRECTORY);
    sendStatisticsMessage(dirStatisticsMessage);
    // partition between files and symbolic links
    // files
    const [symbolic, files] = partition(sourceFiles, entry => entry.type === TYPE_SYMBOLIC);
    const sizeOfFiles = sumBy(files, file => file.size);

    const filesStatisticsMessage = createStatsMessage(sizeOfFiles, files.length, TYPE_FILE);
    sendStatisticsMessage(filesStatisticsMessage);

    // symbolic link
    const symbolicStatisticsMessage = createStatsMessage(0, symbolic.length, TYPE_SYMBOLIC);
    sendStatisticsMessage(symbolicStatisticsMessage);
}

async function getDirectoriesAndFiles(endPoint, relationshipId, srcOrtrg) {
    const endPointProtocol = head(keys(endPoint));
    let endPointList = [];
    await retry(async () => {
        switch (endPointProtocol) {
            case 'local':
                endPointList = await list(endPoint.local.path);
                break;
            case 'snfs':
                endPointList = await listSecureNFS(relationshipId, endPoint.snfs.serverIP, endPoint.snfs.port,
                    META_DATA[relationshipId].snfs.certificate, META_DATA[relationshipId].snfs.serverName, META_DATA[relationshipId].requestId, endPoint.snfs.path);
                break;
            case 's3':
                endPointList = await listObjects(endPoint.s3.region, endPoint.s3.bucket, endPoint.s3.path);
                break;
            case 'sgws':
                endPointList = await listSgwsObjects(endPoint.sgws.host, endPoint.sgws.port, endPoint.sgws.accessKey,
                    endPoint.sgws.secretKey, endPoint.sgws.bucket, endPoint.sgws.path, srcOrtrg, relationshipId);
                break;
            case 'azure':
                endPointList = await listBlobs(endPoint.azure.container, endPoint.azure.path, relationshipId, srcOrtrg);
                break;
            case 'gcp': {
                const credentials = await getGcpCredentials(relationshipId, srcOrtrg);
                endPointList = await listGcpObjects(credentials, endPoint.gcp.bucket, endPoint.gcp.path);
                break;
            }
            default:
                throw new Error(`Got unsupported protocol ${endPoint}`);
        }
    }, {
        retries: config.get('retry.retries'),
        onRetry: err => logger.info('Retrying listing of', endPoint, err)
    });

    const [directories, files] = partition(endPointList, entry => entry.type === TYPE_DIRECTORY);
    return { directories, files };
}

function filterGracePeriod(message, entries) {
    const currentTimeMilliSec = new Date().getTime();
    const { body: { source, relationshipId, grace } } = message;

    return entries.filter(entry => {
        const notGrace = ((entry.mtimeSec + grace) * 1000) < currentTimeMilliSec;
        if (!notGrace) {
            relationshipsLogger.info(`[${relationshipId}] IGNORING - GRACE: ${grace} sec, time: ${currentTimeMilliSec}, source:`, source, 'entry:', entry);
        }
        return notGrace;
    });
}

async function handleScannerMessage(message) {
    const {
        body: {
            source, target, relationshipId, type, deleteOnTarget
        },
        receiptHandle, attributes
    } = message;


    try {
        const newSource = getNewEndPointIfNeeded(relationshipId, META_DATA[relationshipId].requestId, MOUNT_SOURCE_SUFFIX, source);
        const newTarget = getNewEndPointIfNeeded(relationshipId, META_DATA[relationshipId].requestId, MOUNT_TARGET_SUFFIX, target);


        const promises = [];
        switch (type) {
            case SCAN_MODE_SOURCE_AND_TARGET: {
                // list both sides
                const [sourceDirectoriesAndFiles, targetDirectoriesAndFiles] = await Promise.all([getDirectoriesAndFiles(newSource, relationshipId, MOUNT_SOURCE_SUFFIX), getDirectoriesAndFiles(newTarget, relationshipId, MOUNT_TARGET_SUFFIX)]);
                const { directories: sourceDirectories, files: sourceFiles } = sourceDirectoriesAndFiles;
                const { directories: targetDirectories, files: targetFiles } = targetDirectoriesAndFiles;
                logger.debug(`directoriesOfSource ${JSON.stringify(sourceDirectories, null, 4)} files of source ${JSON.stringify(sourceFiles, null, 4)}`);
                logger.debug(`directoriesOfTarget ${JSON.stringify(targetDirectories, null, 4)} files of target ${JSON.stringify(targetFiles, null, 4)}`);

                calculateListStatistics(sourceFiles, relationshipId, attributes);
                // compare files and send messages to transferrer queue
                const filesDifferences = compareFiles(sourceFiles, targetFiles, head(keys(target)));
                const dirDifferences = compareDirs(sourceDirectories, targetDirectories);

                logger.debug('2 side - different files found', filesDifferences);
                const filesFiltered = filterGracePeriod(message, filesDifferences.TASK_TYPE_COPY);
                promises.push(sendTransferrerMessages(message, ACTION_COPY, filesFiltered, source, target));
                promises.push(ignoreEntries(message, filesDifferences.TASK_TYPE_IGNORE));
                if (deleteOnTarget) {
                    promises.push(sendTransferrerMessages(message, ACTION_DELETE, filesDifferences.TASK_TYPE_DELETE, source, target));
                }

                // Directories that were ignored
                promises.push(ignoreEntries(message, dirDifferences.TASK_TYPE_IGNORE));
                // compare directories - equal directories - dive in
                logger.debug('2 side - equal directories found', dirDifferences.TASK_TYPE_EQUAL, ' sending message', JSON.stringify(message, null, 4), ' to scanner with type ', SCAN_MODE_SOURCE_AND_TARGET);
                promises.push(sendScannerMessages(message, SCAN_MODE_SOURCE_AND_TARGET, dirDifferences.TASK_TYPE_EQUAL));

                // compare directories - directories existing in the source, but not in the target - copy from source to target
                logger.debug('2 side - directories to copy', dirDifferences.TASK_TYPE_SOURCE_ONLY, ' sending message', JSON.stringify(message, null, 4), ' to scanner with type ', SCAN_MODE_SOURCE_ONLY);
                promises.push(sendScannerMessages(message, SCAN_MODE_SOURCE_ONLY, dirDifferences.TASK_TYPE_SOURCE_ONLY));
                promises.push(sendTransferrerMessages(message, ACTION_COPY, dirDifferences.TASK_TYPE_SOURCE_ONLY, source, target));

                // compare directories - directories existing in the target, but not in the source - delete from target
                if (deleteOnTarget) {
                    logger.debug('2 side - directories to delete', dirDifferences.TASK_TYPE_TARGET_ONLY, ' sending message', JSON.stringify(message, null, 4), ' to scanner with type ', SCAN_MODE_TARGET_ONLY);
                    promises.push(sendScannerMessages(message, SCAN_MODE_TARGET_ONLY, dirDifferences.TASK_TYPE_TARGET_ONLY));
                    promises.push(sendTransferrerMessages(message, ACTION_DELETE, dirDifferences.TASK_TYPE_TARGET_ONLY, source, target));
                }
                break;
            }
            case SCAN_MODE_SOURCE_ONLY: {
                const { directories, files } = await getDirectoriesAndFiles(newSource, relationshipId, MOUNT_SOURCE_SUFFIX);
                logger.debug(`directoriesOfEndPoint ${JSON.stringify(newSource, null, 4)} ${JSON.stringify(directories, null, 4)} files ${JSON.stringify(files, null, 4)}`);
                calculateListStatistics(files, relationshipId, attributes);
                const filesDifferences = compareFiles(files, [], head(keys(target)));
                const dirDifferences = compareDirs(directories, []);
                const filesFiltered = filterGracePeriod(message, filesDifferences.TASK_TYPE_COPY);
                promises.push(sendTransferrerMessages(message, ACTION_COPY, filesFiltered, source, target));
                promises.push(sendTransferrerMessages(message, ACTION_COPY, dirDifferences.TASK_TYPE_SOURCE_ONLY, source, target));
                promises.push(sendScannerMessages(message, SCAN_MODE_SOURCE_ONLY, dirDifferences.TASK_TYPE_SOURCE_ONLY));
                promises.push(ignoreEntries(message, filesDifferences.TASK_TYPE_IGNORE));
                promises.push(ignoreEntries(message, dirDifferences.TASK_TYPE_IGNORE));
                break;
            }
            case SCAN_MODE_TARGET_ONLY: {
                // if delete on target is false, should never reach here
                const { directories, files } = await getDirectoriesAndFiles(newTarget, relationshipId, MOUNT_TARGET_SUFFIX);
                logger.debug(`directoriesOfEndPoint ${JSON.stringify(newTarget, null, 4)} ${JSON.stringify(directories, null, 4)} files ${JSON.stringify(files, null, 4)}`);
                const filesDifferences = compareFiles([], files);
                const dirDifferences = compareDirs([], directories);
                promises.push(sendTransferrerMessages(message, ACTION_DELETE, filesDifferences.TASK_TYPE_DELETE, source, target));
                promises.push(sendTransferrerMessages(message, ACTION_DELETE, dirDifferences.TASK_TYPE_TARGET_ONLY, source, target));
                promises.push(sendScannerMessages(message, SCAN_MODE_TARGET_ONLY, dirDifferences.TASK_TYPE_TARGET_ONLY));
                promises.push(ignoreEntries(message, filesDifferences.TASK_TYPE_IGNORE));
                promises.push(ignoreEntries(message, dirDifferences.TASK_TYPE_IGNORE));
                break;
            }
            default:
                logger.error(`[${relationshipId}] unknown scan mode ${type}`);
        }
        await Promise.all(promises);
    } catch (err) {
        const errorMessage = {
            step: LIST_STEP,
            relationshipId,
            status: STATUS_FAILED,
            objectType: TYPE_DIRECTORY,
            bytes: 0,
            items: 1,
            reason: JSON.stringify(err, null, 4),
            targetRequiredAction: STATUS_FAILED,
            attributes: message.attributes
        };

        sendStatisticsMessage(errorMessage);
        logger.error(`scan operation failed for rid: ${relationshipId} with error ${err} sending errorMessage`, errorMessage);
        failedRelationshipLogger.addContext('relationshipId', relationshipId);
        failedRelationshipLogger.info(`[${relationshipId}] [${LIST_STEP}] FAILED source`, source, ' and target: ', target, err);
    } finally {
        if (META_DATA[relationshipId]) {
            logger.debug('Deleting message from scanner queue', receiptHandle);
            await deleteMessage(META_DATA[relationshipId].queues.scannerQueue, receiptHandle);
        }
    }
}

async function handleMessageVisibility() {
    const runningTasks = INTERNAL_DISPATCHER_QUEUE.workersList().map(task => task.data);
    const waitingTasks = INTERNAL_DISPATCHER_QUEUE._tasks.toArray(); // eslint-disable-line no-underscore-dangle
    const queues = compact(Object.values(META_DATA).map(value => get(value, 'queues.scannerQueue')));
    logger.info('Internal dispatcher queue:', {
        runningTasks: runningTasks.length,
        waitingTasks: waitingTasks.length,
        queues: queues.length
    });
    const groupedMessagesByRelationshipId = groupBy(runningTasks.concat(waitingTasks), message => message.body.relationshipId);
    Object.keys(groupedMessagesByRelationshipId).forEach(relationshipId => {
        if (META_DATA[relationshipId]) {
            const receiptHandles = groupedMessagesByRelationshipId[relationshipId]
                .filter(task => Date.now() - task.lastVisibilityTimeoutChange > ms(SQS_MESSAGE_UNCHANGED_VISIBILITY_TIMEOUT_LIMIT))
                .map(task => task.receiptHandle);

            INTERNAL_DISPATCHER_QUEUE._tasks.toArray().forEach(message => { if (receiptHandles.includes(message.receiptHandle)) { message.lastVisibilityTimeoutChange = Date.now(); } }); // eslint-disable-line no-param-reassign, no-underscore-dangle
            INTERNAL_DISPATCHER_QUEUE.workersList().forEach(task => { if (receiptHandles.includes(task.data.receiptHandle)) { task.data.lastVisibilityTimeoutChange = Date.now(); } }); // eslint-disable-line no-param-reassign
            changeMessageVisibilityBatch(META_DATA[relationshipId].queues.scannerQueue, receiptHandles);
        }
    });
}

async function getWork(concurrency) {
    const queues = compact(Object.values(META_DATA).map(value => get(value, 'queues.scannerQueue')));

    if (queues.length > 0) {
        const waitingWorkersThreshold = config.get('workers.scanner.concurrency-eagerness-ratio') * concurrency;
        if (waitingWorkersThreshold > INTERNAL_DISPATCHER_QUEUE.length()) {
            const scannerQueue = head(shuffle(queues));
            const messages = await receiveMessages(scannerQueue);

            logger.debug(`Got ${messages.length} messages`, {
                running: INTERNAL_DISPATCHER_QUEUE.running(),
                length: INTERNAL_DISPATCHER_QUEUE.length()
            });
            messages.forEach(async message => {
                try {
                    if (validateMessage(message.body, 'SQS', scannerIncomingSQSSchema)) {
                        INTERNAL_DISPATCHER_QUEUE.push(message);
                    }
                } catch (err) {
                    logger.warn('Failed to validate scanner message', err);
                    deleteMessage(scannerQueue, message.receiptHandle);
                }
            });
        }
        setImmediate(() => getWork(concurrency));
    } else {
        setTimeout(() => getWork(concurrency), SQS_DEFAULT_WAIT_TIMEOUT * 1000);
    }
}

function handleCopyMessage(message) {
    const {
        relationshipId, requestId, scannerQueue, transferrerQueue, serverName, certificate
    } = message;
    logger.info(`handling IPC copy message for rid: ${relationshipId}, requestId: ${requestId}`);
    META_DATA[relationshipId] = {
        queues: {
            scannerQueue,
            transferrerQueue
        },
        status: IPC_TYPE_COPY,
        requestId
    };
    if (serverName && certificate) {
        META_DATA[relationshipId].snfs = { serverName, certificate };
    }
}

function handleDoneRemoveAbortUpdateMessage(message) {
    const { relationshipId, type } = message;
    delete META_DATA[relationshipId];
    INTERNAL_DISPATCHER_QUEUE.remove(task => task.data.body.relationshipId === relationshipId);
    logger.info(`Done handling command: ${type} for rid: ${relationshipId}`);
}

function boot() {
    const concurrency = config.get('workers.scanner.concurrency');
    logger.info(`Starting scanner #${cluster.worker.id}, PID ${process.pid}, concurrency ${concurrency} ...`);

    // queue should have 'async' keyword otherwise it does not work
    INTERNAL_DISPATCHER_QUEUE = queue(async scannerMessage => {
        const { body: { relationshipId } } = scannerMessage;
        if (META_DATA[relationshipId]) {
            return handleScannerMessage(scannerMessage);
        }
    }, concurrency);
    INTERNAL_DISPATCHER_QUEUE.drain = () => logger.info('Drain -> no more work');
    INTERNAL_DISPATCHER_QUEUE.error = (err, scannerMessage) => logger.error(`Got error for scanner message ${scannerMessage}`, err);
    getWork(concurrency);
    setInterval(() => handleMessageVisibility(), ms(SQS_CHANGE_MESSAGE_VISIBILITY_TIMEOUT_INTERVAL));
}

process.on('message', message => {
    logger.info('Got IPC message', message);
    if (validateMessage(message, 'IPC', scannerIncomingIPCMessagesSchemes[message.type])) {
        switch (message.type) {
            case IPC_TYPE_COPY:
                handleCopyMessage(message);
                break;
            case IPC_TYPE_DONE:
            case IPC_TYPE_UPDATE:
            case IPC_TYPE_REMOVE:
            case IPC_TYPE_ABORT:
                handleDoneRemoveAbortUpdateMessage(message);
                break;
            default:
                logger.error('Got unexpected message type', message.type);
        }
    }
});

if (process.env.NODE_ENV !== 'test') {
    boot();
}

module.exports = { compareFiles, compareDirs };
