const cluster = require('cluster');
const config = require('config');
const { queue } = require('async');
const { pipeline } = require('stream');
const retry = require('async-retry');
const {
    readlink, symlink, mkdirs, createReadStream, createWriteStream, remove, rename, chown, chmod, utimes, open, close, constants
} = require('fs-extra');
const { setTimeout } = require('timers');
const ms = require('ms');
const {
    shuffle, head, find, keys, _, compact, get, groupBy, uniq, isNil, cloneDeep
} = require('lodash');
const https = require('https');
const http = require('http');
const prettyBytes = require('bytes');
const { escape } = require('querystring');
const {
    gracefulJoin, mkdirsChmodAndChown, chownIfNeeded, copyNfsAcl
} = require('../lib/fs/fs');
const { getNewEndPointIfNeeded } = require('../lib/fs/mount_auxiliary');
const getLogger = require('../lib/logger/logger');
const {
    receiveMessages, changeMessageVisibilityBatch, deleteMessage, sendMessage
} = require('../lib/aws/sqs');
const {
    deleteObject, getS3ReadStream, upload, calculateObjectStoragePath, createTagsArray, getS3Object, copyObject, copyDirectory
} = require('../lib/aws/s3');
const {
    deleteBlob, createBlobReadStream, createWriteStreamToBlockBlob, getBlobToText, createMetadata, createBlockBlobFromText,
    startBlobCopy, setBlobMetadata, getBlobMetadata, waitForCopy, abortCopyBlob, setBlobTier
} = require('../lib/azure/blob');
const {
    getGcpCredentials, deleteGcpObject, getGcpReadStream, getGcpObjectMetadata, getGcpWriteStream, createGcpFolder, createGcpSymlink, createGcpMetadata, isObjectExists
} = require('../lib/gcp/storage');

const {
    getSecureReadStream, getSecureSymlinkContent, connectToServer, uploadSymlink,
    uploadDirectory, getSecureUploadStream, deleteFile
} = require('../lib/snfs/snfs');

const {
    TYPE_FILE,
    TYPE_DIRECTORY,
    TYPE_SYMBOLIC,
    IPC_TYPE_COPY,
    IPC_TYPE_REMOVE,
    IPC_TYPE_ABORT,
    IPC_TYPE_DONE,
    IPC_TYPE_UPDATE, ACTION_DELETE,
    ACTION_COPY,
    STATUS_SUCCEEDED,
    STATUS_FAILED,
    TRANSFER_STEP,
    ABORT_ERROR,
    CLEAN_RELATIONSHIP_RESOURCES_TIMEOUT,
    MOUNT_SOURCE_SUFFIX,
    MOUNT_TARGET_SUFFIX,
    SQS_MESSAGE_UNCHANGED_VISIBILITY_TIMEOUT_LIMIT,
    SQS_CHANGE_MESSAGE_VISIBILITY_TIMEOUT_INTERVAL,
    SQS_DEFAULT_WAIT_TIMEOUT
} = require('../lib/consts');
const sqsTransferrerIncomingMessagesSchemes = require('../messages/transferrer/sqs-incoming-messages');
const transferrerIncomingIPCMessagesSchemes = require('../messages/transferrer/ipc-incoming-messages');
const { copyStatisticsSchema } = require('../messages/schemas/statuses-outgoing-messages');
const validateMessage = require('../messages/schemas/message-validator');
const { startHealthLogInterval } = require('../lib/health');
const {
    deleteSgwsObject, uploadSgws, getSgwsObject, getSgwsReadStream, createSgwsMetadata
} = require('../lib/sgws/sgws');

const logger = getLogger();
const relationshipsLogger = getLogger('relationships');
const failedRelationshipLogger = getLogger('failed-relationships');

http.globalAgent.maxSockets = 500;
https.globalAgent.maxSockets = 500;

startHealthLogInterval();

process.on('uncaughtException', err => logger.error('Got uncaught exception', err));
process.on('unhandledRejection', event => logger.error('Got unhandled rejection', event));

const NETAPP_TEMP_FILE_REGEX = /^[.].+_[a-zA-Z\d]{7}_netapp$/;

const META_DATA = {};
let INTERNAL_DISPATCHER_QUEUE;

function isNetappEntry(path) {
    return NETAPP_TEMP_FILE_REGEX.test(path);
}

function getAbortError(path) {
    const error = new Error(`item: ${path} is aborted`);
    error.code = ABORT_ERROR;
    return error;
}

function isRelationshipAborted(relationshipId) {
    const metaData = META_DATA[relationshipId];
    return !metaData || metaData.status === IPC_TYPE_ABORT;
}

function addInTransitItem(relationshipId, targetPath, item) {
    META_DATA[relationshipId].items[targetPath] = item;
}

function removeInTransitItem(relationshipId, targetPath) {
    delete META_DATA[relationshipId].items[targetPath];
}

async function abortInTransitItem(target, entry, relationshipId) {
    const targetPath = gracefulJoin(target.local.path, entry.path);
    const metaData = META_DATA[relationshipId].items[targetPath];
    if (metaData) {
        const { tempTargetPath } = metaData;
        logger.info('Aborting file:', entry.path, 'With temp Path:', tempTargetPath);
        await remove(tempTargetPath);
        removeInTransitItem(relationshipId, targetPath);
    } else {
        logger.warn(`Metadata does not contain path: ${targetPath} could not abort item`);
    }
}

async function abortInTransitItemSNFSUpload(target, entry, relationshipId) {
    const targetPath = gracefulJoin(target.snfs.path, entry.path);
    const metaData = META_DATA[relationshipId].items[targetPath];
    if (metaData) {
        logger.info('Aborting file:', entry.path);
        removeInTransitItem(relationshipId, targetPath);
    } else {
        logger.warn(`Metadata does not contain path: ${targetPath} could not abort item`);
    }
}

async function abortInTransitUploadItem(target, entry, relationshipId) {
    const protocol = _(target).keys().head();
    const targetPath = calculateObjectStoragePath(target[protocol].path, entry.path);
    logger.info(`Aborting ${protocol} object: ${targetPath}`);
    removeInTransitItem(relationshipId, targetPath);
}

async function triggerAbort(relationshipId, targetPath) {
    const metaData = META_DATA[relationshipId].items[targetPath];
    if (metaData) {
        const {
            secureReadStream, client, readStream, writeStream, copyId, targetContainer
        } = metaData;
        logger.info('trigger aborting object path:', targetPath);
        if (copyId && targetContainer) {
            await abortCopyBlob(relationshipId, targetContainer, targetPath, copyId);
        } else if (secureReadStream && client) {
            const err = getAbortError(targetPath);
            secureReadStream.destroy(err);
            writeStream.destroy(err);
            client.close();
        } else {
            const err = getAbortError(targetPath);
            writeStream.destroy(err);
            readStream.destroy(err);
            if (client) {
                client.close();
            }
        }
    } else {
        logger.warn(`Metadata does not contain path: ${targetPath} could Not trigger abort item`);
    }
}

async function pipe(readStream, writeStream) {
    return new Promise((resolve, reject) => {
        pipeline(
            readStream,
            writeStream,
            err => {
                if (err) {
                    reject(err);
                } else {
                    resolve();
                }
            }
        );
    });
}

async function copyS3ToS3(source, target, entry) {
    const sourcePath = calculateObjectStoragePath(source.s3.path, entry.path);
    const targetPath = calculateObjectStoragePath(target.s3.path, entry.path);
    const sourceBucket = get(source, 's3.bucket');
    const kmsKeyId = get(target, 's3.encryption.kms.keyId');
    const storageClass = get(target, 's3.storageClass');
    const tagsToAdd = createTagsArray(target.s3.tags, entry);
    return copyObject(source.s3.region, target.s3.bucket, escape(`${sourceBucket}/${sourcePath}`), targetPath, storageClass, tagsToAdd, kmsKeyId, entry.size, entry.type);
}

async function sendTransferrerMessages(source, target, relationshipId, entry, attributes) {
    const queues = META_DATA[relationshipId] ? META_DATA[relationshipId].queues : undefined;

    if (queues) {
        if (entry.partsDetails) {
            const messageToSend = {
                type: ACTION_COPY,
                source,
                target,
                entry,
                relationshipId
            };
            return sendMessage(queues.transferrerQueue, messageToSend, attributes);
        }
    }
}

const DELETE_TABLE = [
    {
        target: 's3',
        type: TYPE_FILE,
        deleteHandler: async (target, entry) => {
            const newPath = calculateObjectStoragePath(target.s3.path, entry.path);
            await deleteObject(target.s3.region, target.s3.bucket, `${newPath}`);
        }
    },
    {
        target: 's3',
        type: TYPE_DIRECTORY,
        deleteHandler: async (target, entry) => {
            const newPath = calculateObjectStoragePath(target.s3.path, entry.path);
            await deleteObject(target.s3.region, target.s3.bucket, `${newPath}/`);
        }
    },
    {
        target: 's3',
        type: TYPE_SYMBOLIC,
        deleteHandler: async (target, entry) => {
            const newPath = calculateObjectStoragePath(target.s3.path, entry.path);
            await deleteObject(target.s3.region, target.s3.bucket, `${newPath}`);
        }
    },
    {
        target: 'sgws',
        type: TYPE_FILE,
        deleteHandler: async (target, entry, relationshipId) => {
            const newPath = calculateObjectStoragePath(target.sgws.path, entry.path);
            await deleteSgwsObject(target.sgws.host, target.sgws.port, target.sgws.accessKey, target.sgws.secretKey, target.sgws.bucket, newPath, MOUNT_TARGET_SUFFIX, relationshipId);
        }
    },
    {
        target: 'sgws',
        type: TYPE_DIRECTORY,
        deleteHandler: async (target, entry, relationshipId) => {
            const newPath = calculateObjectStoragePath(target.sgws.path, entry.path);
            await deleteSgwsObject(target.sgws.host, target.sgws.port, target.sgws.accessKey, target.sgws.secretKey, target.sgws.bucket, `${newPath}/`, MOUNT_TARGET_SUFFIX, relationshipId);
        }
    },
    {
        target: 'sgws',
        type: TYPE_SYMBOLIC,
        deleteHandler: async (target, entry, relationshipId) => {
            const newPath = calculateObjectStoragePath(target.sgws.path, entry.path);
            await deleteSgwsObject(target.sgws.host, target.sgws.port, target.sgws.accessKey, target.sgws.secretKey, target.sgws.bucket, newPath, MOUNT_TARGET_SUFFIX, relationshipId);
        }
    },

    {
        target: 'snfs',
        type: TYPE_FILE,
        deleteHandler: async (target, entry, relationshipId) => {
            await deleteFile(relationshipId, target.snfs.serverIP, target.snfs.port, META_DATA[relationshipId].snfs.certificate,
                META_DATA[relationshipId].snfs.serverName, META_DATA[relationshipId].requestId, gracefulJoin(target.snfs.path, entry.path));
        }
    },
    {
        target: 'snfs',
        type: TYPE_DIRECTORY,
        deleteHandler: async (target, entry, relationshipId) => {
            await deleteFile(relationshipId, target.snfs.serverIP, target.snfs.port, META_DATA[relationshipId].snfs.certificate,
                META_DATA[relationshipId].snfs.serverName, META_DATA[relationshipId].requestId, gracefulJoin(target.snfs.path, entry.path));
        }
    },
    {
        target: 'snfs',
        type: TYPE_SYMBOLIC,
        deleteHandler: async (target, entry, relationshipId) => {
            await deleteFile(relationshipId, target.snfs.serverIP, META_DATA[relationshipId].requestId, gracefulJoin(target.snfs.path, entry.path));
        }
    },

    {
        target: 'local',
        type: TYPE_FILE,
        deleteHandler: async (target, entry) => {
            await remove(gracefulJoin(target.local.path, entry.path));
        }
    },
    {
        target: 'local',
        type: TYPE_DIRECTORY,
        deleteHandler: async (target, entry) => {
            await remove(gracefulJoin(target.local.path, entry.path));
        }
    },
    {
        target: 'local',
        type: TYPE_SYMBOLIC,
        deleteHandler: async (target, entry) => {
            await remove(gracefulJoin(target.local.path, entry.path));
        }
    },
    {
        target: 'azure',
        type: TYPE_FILE,
        deleteHandler: async (target, entry, relationshipId) => {
            const newPath = calculateObjectStoragePath(target.azure.path, entry.path);
            await deleteBlob(target.azure.container, newPath, relationshipId);
        }
    },
    {
        target: 'azure',
        type: TYPE_DIRECTORY,
        deleteHandler: async (target, entry, relationshipId) => {
            const newPath = calculateObjectStoragePath(target.azure.path, entry.path);
            await deleteBlob(target.azure.container, `${newPath}/`, relationshipId);
        }
    },
    {
        target: 'azure',
        type: TYPE_SYMBOLIC,
        deleteHandler: async (target, entry, relationshipId) => {
            const newPath = calculateObjectStoragePath(target.azure.path, entry.path);
            await deleteBlob(target.azure.container, newPath, relationshipId);
        }
    },
    {
        target: 'gcp',
        type: TYPE_FILE,
        deleteHandler: async (target, entry, relationshipId) => {
            const newPath = calculateObjectStoragePath(target.gcp.path, entry.path);
            const credentials = await getGcpCredentials(relationshipId, MOUNT_TARGET_SUFFIX);
            await deleteGcpObject(credentials, target.gcp.bucket, `${newPath}`);
        }
    },
    {
        target: 'gcp',
        type: TYPE_DIRECTORY,
        deleteHandler: async (target, entry, relationshipId) => {
            const newPath = calculateObjectStoragePath(target.gcp.path, entry.path);
            const credentials = await getGcpCredentials(relationshipId, MOUNT_TARGET_SUFFIX);
            const noNameFileExists = await isObjectExists(credentials, target.gcp.bucket, `${newPath}/<no name>`);
            if (noNameFileExists) {
                await deleteGcpObject(credentials, target.gcp.bucket, `${newPath}/<no name>`);
            }
            await deleteGcpObject(credentials, target.gcp.bucket, `${newPath}/`);
        }
    },
    {
        target: 'gcp',
        type: TYPE_SYMBOLIC,
        deleteHandler: async (target, entry, relationshipId) => {
            const newPath = calculateObjectStoragePath(target.gcp.path, entry.path);
            const credentials = await getGcpCredentials(relationshipId, MOUNT_TARGET_SUFFIX);
            await deleteGcpObject(credentials, target.gcp.bucket, `${newPath}`);
        }
    }
];

const COPY_TABLE = [

    // ----------------------- local to S3 ----------------------//
    {
        source: 'local',
        target: 's3',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId) => {
            const targetPath = calculateObjectStoragePath(target.s3.path, entry.path);
            const kmsKeyId = get(target, 's3.encryption.kms.keyId');
            const storageClass = get(target, 's3.storageClass');

            if (entry.size === 0) {
                await upload(target.s3.region, target.s3.bucket, '', targetPath, createTagsArray(target.s3.tags, entry), kmsKeyId, storageClass).promise();
            } else {
                const readStream = createReadStream(gracefulJoin(source.local.path, entry.path));

                const managedUpload = upload(target.s3.region, target.s3.bucket, readStream, targetPath, createTagsArray(target.s3.tags, entry), kmsKeyId, storageClass);
                addInTransitItem(relationshipId, targetPath, { readStream, managedUpload });
                await managedUpload.promise();
                removeInTransitItem(relationshipId, targetPath);
            }
        },
        abortHandler: abortInTransitUploadItem
    },
    {
        source: 'local',
        target: 's3',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry) => {
            const newPath = calculateObjectStoragePath(target.s3.path, entry.path);
            const kmsKeyId = get(target, 's3.encryption.kms.keyId');
            const storageClass = get(target, 's3.storageClass');
            await upload(target.s3.region, target.s3.bucket, '', `${newPath}/`, createTagsArray(target.s3.tags, entry), kmsKeyId, storageClass).promise();
        }
    },
    {
        source: 'local',
        target: 's3',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry) => {
            const linkFileContent = await readlink(gracefulJoin(source.local.path, entry.path));
            const newPath = calculateObjectStoragePath(target.s3.path, entry.path);
            const kmsKeyId = get(target, 's3.encryption.kms.keyId');
            const storageClass = get(target, 's3.storageClass');
            await upload(target.s3.region, target.s3.bucket, linkFileContent, newPath, createTagsArray(target.s3.tags, entry), kmsKeyId, storageClass).promise();
        }
    },

    // ----------------------- snfs to local (download) ----------------------//

    {
        source: 'snfs',
        target: 'local',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId) => {
            await mkdirs(target.local.path);
            const targetPath = gracefulJoin(target.local.path, entry.path);

            if (entry.size === 0) {
                const fd = await open(targetPath, constants.O_CREAT);
                await Promise.all([
                    utimes(targetPath, entry.mtimeSec, entry.mtimeSec),
                    chmod(targetPath, entry.mode),
                    chownIfNeeded(targetPath, entry.uid, entry.gid)
                ]);
                await close(fd);
            } else {
                const tempTargetPath = gracefulJoin(target.local.path, entry.tempEntryName);
                const client = connectToServer(source.snfs.serverIP, source.snfs.port,
                    META_DATA[relationshipId].snfs.certificate, META_DATA[relationshipId].snfs.serverName);
                const secureReadStream = await getSecureReadStream(relationshipId, client,
                    META_DATA[relationshipId].requestId, gracefulJoin(source.snfs.path, entry.path));
                const writeStream = createWriteStream(tempTargetPath, { mode: entry.mode });
                addInTransitItem(relationshipId, targetPath, {
                    secureReadStream, writeStream, tempTargetPath, client
                });
                await pipe(secureReadStream, writeStream);
                secureReadStream.close();
                client.close();
                removeInTransitItem(relationshipId, targetPath);
                await Promise.all([
                    utimes(tempTargetPath, entry.mtimeSec, entry.mtimeSec),
                    chmod(tempTargetPath, entry.mode),
                    chownIfNeeded(tempTargetPath, entry.uid, entry.gid)
                ]);
                await rename(tempTargetPath, targetPath);
            }
        },
        abortHandler: abortInTransitItem
    },
    {
        source: 'snfs',
        target: 'local',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry) => {
            await mkdirsChmodAndChown(target.local.path, entry);
        }
    },
    {
        source: 'snfs',
        target: 'local',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry, relationshipId) => {
            await mkdirs(target.local.path);

            const sourcePath = gracefulJoin(source.snfs.path, entry.path);
            const targetPath = gracefulJoin(target.local.path, entry.path);

            const linkFileContent = await getSecureSymlinkContent(relationshipId, source.snfs.serverIP, source.snfs.port,
                META_DATA[relationshipId].snfs.certificate, META_DATA[relationshipId].snfs.serverName, META_DATA[relationshipId].requestId, sourcePath);

            await remove(targetPath);
            await symlink(linkFileContent, targetPath);
        }
    },

    // ----------------------- local to snfs (upload) ----------------------//

    {
        source: 'local',
        target: 'snfs',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId) => {
            const client = connectToServer(target.snfs.serverIP, target.snfs.port, META_DATA[relationshipId].snfs.certificate,
                META_DATA[relationshipId].snfs.serverName);
            const writeStream = await getSecureUploadStream(relationshipId, client,
                META_DATA[relationshipId].requestId, target.snfs.path, entry);
            const readStream = createReadStream(gracefulJoin(source.local.path, entry.path));
            const targetPath = gracefulJoin(target.snfs.path, entry.path);
            addInTransitItem(relationshipId, targetPath, { readStream, writeStream });
            await pipe(readStream, writeStream);
            removeInTransitItem(relationshipId, targetPath);
            client.close();
        },
        abortHandler: abortInTransitItemSNFSUpload
    },
    {
        source: 'local',
        target: 'snfs',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry, relationshipId) => {
            await uploadDirectory(relationshipId, target.snfs.serverIP, target.snfs.port, META_DATA[relationshipId].snfs.certificate,
                META_DATA[relationshipId].snfs.serverName, META_DATA[relationshipId].requestId, target.snfs.path, entry);
        }
    },
    {
        source: 'local',
        target: 'snfs',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry, relationshipId) => {
            const symlinkContent = await readlink(gracefulJoin(source.local.path, entry.path));
            await uploadSymlink(relationshipId, target.snfs.serverIP, target.snfs.port, META_DATA[relationshipId].snfs.certificate,
                META_DATA[relationshipId].snfs.serverName, META_DATA[relationshipId].requestId,
                target.snfs.path, entry, symlinkContent);
        }
    },

    // ----------------------- local to local ----------------------//

    {
        source: 'local',
        target: 'local',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId, attributes, originalCopy, originalTarget) => {
            await mkdirs(target.local.path);
            const sourcePath = gracefulJoin(source.local.path, entry.path);
            const targetPath = gracefulJoin(target.local.path, entry.path);
            if (entry.size === 0) {
                const fd = await open(targetPath, constants.O_CREAT);
                await Promise.all([
                    utimes(targetPath, entry.mtimeSec, entry.mtimeSec),
                    chmod(targetPath, entry.mode),
                    chownIfNeeded(targetPath, entry.uid, entry.gid)
                ]);
                await close(fd);
            } else {
                const tempTargetPath = gracefulJoin(target.local.path, entry.tempEntryName);
                if (entry.partsDetails) {
                    const {
                        partsDetails: {
                            start, end, partNumber, partSize
                        }, size
                    } = entry;
                    logger.info('Received message parts', JSON.stringify({
                        partNumber, start, end, partSize
                    }));
                    const readStream = createReadStream(sourcePath, {
                        start,
                        end
                    });
                    const writeStream = createWriteStream(tempTargetPath, { mode: entry.mode, start: entry.start, flags: 'a' });
                    addInTransitItem(relationshipId, targetPath, { readStream, writeStream, tempTargetPath });
                    await pipe(readStream, writeStream);
                    removeInTransitItem(relationshipId, targetPath);
                    // calcualte new partDetails
                    const newPartDetails = {
                        partNumber: partNumber + 1,
                        start: Math.min(end + 1, size), // because both start and end are inclusive
                        end: Math.min(end + partSize, size), // on node.js start and end are inclusive + we will always have start
                        partSize
                    };
                    logger.info('calculating new part details', JSON.stringify(newPartDetails));
                    if (newPartDetails.start === size) {
                        logger.info(`Done coping parts of file ${entry.path}`);
                        await Promise.all([
                            utimes(tempTargetPath, entry.mtimeSec, entry.mtimeSec),
                            chmod(tempTargetPath, entry.mode),
                            chownIfNeeded(tempTargetPath, entry.uid, entry.gid)
                        ]);
                        await rename(tempTargetPath, targetPath);
                    } else {
                        const entryToSend = cloneDeep(entry);
                        entryToSend.partsDetails = newPartDetails;
                        await sendTransferrerMessages(originalCopy, originalTarget, relationshipId, entryToSend, attributes);
                    }
                } else {
                    const readStream = createReadStream(gracefulJoin(source.local.path, entry.path));
                    const writeStream = createWriteStream(tempTargetPath, { mode: entry.mode });

                    // TODO - race condition - abort command between removeItem and copyToFS will delete file and create chmod/n rename errors..
                    addInTransitItem(relationshipId, targetPath, { readStream, writeStream, tempTargetPath });
                    await pipe(readStream, writeStream);
                    removeInTransitItem(relationshipId, targetPath);

                    await Promise.all([
                        utimes(tempTargetPath, entry.mtimeSec, entry.mtimeSec),
                        chmod(tempTargetPath, entry.mode),
                        chownIfNeeded(tempTargetPath, entry.uid, entry.gid)
                    ]);

                    await rename(tempTargetPath, targetPath);
                }
            }
            if (config.get('protocols.nfs.copy-acl')) {
                await copyNfsAcl(sourcePath, targetPath);
            }
        },
        abortHandler: abortInTransitItem
    },
    {
        source: 'local',
        target: 'local',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry) => {
            await mkdirsChmodAndChown(target.local.path, entry);
            if (config.get('protocols.nfs.copy-acl')) {
                await copyNfsAcl(gracefulJoin(source.local.path, entry.path), gracefulJoin(target.local.path, entry.path));
            }
        }
    },
    // We can't change mtime to symlink and as a result, it will be copied again and again
    {
        source: 'local',
        target: 'local',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry) => {
            await mkdirs(target.local.path);
            const sourcePath = gracefulJoin(source.local.path, entry.path);
            const targetPath = gracefulJoin(target.local.path, entry.path);
            const [linkFileContent] = await Promise.all([readlink(sourcePath), remove(targetPath)]);
            await symlink(linkFileContent, targetPath);
            if (config.get('protocols.nfs.copy-acl')) {
                await copyNfsAcl(sourcePath, targetPath);
            }
        }
    },
    {
        source: 's3',
        target: 'local',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId) => {
            await mkdirs(target.local.path);
            const tempTargetPath = gracefulJoin(target.local.path, entry.tempEntryName);
            const targetPath = gracefulJoin(target.local.path, entry.path);


            if (entry.size === 0) {
                const fd = await open(targetPath, constants.O_CREAT);
                await Promise.all([
                    utimes(targetPath, entry.mtimeSec, entry.mtimeSec),
                    chmod(targetPath, entry.mode),
                    chownIfNeeded(targetPath, entry.uid, entry.gid)
                ]);
                await close(fd);
            } else {
                const newPath = calculateObjectStoragePath(source.s3.path, entry.path);
                const readStream = getS3ReadStream(source.s3.region, source.s3.bucket, newPath);
                const writeStream = createWriteStream(tempTargetPath, { mode: entry.mode });
                addInTransitItem(relationshipId, targetPath, { readStream, writeStream });
                await pipe(readStream, writeStream);
                removeInTransitItem(relationshipId, targetPath);
                await Promise.all([utimes(tempTargetPath, entry.mtimeSec, entry.mtimeSec), chmod(tempTargetPath, entry.mode), chown(tempTargetPath, entry.uid, entry.gid)]);
                await rename(tempTargetPath, targetPath);
            }
        },
        abortHandler: abortInTransitItem
    },
    {
        source: 's3',
        target: 'local',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry) => {
            await mkdirsChmodAndChown(target.local.path, entry);
        }
    },
    {
        source: 's3',
        target: 'local',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry) => {
            await mkdirs(target.local.path);
            const symlinkSrcPath = calculateObjectStoragePath(source.s3.path, entry.path);
            const FilePointedFromSymlinkObject = await getS3Object(source.s3.region, source.s3.bucket, symlinkSrcPath);
            const FilePointedFromSymlinkRelativePath = FilePointedFromSymlinkObject.toString();
            const symlinkTrgPath = gracefulJoin(target.local.path, entry.path);
            await remove(symlinkTrgPath);
            await symlink(FilePointedFromSymlinkRelativePath, symlinkTrgPath);
        }
    },

    // ----------------------- AZURE to LOCAL ----------------------//
    {
        source: 'azure',
        target: 'local',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId) => {
            await mkdirs(target.local.path);
            const tempTargetPath = gracefulJoin(target.local.path, entry.tempEntryName);
            const targetPath = gracefulJoin(target.local.path, entry.path);
            if (entry.size === 0) {
                const fd = await open(targetPath, constants.O_CREAT);
                await Promise.all([
                    utimes(targetPath, entry.mtimeSec, entry.mtimeSec),
                    chmod(targetPath, entry.mode),
                    chownIfNeeded(targetPath, entry.uid, entry.gid)
                ]);
                await close(fd);
            } else {
                const newPath = calculateObjectStoragePath(source.azure.path, entry.path);
                const readStream = await createBlobReadStream(source.azure.container, newPath, relationshipId);
                const writeStream = createWriteStream(tempTargetPath, { mode: entry.mode });
                addInTransitItem(relationshipId, targetPath, { readStream, writeStream });
                await pipe(readStream, writeStream);
                removeInTransitItem(relationshipId, targetPath);
                await Promise.all([utimes(tempTargetPath, entry.mtimeSec, entry.mtimeSec), chmod(tempTargetPath, entry.mode), chown(tempTargetPath, entry.uid, entry.gid)]);
                await rename(tempTargetPath, targetPath);
            }
        },
        abortHandler: abortInTransitItem
    },
    {
        source: 'azure',
        target: 'local',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry) => {
            await mkdirsChmodAndChown(target.local.path, entry);
        }
    },
    {
        source: 'azure',
        target: 'local',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry, relationshipId) => {
            await mkdirs(target.local.path);
            const newPath = calculateObjectStoragePath(source.azure.path, entry.path);
            const linkFileContent = await getBlobToText(source.azure.container, newPath, relationshipId);
            const targetPath = gracefulJoin(target.local.path, entry.path);
            await remove(targetPath);
            await symlink(linkFileContent, targetPath);
        }
    },

    // ----------------------- LOCAL to AZURE ----------------------//
    {
        source: 'local',
        target: 'azure',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId) => {
            const targetPath = calculateObjectStoragePath(target.azure.path, entry.path);
            const metadata = createMetadata(target.azure.tags, entry);
            if (entry.size === 0) {
                await createBlockBlobFromText(target.azure.container, targetPath, undefined, metadata, relationshipId);
            } else {
                try {
                    const readStream = createReadStream(gracefulJoin(source.local.path, entry.path));
                    const writeStream = await createWriteStreamToBlockBlob(target.azure.container, targetPath, metadata, relationshipId);
                    addInTransitItem(relationshipId, targetPath, { readStream, writeStream });
                    await pipe(readStream, writeStream);
                    removeInTransitItem(relationshipId, targetPath);
                } catch (err) {
                    // https://github.com/Azure/azure-storage-node/issues/335 - solution to the error : Error StorageError: The uncommitted block count cannot exceed the maximum limit of 100,000 blocks
                    logger.warn('failed writingStreamToBlockBlob cleaning block blobs before throwing an error');
                    await createBlockBlobFromText(target.azure.container, targetPath, undefined, metadata, relationshipId);
                    await deleteBlob(target.azure.container, targetPath, relationshipId);
                    throw err;
                }
            }
            await setBlobTier(target.azure.container, targetPath, relationshipId, get(target, 'azure.blobTier'));
        },
        abortHandler: abortInTransitUploadItem
    },
    {
        source: 'local',
        target: 'azure',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry, relationshipId) => {
            const newPath = calculateObjectStoragePath(target.azure.path, entry.path);
            const metadata = createMetadata(target.azure.tags, entry);
            await createBlockBlobFromText(target.azure.container, `${newPath}/`, undefined, metadata, relationshipId);
            await setBlobTier(target.azure.container, `${newPath}/`, relationshipId, get(target, 'azure.blobTier'));
        }
    },
    {
        source: 'local',
        target: 'azure',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry, relationshipId) => {
            const linkFileContent = await readlink(gracefulJoin(source.local.path, entry.path));
            const newPath = calculateObjectStoragePath(target.azure.path, entry.path);
            const metadata = createMetadata(target.azure.tags, entry);
            await createBlockBlobFromText(target.azure.container, newPath, linkFileContent, metadata, relationshipId);
            await setBlobTier(target.azure.container, newPath, relationshipId, get(target, 'azure.blobTier'));
        }
    },

    // ----------------------- LOCAL to SGWS ----------------------//
    {
        source: 'local',
        target: 'sgws',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId) => {
            const targetPath = calculateObjectStoragePath(target.sgws.path, entry.path);
            if (entry.size === 0) {
                const managedUpload = await uploadSgws(target.sgws.host, target.sgws.port, target.sgws.accessKey,
                    target.sgws.secretKey, target.sgws.bucket, '', targetPath, createSgwsMetadata(entry, target.sgws.tags || []), relationshipId);
                await managedUpload.promise();
            } else {
                const readStream = createReadStream(gracefulJoin(source.local.path, entry.path));
                const managedUpload = await uploadSgws(target.sgws.host, target.sgws.port, target.sgws.accessKey,
                    target.sgws.secretKey, target.sgws.bucket, readStream, targetPath, createSgwsMetadata(entry, target.sgws.tags || []), relationshipId);
                addInTransitItem(relationshipId, targetPath, { readStream, managedUpload });
                await managedUpload.promise();
                removeInTransitItem(relationshipId, targetPath);
            }
        },
        abortHandler: abortInTransitUploadItem
    },
    {
        source: 'local',
        target: 'sgws',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry, relationshipId) => {
            const newPath = calculateObjectStoragePath(target.sgws.path, entry.path);
            const managedUpload = await uploadSgws(target.sgws.host, target.sgws.port, target.sgws.accessKey,
                target.sgws.secretKey, target.sgws.bucket, '', `${newPath}/`, createSgwsMetadata(entry, target.sgws.tags || []), relationshipId);
            await managedUpload.promise();
        }
    },
    {
        source: 'local',
        target: 'sgws',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry, relationshipId) => {
            const linkFileContent = await readlink(gracefulJoin(source.local.path, entry.path));
            const newPath = calculateObjectStoragePath(target.sgws.path, entry.path);
            const managedUpload = await uploadSgws(target.sgws.host, target.sgws.port, target.sgws.accessKey,
                target.sgws.secretKey, target.sgws.bucket, linkFileContent, newPath, createSgwsMetadata(entry, target.sgws.tags || []), relationshipId);
            await managedUpload.promise();
        }
    },
    // ----------------------- SGWS to local ----------------------//
    {
        source: 'sgws',
        target: 'local',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId) => {
            await mkdirs(target.local.path);
            const tempTargetPath = gracefulJoin(target.local.path, entry.tempEntryName);
            const targetPath = gracefulJoin(target.local.path, entry.path);

            if (entry.size === 0) {
                const fd = await open(targetPath, constants.O_CREAT);
                await Promise.all([
                    utimes(targetPath, entry.mtimeSec, entry.mtimeSec),
                    chmod(targetPath, entry.mode),
                    chownIfNeeded(targetPath, entry.uid, entry.gid)
                ]);
                await close(fd);
            } else {
                const newPath = calculateObjectStoragePath(source.sgws.path, entry.path);
                const readStream = await getSgwsReadStream(source.sgws.host, source.sgws.port, source.sgws.accessKey, source.sgws.secretKey, source.sgws.bucket, newPath, relationshipId);
                const writeStream = createWriteStream(tempTargetPath, { mode: entry.mode });
                addInTransitItem(relationshipId, targetPath, { readStream, writeStream });
                await pipe(readStream, writeStream);
                removeInTransitItem(relationshipId, targetPath);
                await Promise.all([utimes(tempTargetPath, entry.mtimeSec, entry.mtimeSec), chmod(tempTargetPath, entry.mode), chown(tempTargetPath, entry.uid, entry.gid)]);
                await rename(tempTargetPath, targetPath);
            }
        },
        abortHandler: abortInTransitItem
    },
    {
        source: 'sgws',
        target: 'local',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry) => {
            await mkdirsChmodAndChown(target.local.path, entry);
        }
    },
    {
        source: 'sgws',
        target: 'local',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry, relationshipId) => {
            await mkdirs(target.local.path);
            const newPath = calculateObjectStoragePath(source.sgws.path, entry.path);
            const linkFileContent = await getSgwsObject(source.sgws.host, source.sgws.port, source.sgws.accessKey, source.sgws.secretKey, source.sgws.bucket, newPath, MOUNT_SOURCE_SUFFIX, relationshipId);
            const targetPath = gracefulJoin(target.local.path, entry.path);
            await remove(targetPath);
            await symlink(linkFileContent, targetPath);
        }
    },
    // ----------------------- gcp to local ----------------------//
    {
        source: 'gcp',
        target: 'local',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId) => {
            await mkdirs(target.local.path);
            const tempTargetPath = gracefulJoin(target.local.path, entry.tempEntryName);
            const targetPath = gracefulJoin(target.local.path, entry.path);
            if (entry.size === 0) {
                const fd = await open(targetPath, constants.O_CREAT);
                await Promise.all([
                    utimes(targetPath, entry.mtimeSec, entry.mtimeSec),
                    chmod(targetPath, entry.mode),
                    chownIfNeeded(targetPath, entry.uid, entry.gid)
                ]);
                await close(fd);
            } else {
                const newPath = calculateObjectStoragePath(source.gcp.path, entry.path);
                const credentials = await getGcpCredentials(relationshipId, MOUNT_SOURCE_SUFFIX);
                const readStream = getGcpReadStream(credentials, source.gcp.bucket, newPath, relationshipId);
                const writeStream = createWriteStream(tempTargetPath, { mode: entry.mode });
                addInTransitItem(relationshipId, targetPath, { readStream, writeStream });
                await pipe(readStream, writeStream);
                removeInTransitItem(relationshipId, targetPath);
                await Promise.all([utimes(tempTargetPath, entry.mtimeSec, entry.mtimeSec), chmod(tempTargetPath, entry.mode), chown(tempTargetPath, entry.uid, entry.gid)]);
                await rename(tempTargetPath, targetPath);
            }
        },
        abortHandler: abortInTransitItem
    },
    {
        source: 'gcp',
        target: 'local',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry) => {
            await mkdirsChmodAndChown(target.local.path, entry);
        }
    },
    {
        source: 'gcp',
        target: 'local',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry, relationshipId) => {
            // TODO check if we could use getFile or how we could get the file content ( not stream )
            await mkdirs(target.local.path);
            const newPath = calculateObjectStoragePath(source.gcp.path, entry.path);
            const credentials = await getGcpCredentials(relationshipId, MOUNT_SOURCE_SUFFIX);
            const linkFileMetaData = await getGcpObjectMetadata(credentials, source.gcp.bucket, newPath, relationshipId, MOUNT_SOURCE_SUFFIX);
            const linkFileContent = linkFileMetaData.metadata.symlink;
            const targetPath = gracefulJoin(target.local.path, entry.path);
            await remove(targetPath);
            await symlink(linkFileContent, targetPath);
        }
    },
    // ----------------------- local to gcp ----------------------//
    {
        source: 'local',
        target: 'gcp',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId) => {
            const readStream = createReadStream(gracefulJoin(source.local.path, entry.path));
            const targetPath = calculateObjectStoragePath(target.gcp.path, entry.path);
            const credentials = await getGcpCredentials(relationshipId, MOUNT_TARGET_SUFFIX);
            const writeStream = getGcpWriteStream(credentials, target.gcp.bucket, targetPath, createGcpMetadata(target.gcp.tags, entry));
            addInTransitItem(relationshipId, targetPath, { readStream, writeStream });
            await pipe(readStream, writeStream);
            removeInTransitItem(relationshipId, targetPath);
        },
        abortHandler: abortInTransitUploadItem
    },
    {
        source: 'local',
        target: 'gcp',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry, relationshipId) => {
            const targetPath = calculateObjectStoragePath(target.gcp.path, entry.path);
            const credentials = await getGcpCredentials(relationshipId, MOUNT_TARGET_SUFFIX);
            await createGcpFolder(credentials, target.gcp.bucket, targetPath, createGcpMetadata(target.gcp.tags, entry), relationshipId);
        }
    },
    {
        source: 'local',
        target: 'gcp',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry, relationshipId) => {
            const linkFileContent = await readlink(gracefulJoin(source.local.path, entry.path));
            const targetPath = calculateObjectStoragePath(target.gcp.path, entry.path);
            const credentials = await getGcpCredentials(relationshipId, MOUNT_TARGET_SUFFIX);
            await createGcpSymlink(credentials, target.gcp.bucket, targetPath, linkFileContent, createGcpMetadata(target.gcp.tags, entry, linkFileContent), relationshipId);
        }
    },
    // ----------------------- AZURE to AZURE ----------------------//
    {
        source: 'azure',
        target: 'azure',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId) => {
            const targetPath = calculateObjectStoragePath(target.azure.path, entry.path);
            const sourcePath = calculateObjectStoragePath(source.azure.path, entry.path);
            const metadata = createMetadata(target.azure.tags, entry);
            // finding the current metadata of the copied blob
            const sourceBlobMetadata = await getBlobMetadata(relationshipId, MOUNT_SOURCE_SUFFIX, source.azure.container, sourcePath);
            const targetBlobMetadata = Object.assign({}, metadata, sourceBlobMetadata);
            const copyId = await startBlobCopy(relationshipId,
                source.azure.container, target.azure.container, sourcePath, targetPath);
            addInTransitItem(relationshipId, targetPath, { copyId, targetContainer: target.azure.container });
            await waitForCopy(relationshipId, target.azure.container, targetPath);
            removeInTransitItem(relationshipId, targetPath);
            await setBlobMetadata(relationshipId, MOUNT_TARGET_SUFFIX,
                target.azure.container, targetPath, targetBlobMetadata);
            await setBlobTier(target.azure.container, targetPath, relationshipId, get(target, 'azure.blobTier'));
        },
        abortHandler: abortInTransitUploadItem
    },
    {
        source: 'azure',
        target: 'azure',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry, relationshipId) => {
            const newPath = calculateObjectStoragePath(target.azure.path, entry.path);
            const metadata = createMetadata(target.azure.tags, entry);
            await createBlockBlobFromText(target.azure.container, `${newPath}/`, undefined, metadata, relationshipId);
            await setBlobTier(target.azure.container, `${newPath}/`, relationshipId, get(target, 'azure.blobTier'));
        }
    },
    {
        source: 'azure',
        target: 'azure',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry, relationshipId) => {
            const targetPath = calculateObjectStoragePath(target.azure.path, entry.path);
            const sourcePath = calculateObjectStoragePath(source.azure.path, entry.path);
            const metadata = createMetadata(target.azure.tags, entry);
            // finding the current metadata of the copied blob
            const sourceBlobMetadata = await getBlobMetadata(relationshipId, MOUNT_SOURCE_SUFFIX, source.azure.container, sourcePath);
            const targetBlobMetadata = Object.assign({}, metadata, sourceBlobMetadata);
            const copyId = await startBlobCopy(relationshipId,
                source.azure.container, target.azure.container, sourcePath, targetPath);
            addInTransitItem(relationshipId, targetPath, { copyId, targetContainer: target.azure.container });
            await waitForCopy(relationshipId, target.azure.container, targetPath);
            removeInTransitItem(relationshipId, targetPath);
            await setBlobMetadata(relationshipId, MOUNT_TARGET_SUFFIX,
                target.azure.container, targetPath, targetBlobMetadata);
            await setBlobTier(target.azure.container, targetPath, relationshipId, get(target, 'azure.blobTier'));
        },
        abortHandler: abortInTransitUploadItem
    },
    // ----------------------- S3 to S3 ----------------------//

    {
        source: 's3',
        target: 's3',
        type: TYPE_FILE,
        copyHandler: (source, target, entry) => copyS3ToS3(source, target, entry),
        abortHandler: abortInTransitUploadItem
    },
    {
        source: 's3',
        target: 's3',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry) => {
            const sourcePath = calculateObjectStoragePath(source.s3.path, entry.path);
            const targetPath = calculateObjectStoragePath(target.s3.path, entry.path);
            const sourceBucket = get(source, 's3.bucket');
            const kmsKeyId = get(target, 's3.encryption.kms.keyId');
            const storageClass = get(target, 's3.storageClass');
            const tagsToAdd = createTagsArray(target.s3.tags, entry);
            await copyDirectory(source.s3.region, target.s3.bucket, encodeURI(`${sourceBucket}/${sourcePath}/`), `${targetPath}/`, storageClass, tagsToAdd, kmsKeyId);
        }
    },
    {
        source: 's3',
        target: 's3',
        type: TYPE_SYMBOLIC,
        copyHandler: (source, target, entry) => copyS3ToS3(source, target, entry),
        abortHandler: abortInTransitUploadItem
    },
    // ----------------------- S3 to SGWS ----------------------//
    {
        source: 's3',
        target: 'sgws',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId) => {
            const targetPath = calculateObjectStoragePath(target.sgws.path, entry.path);
            if (entry.size === 0) {
                const managedUpload = await uploadSgws(target.sgws.host, target.sgws.port, target.sgws.accessKey,
                    target.sgws.secretKey, target.sgws.bucket, '', targetPath, createSgwsMetadata(entry, target.sgws.tags || []), relationshipId);
                await managedUpload.promise();
            } else {
                const newPath = calculateObjectStoragePath(source.s3.path, entry.path);
                const readStream = getS3ReadStream(source.s3.region, source.s3.bucket, newPath);
                const managedUpload = await uploadSgws(target.sgws.host, target.sgws.port, target.sgws.accessKey,
                    target.sgws.secretKey, target.sgws.bucket, readStream, targetPath, createSgwsMetadata(entry, target.sgws.tags || []), relationshipId);
                addInTransitItem(relationshipId, targetPath, { readStream, managedUpload });
                await managedUpload.promise();
                removeInTransitItem(relationshipId, targetPath);
            }
        },
        abortHandler: abortInTransitUploadItem
    },
    {
        source: 's3',
        target: 'sgws',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry, relationshipId) => {
            const newPath = calculateObjectStoragePath(target.sgws.path, entry.path);
            const managedUpload = await uploadSgws(target.sgws.host, target.sgws.port, target.sgws.accessKey,
                target.sgws.secretKey, target.sgws.bucket, '', `${newPath}/`, createSgwsMetadata(entry, target.sgws.tags || []), relationshipId);
            await managedUpload.promise();
        }
    },
    {
        source: 's3',
        target: 'sgws',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry, relationshipId) => {
            const symlinkSrcPath = calculateObjectStoragePath(source.s3.path, entry.path);
            const FilePointedFromSymlinkObject = await getS3Object(source.s3.region, source.s3.bucket, symlinkSrcPath);
            const linkFileContent = FilePointedFromSymlinkObject.toString();
            const newPath = calculateObjectStoragePath(target.sgws.path, entry.path);
            const managedUpload = await uploadSgws(target.sgws.host, target.sgws.port, target.sgws.accessKey,
                target.sgws.secretKey, target.sgws.bucket, linkFileContent, newPath, createSgwsMetadata(entry, target.sgws.tags || []), relationshipId);
            await managedUpload.promise();
        }
    },
    // ----------------------- SGWS to S3 ----------------------//
    {
        source: 'sgws',
        target: 's3',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId) => {
            const targetPath = calculateObjectStoragePath(target.s3.path, entry.path);
            const kmsKeyId = get(target, 's3.encryption.kms.keyId');
            const storageClass = get(target, 's3.storageClass');
            if (entry.size === 0) {
                await upload(target.s3.region, target.s3.bucket, '', targetPath, createTagsArray(target.s3.tags, entry), kmsKeyId, storageClass).promise();
            } else {
                const newPath = calculateObjectStoragePath(source.sgws.path, entry.path);
                const readStream = await getSgwsReadStream(source.sgws.host, source.sgws.port, source.sgws.accessKey, source.sgws.secretKey, source.sgws.bucket, newPath, relationshipId);
                const managedUpload = upload(target.s3.region, target.s3.bucket, readStream, targetPath, createTagsArray(target.s3.tags, entry), kmsKeyId, storageClass);
                addInTransitItem(relationshipId, targetPath, { readStream, managedUpload });
                await managedUpload.promise();
                removeInTransitItem(relationshipId, targetPath);
            }
        },
        abortHandler: abortInTransitUploadItem
    },
    {
        source: 'sgws',
        target: 's3',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry) => {
            const newPath = calculateObjectStoragePath(target.s3.path, entry.path);
            const kmsKeyId = get(target, 's3.encryption.kms.keyId');
            const storageClass = get(target, 's3.storageClass');
            await upload(target.s3.region, target.s3.bucket, '', `${newPath}/`, createTagsArray(target.s3.tags, entry), kmsKeyId, storageClass).promise();
        }
    },
    {
        source: 'sgws',
        target: 's3',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry, relationshipId) => {
            const newPath = calculateObjectStoragePath(source.sgws.path, entry.path);
            const linkFileContent = await getSgwsObject(source.sgws.host, source.sgws.port, source.sgws.accessKey, source.sgws.secretKey, source.sgws.bucket, newPath, MOUNT_SOURCE_SUFFIX, relationshipId);
            const kmsKeyId = get(target, 's3.encryption.kms.keyId');
            const storageClass = get(target, 's3.storageClass');
            const targetPath = calculateObjectStoragePath(target.s3.path, entry.path);
            await upload(target.s3.region, target.s3.bucket, linkFileContent, targetPath, createTagsArray(target.s3.tags, entry), kmsKeyId, storageClass).promise();
        }
    },
    // ----------------------- SGWS to SGWS----------------------//
    {
        source: 'sgws',
        target: 'sgws',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId) => {
            const targetPath = calculateObjectStoragePath(target.sgws.path, entry.path);
            if (entry.size === 0) {
                const managedUpload = await uploadSgws(target.sgws.host, target.sgws.port, target.sgws.accessKey,
                    target.sgws.secretKey, target.sgws.bucket, '', targetPath, createSgwsMetadata(entry, target.sgws.tags || []), relationshipId);
                await managedUpload.promise();
            } else {
                const newPath = calculateObjectStoragePath(source.sgws.path, entry.path);
                const readStream = await getSgwsReadStream(source.sgws.host, source.sgws.port, source.sgws.accessKey, source.sgws.secretKey, source.sgws.bucket, newPath, relationshipId);
                const managedUpload = await uploadSgws(target.sgws.host, target.sgws.port, target.sgws.accessKey,
                    target.sgws.secretKey, target.sgws.bucket, readStream, targetPath, createSgwsMetadata(entry, target.sgws.tags || []), relationshipId);
                addInTransitItem(relationshipId, targetPath, { readStream, managedUpload });
                await managedUpload.promise();
                removeInTransitItem(relationshipId, targetPath);
            }
        },
        abortHandler: abortInTransitUploadItem
    },
    {
        source: 'sgws',
        target: 'sgws',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry, relationshipId) => {
            const newPath = calculateObjectStoragePath(target.sgws.path, entry.path);
            const managedUpload = await uploadSgws(target.sgws.host, target.sgws.port, target.sgws.accessKey,
                target.sgws.secretKey, target.sgws.bucket, '', `${newPath}/`, createSgwsMetadata(entry, target.sgws.tags || []), relationshipId);
            await managedUpload.promise();
        }
    },
    {
        source: 'sgws',
        target: 'sgws',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry, relationshipId) => {
            const newPath = calculateObjectStoragePath(source.sgws.path, entry.path);
            const linkFileContent = await getSgwsObject(source.sgws.host, source.sgws.port, source.sgws.accessKey, source.sgws.secretKey, source.sgws.bucket, newPath, MOUNT_SOURCE_SUFFIX, relationshipId);

            const targetPath = calculateObjectStoragePath(target.sgws.path, entry.path);
            const managedUpload = await uploadSgws(target.sgws.host, target.sgws.port, target.sgws.accessKey,
                target.sgws.secretKey, target.sgws.bucket, linkFileContent, targetPath, createSgwsMetadata(entry, target.sgws.tags || []), relationshipId);
            await managedUpload.promise();
        }
    }
];

function sendStatisticsMessage(statisticsMessage) {
    // send status message back
    const isValid = validateMessage(statisticsMessage, 'SQS', copyStatisticsSchema);
    if (isValid) {
        process.send(statisticsMessage);
    }
}

async function handleDoneRemoveAbortUpdateMessage(message) {
    const {
        relationshipId, type
    } = message;
    if (META_DATA[relationshipId]) {
        META_DATA[relationshipId].status = IPC_TYPE_ABORT;
        delete META_DATA[relationshipId].queues;
        INTERNAL_DISPATCHER_QUEUE.remove(task => task.data.body.relationshipId === relationshipId);
        const itemsInProgress = META_DATA[relationshipId].items;
        logger.info(`For Command ${type}, aborting ${keys(itemsInProgress).length} items`);
        keys(itemsInProgress).forEach(async targetPath => {
            try {
                await triggerAbort(relationshipId, targetPath);
            } catch (err) {
                logger.error('error after abort operation', err);
            }
        });
    }
    // TODO - delete relationship from realtime when all items Done aborting (For now after 'X' minute - need fixing).
    setTimeout(() => {
        logger.info(`Done handling command: ${type} for rid: ${relationshipId}`);
        delete META_DATA[relationshipId];
    }, CLEAN_RELATIONSHIP_RESOURCES_TIMEOUT);
}

async function handleCopy(source, target, entry, relationshipId, attributes, originalCopy, originalTarget) {
    const sourceProtocol = _(source).keys().head();
    const targetProtocol = _(target).keys().head();
    const match = find(COPY_TABLE, {
        source: sourceProtocol,
        target: targetProtocol,
        type: entry.type
    });
    const startTime = Date.now();
    try {
        await retry(async bail => {
            if (isRelationshipAborted(relationshipId)) {
                if (sourceProtocol === 'local' && targetProtocol === 'local' && entry.partSize) {
                    relationshipsLogger.info('Removing temporary file due to copy abort:', entry.tempTargetPath);
                    await remove(entry.tempTargetPath);
                }
                bail(getAbortError(gracefulJoin(source.local.path, entry.path)));
            } else {
                await match.copyHandler(source, target, entry, relationshipId, attributes, originalCopy, originalTarget);
            }
        }, {
            retries: config.get('retry.retries'),
            onRetry: async err => {
                relationshipsLogger.warn(`[${relationshipId}] RETRYING copy from:`, JSON.stringify(source), 'to:', JSON.stringify(target), 'entry:', JSON.stringify(entry), err);
                const targetPath = get(target, 'local.path');
                if (targetPath) {
                    const tempTargetPath = gracefulJoin(targetPath, entry.tempEntryName);
                    try {
                        relationshipsLogger.info('Removing temporary file due to copy failure:', tempTargetPath);
                        await retry(async () => { await remove(tempTargetPath); }, { retries: config.get('retry.retries') });
                    } catch (error) {
                        relationshipsLogger.error(`Failed removing temporary file ${tempTargetPath}`, error);
                    }
                }
            }
        });
        relationshipsLogger.info(`[${relationshipId}] COMPLETED copy from:`, JSON.stringify(source), 'to:', JSON.stringify(target), 'entry:', JSON.stringify(entry), `took: ${Date.now() - startTime}`);
        sendStatisticsMessage({
            step: TRANSFER_STEP,
            relationshipId,
            targetRequiredAction: ACTION_COPY,
            bytes: !(isNil(entry.partsDetails)) ? entry.partsDetails.end - entry.partsDetails.start : entry.size,
            items: (entry.partsDetails && entry.size === entry.partsDetails.end) || (!entry.partsDetails) ? 1 : 0,
            objectType: entry.type,
            path: entry.path,
            status: STATUS_SUCCEEDED,
            attributes
        });
    } catch (err) {
        if (err.code === ABORT_ERROR) {
            await match.abortHandler(target, entry, relationshipId);
            relationshipsLogger.info(`[${relationshipId}] ABORTED copy from:`, JSON.stringify(source), 'to:', JSON.stringify(target), 'entry:', JSON.stringify(entry));
        } else {
            relationshipsLogger.error(`[${relationshipId}] FAILED copy from:`, JSON.stringify(source), 'to:', JSON.stringify(target), 'entry:', JSON.stringify(entry), err);
            throw err;
        }
    }
}

async function handleDelete(target, entry, relationshipId, attributes) {
    const match = find(DELETE_TABLE, { target: _(target).keys().head(), type: entry.type });

    relationshipsLogger.info(`[${relationshipId}] STARTING deletion from:`, target, 'entry:', entry);

    try {
        await retry(() => match.deleteHandler(target, entry, relationshipId), {
            retries: config.get('retry.retries'),
            onRetry: err => relationshipsLogger.info(`[${relationshipId}] RETRYING deletion from:`, target, 'entry:', entry, err)
        });
        relationshipsLogger.info(`[${relationshipId}] COMPLETED deletion from:`, target, 'entry:', entry);
        if (!isNetappEntry(entry.path)) {
            sendStatisticsMessage({
                step: TRANSFER_STEP,
                relationshipId,
                targetRequiredAction: ACTION_DELETE,
                bytes: entry.size,
                items: 1,
                objectType: entry.type,
                status: STATUS_SUCCEEDED,
                path: entry.path,
                attributes
            });
        }
    } catch (err) {
        relationshipsLogger.error(`[${relationshipId}] FAILED deletion from:`, target, 'entry:', entry, err);
        throw err;
    }
}

async function handleTransferrerMessage(message) {
    // TODO think about moving type into attributes in all SQS messages
    const {
        body: {
            source, target, entry, relationshipId, type
        },
        receiptHandle, attributes
    } = message;

    try {
        const newSource = getNewEndPointIfNeeded(relationshipId, META_DATA[relationshipId].requestId, MOUNT_SOURCE_SUFFIX, source);
        const newTarget = getNewEndPointIfNeeded(relationshipId, META_DATA[relationshipId].requestId, MOUNT_TARGET_SUFFIX, target);

        switch (type) {
            case ACTION_COPY:
                await handleCopy(newSource, newTarget, entry, relationshipId, attributes, source, target);
                break;
            case ACTION_DELETE:
                await handleDelete(newTarget, entry, relationshipId, attributes);
                // TODO - think - we should delay directories delete messages in scanner
                break;
            default:
                throw new Error(`Got unexpected operation type: ${type}`);
        }
    } catch (err) {
        const errorMessage = {
            step: TRANSFER_STEP,
            relationshipId,
            targetRequiredAction: type,
            bytes: entry.size,
            items: 1,
            objectType: entry.type,
            path: entry.path,
            status: STATUS_FAILED,
            reason: err,
            attributes
        };

        sendStatisticsMessage(errorMessage);
        failedRelationshipLogger.addContext('relationshipId', relationshipId);
        failedRelationshipLogger.info(`[${relationshipId}] [${type}] FAILED: source`, source, 'target:', target, 'entry:', entry, err);
    } finally {
        const transferrerQueue = get(META_DATA, `${relationshipId}.queues.transferrerQueue`);
        if (transferrerQueue) {
            logger.debug('Deleting transferrer message', receiptHandle);
            deleteMessage(transferrerQueue, receiptHandle);
        }
    }
}

async function handleMessageVisibility() {
    const runningTasks = INTERNAL_DISPATCHER_QUEUE.workersList().map(task => task.data);
    const waitingTasks = INTERNAL_DISPATCHER_QUEUE._tasks.toArray(); // eslint-disable-line no-underscore-dangle
    const queues = compact(Object.values(META_DATA).map(value => get(value, 'queues.transferrerQueue')));
    logger.info('Internal dispatcher queue:', { runningTasks: runningTasks.length, waitingTasks: waitingTasks.length, queues: queues.length });
    const groupedMessagesByRelationshipId = groupBy(runningTasks.concat(waitingTasks), message => message.body.relationshipId);
    Object.keys(groupedMessagesByRelationshipId).forEach(relationshipId => {
        if (META_DATA[relationshipId]) {
            const receiptHandles = groupedMessagesByRelationshipId[relationshipId]
                .filter(task => Date.now() - task.lastVisibilityTimeoutChange > ms(SQS_MESSAGE_UNCHANGED_VISIBILITY_TIMEOUT_LIMIT))
                .map(task => task.receiptHandle);

            INTERNAL_DISPATCHER_QUEUE._tasks.toArray().forEach(message => { if (receiptHandles.includes(message.receiptHandle)) { message.lastVisibilityTimeoutChange = Date.now(); } }); // eslint-disable-line no-param-reassign, no-underscore-dangle
            INTERNAL_DISPATCHER_QUEUE.workersList().forEach(task => { if (receiptHandles.includes(task.data.receiptHandle)) { task.data.lastVisibilityTimeoutChange = Date.now(); } }); // eslint-disable-line no-param-reassign
            changeMessageVisibilityBatch(META_DATA[relationshipId].queues.transferrerQueue, uniq(receiptHandles));
        }
    });
}

// Sums up the total size of all the waiting and running tasks (objects)
function totalTasksSize() {
    const reducer = (accumulator, currentValue) => accumulator + currentValue;
    return INTERNAL_DISPATCHER_QUEUE._tasks.toArray().map(task => task.body.entry.size).concat(INTERNAL_DISPATCHER_QUEUE.workersList().map(task => task.data.body.entry.size)).reduce(reducer, 0); // eslint-disable-line no-underscore-dangle
}

async function getWork(concurrency) {
    const queues = compact(Object.values(META_DATA).map(value => get(value, 'queues.transferrerQueue')));

    if (queues.length > 0) {
        const waitingWorkersThreshold = config.get('workers.transferrer.concurrency-eagerness-ratio') * concurrency;
        if (waitingWorkersThreshold > INTERNAL_DISPATCHER_QUEUE.length()) {
            const transferrerQueue = head(shuffle(queues));
            const messages = await receiveMessages(transferrerQueue);
            messages.sort((a, b) => a.body.entry.size - b.body.entry.size); // ascending sorted array. We want to push first the smaller files in order to fill the queue as much as possible (helpful when the target is S3)
            logger.debug(`Got ${messages.length} messages`, { running: INTERNAL_DISPATCHER_QUEUE.running(), length: INTERNAL_DISPATCHER_QUEUE.length() });
            messages.forEach(message => {
                try {
                    if (validateMessage(message.body, 'SQS', sqsTransferrerIncomingMessagesSchemes[message.body.type])) {
                        if (!message.body.target.s3 || prettyBytes(config.get('protocols.s3.max-object-size')) > totalTasksSize() + message.body.entry.size) {
                            INTERNAL_DISPATCHER_QUEUE.push(message);
                        }
                    }
                } catch (err) {
                    logger.warn('Failed to validate transferrer message', err);
                    deleteMessage(transferrerQueue, message.receiptHandle);
                }
            });
        }
        setImmediate(() => getWork(concurrency));
    } else {
        setTimeout(() => getWork(concurrency), SQS_DEFAULT_WAIT_TIMEOUT * 1000);
    }
}

function boot() {
    const concurrency = config.get('workers.transferrer.concurrency');
    logger.info(`Starting transferrer #${cluster.worker.id}, PID ${process.pid}, concurrency ${concurrency} ...`);

    // queue should have 'async' keyword otherwise it does not work
    INTERNAL_DISPATCHER_QUEUE = queue(async transferrerMessage => {
        const { body: { relationshipId } } = transferrerMessage;
        if (META_DATA[relationshipId]) {
            return handleTransferrerMessage(transferrerMessage);
        }
    }, concurrency);
    INTERNAL_DISPATCHER_QUEUE.drain = () => logger.info('Drain -> no more work');
    INTERNAL_DISPATCHER_QUEUE.error = (err, transferrerMessage) => logger.error(`Got error for transferrer message ${transferrerMessage}`, err);
    getWork(concurrency);
    setInterval(() => handleMessageVisibility(), ms(SQS_CHANGE_MESSAGE_VISIBILITY_TIMEOUT_INTERVAL));
}

process.on('message', message => {
    logger.info('Got IPC message', message);

    if (validateMessage(message, 'IPC', transferrerIncomingIPCMessagesSchemes[message.type])) {
        switch (message.type) {
            case IPC_TYPE_COPY:
                META_DATA[message.relationshipId] = {
                    queues: {
                        transferrerQueue: message.transferrerQueue
                    },
                    status: IPC_TYPE_COPY,
                    requestId: message.requestId,
                    items: {}
                };
                if (message.certificate && message.serverName) {
                    META_DATA[message.relationshipId].snfs = { certificate: message.certificate, serverName: message.serverName };
                }
                break;
            case IPC_TYPE_DONE:
            case IPC_TYPE_UPDATE:
            case IPC_TYPE_REMOVE:
            case IPC_TYPE_ABORT:
                handleDoneRemoveAbortUpdateMessage(message);
                break;
            default:
                logger.error('Got unexpected message type', message.type);
        }
    }
});

if (process.env.NODE_ENV !== 'test') {
    boot();
}

module.exports = { DELETE_TABLE, COPY_TABLE, isNetappEntry };
