const AWS = require('aws-sdk-mock');
const {
    listBuckets, getBucketTagging, getBucketLocation, listObjects, getObjectTagging, createTagsArray
} = require('../../../src/lib/aws/s3');
const { TYPE_FILE, TYPE_DIRECTORY, TYPE_SYMBOLIC } = require('../../../src/lib/consts');

const date = new Date('2018-01-30T13:09:52.000Z');
const mtimeSec = parseInt(date.getTime() / 1000);

describe('listBuckets', () => {
    afterEach(() => {
        AWS.restore('S3', 'listBuckets');
        AWS.restore('S3', 'getBucketLocation');
        AWS.restore('S3', 'getBucketTagging');
    });

    test('listBuckets - no buckets', async () => {
        AWS.mock('S3', 'listBuckets', callback => callback(null, { Buckets: [] }));
        const buckets = await listBuckets('us-west-1');
        expect(buckets).toHaveLength(0);
    });

    test('listBuckets - filter fabric pool', async () => {
        AWS.mock('S3', 'listBuckets', callback => callback(null, { Buckets: [{ Name: 'fabric-pool-111' }, { Name: 'test1' }] }));
        AWS.mock('S3', 'getBucketLocation', (params, callback) => callback(null, { LocationConstraint: 'eu-west-1' }));
        AWS.mock('S3', 'getBucketTagging', (params, callback) => callback(null, { TagSet: [] }));
        const buckets = await listBuckets('us-west-1');
        expect(buckets).toHaveLength(1);
    });
});

describe('getObjectTagging', () => {
    afterEach(() => AWS.restore('S3', 'getObjectTagging'));

    test('getObjectTagging - key doesnt exist', async () => {
        AWS.mock('S3', 'getObjectTagging', (params, callback) => callback({
            message: 'The specified key does not exist.',
            code: 'NoSuchKey'
        }, {}));

        const tags = await getObjectTagging('us-west-1', 'testBucket', 'test.js');
        expect(tags).toHaveLength(0);
    });

    test('getObjectTagging - bucket doesnt exist', async () => {
        AWS.mock('S3', 'getObjectTagging', (params, callback) => callback({
            message: 'The specified bucket does not exist',
            code: 'NoSuchBucket'
        }, {}));
        await expect(getObjectTagging('us-west-1', 'doesntexist', 'test.js')).rejects.toEqual({
            message: 'The specified bucket does not exist',
            code: 'NoSuchBucket'
        });
    });

    test('getObjectTagging -tags returns', async () => {
        AWS.mock('S3', 'getObjectTagging', (params, callback) => callback(null, { TagSet: [{ Key: 'a', Value: 'b' }] }));
        const tags = await getObjectTagging('us-east-1', 'dfiochen', 'secondfile.txt');
        expect(tags).toHaveLength(1);
    });
});

describe('getBucketTagging', () => {
    afterEach(() => AWS.restore('S3', 'getBucketTagging'));

    test('getBucketTagging - no tags on bucket', async () => {
        AWS.mock('S3', 'getBucketTagging', (params, callback) => callback({ message: 'The TagSet does not exist', code: 'NoSuchTagSet' }, {}));
        const tags = await getBucketTagging('us-west-1', 'test');
        expect(tags).toHaveLength(0);
    });
});

describe('getBucketLocation', () => {
    afterEach(() => AWS.restore('S3', 'getBucketLocation'));

    test('getBucketLocation -access denied', async () => {
        AWS.mock('S3', 'getBucketLocation', (params, callback) => callback({ message: 'Access Denied', code: 'AccessDenied' }, {}));
        await expect(getBucketLocation('us-west-1', 'testBucket')).rejects.toEqual({ code: 'AccessDenied', message: 'Access Denied' });
    });
});

describe('listObjects', () => {
    function s3FileObject(path, size, isSymbolic = false, uid = 0, gid = 0) {
        return {
            path,
            size,
            type: (isSymbolic) ? TYPE_SYMBOLIC : TYPE_FILE,
            mode: 33206,
            mtimeSec,
            uid,
            gid
        };
    }
    function s3DirectoryObject(path, uid = 0, gid = 0) {
        return {
            path,
            size: 0,
            type: TYPE_DIRECTORY,
            mode: 16822,
            mtimeSec,
            uid,
            gid
        };
    }
    afterEach(() => AWS.restore('S3', 'listObjectsV2'));

    test('listObjects - no valid bucket', async () => {
        AWS.mock('S3', 'listObjectsV2', (params, callback) => callback({
            message: 'The specified bucket is not valid.',
            code: 'InvalidBucketName'
        }, {}));

        await expect(listObjects('us-east-1', 'test')).rejects.toEqual({
            message: 'The specified bucket is not valid.',
            code: 'InvalidBucketName'
        });
    });


    test('listObjects - Access Denied', async () => {
        AWS.mock('S3', 'listObjectsV2', (params, callback) => callback({ code: 'AccessDenied', message: 'Access Denied' }, {}));
        await expect(listObjects('us-east-1', 'test')).rejects.toEqual({ code: 'AccessDenied', message: 'Access Denied' });
    });

    test('listObjects- with recursive calls', async () => {
        const listObjectsFunction = (params, callback) => {
            let result;
            if (params.ContinuationToken) {
                result = {
                    IsTruncated: false,
                    Contents: [{
                        Key: 'temp/lastObject.png',
                        LastModified: date,
                        ETag: '71e749519a5cdf0a437aa540d98d079a',
                        Size: 815104,
                        StorageClass: 'STANDARD'
                    },
                    {
                        Key: 'temp/',
                        LastModified: date,
                        ETag: '71e749519a5cdf0a437aa540d98d079a',
                        Size: 0,
                        StorageClass: 'STANDARD'
                    }],
                    Name: 'test',
                    Prefix: 'temp/',
                    Delimiter: '/',
                    MaxKeys: 1000,
                    CommonPrefixes: [{ Prefix: 'temp/dir1/dir2' }],
                    KeyCount: 3
                };
            } else {
                result = {
                    IsTruncated: true,
                    Contents: [{
                        Key: 'temp/testdoc.doc',
                        LastModified: date,
                        ETag: '71e749519a5cdf0a437aa540d98d079a',
                        Size: 815104,
                        StorageClass: 'STANDARD'
                    },
                    {
                        Key: 'temp/testdoc.html',
                        LastModified: date,
                        ETag: 'cab26c0e9c32031ee4d679b83eea13a8',
                        Size: 75121,
                        StorageClass: 'STANDARD'
                    }],
                    Name: 'test',
                    Prefix: 'temp/',
                    Delimiter: '/',
                    MaxKeys: 1000,
                    CommonPrefixes: [{ Prefix: 'temp/dir1/' }],
                    KeyCount: 3,
                    NextContinuationToken: 'next'
                };
            }
            callback(null, result);
        };
        const getObjectTaggingFunc = (params, callback) => {
            const files = ['temp/testdoc.html', 'temp/lastObject.png'];
            let result;
            if (files.includes(params.Key)) {
                result = { TagSet: [{ Key: 'netapp-cs-mode', Value: '33206' }, { Key: 'netapp-cs-uid', Value: '0' }, { Key: 'netapp-cs-gid', Value: '0' }, { Key: 'netapp-cs-mtimeSec', Value: mtimeSec.toString() }] };
            } else if (params.Key === 'temp/testdoc.doc') {
                result = { TagSet: [{ Key: 'netapp-cs-mode', Value: '33206' }, { Key: 'netapp-cs-uid', Value: '0' }, { Key: 'netapp-cs-gid', Value: '0' }, { Key: 'netapp-cs-slink', Value: 'True' }, { Key: 'netapp-cs-mtimeSec', Value: mtimeSec.toString() }] };
            } else {
                result = { TagSet: [{ Key: 'netapp-cs-mode', Value: '16822' }, { Key: 'netapp-cs-uid', Value: '0' }, { Key: 'netapp-cs-gid', Value: '0' }, { Key: 'netapp-cs-mtimeSec', Value: mtimeSec.toString() }] };
            }
            callback(null, result);
        };

        AWS.mock('S3', 'listObjectsV2', listObjectsFunction);
        AWS.mock('S3', 'getObjectTagging', getObjectTaggingFunc);
        const listObjectResult = await listObjects('us-east-1', 'test', 'temp');

        expect(listObjectResult).toEqual(expect.arrayContaining([
            s3FileObject('testdoc.doc', 815104, true), s3FileObject('testdoc.html', 75121), s3FileObject('lastObject.png', 815104),
            s3DirectoryObject('dir1'), s3DirectoryObject('dir2')]));
    });
});


describe('createTagsArray', () => {
    function createTag(key, value) {
        return { Key: key, Value: `${value}` };
    }

    test('createTags-HappyPath without symbolic link', () => {
        const entry = {
            uid: 0, gid: 1, mode: 16822, type: TYPE_FILE, mtimeSec
        };
        expect(createTagsArray(undefined, entry)).toEqual([createTag('netapp-cs-uid', 0), createTag('netapp-cs-gid', 1), createTag('netapp-cs-mode', 16822), createTag('netapp-cs-mtimeSec', mtimeSec)]);
    });

    test('createTags-HappyPath with symbolic link', () => {
        const entry = {
            uid: 0, gid: 1, mode: 16822, type: TYPE_SYMBOLIC, mtimeSec
        };
        expect(createTagsArray(undefined, entry)).toEqual([createTag('netapp-cs-uid', 0), createTag('netapp-cs-gid', 1), createTag('netapp-cs-mode', 16822), createTag('netapp-cs-mtimeSec', mtimeSec),
            createTag('netapp-cs-slink', 'True')]);
    });

    test('createTags-HappyPath with symbolic link and tags', () => {
        const entry = {
            uid: 0, gid: 1, mode: 16822, type: TYPE_SYMBOLIC, mtimeSec
        };
        const tags = [createTag('testKey', 'testValue')];
        expect(createTagsArray(tags, entry)).toEqual([createTag('testKey', 'testValue'), createTag('netapp-cs-uid', 0), createTag('netapp-cs-gid', 1), createTag('netapp-cs-mode', 16822), createTag('netapp-cs-mtimeSec', mtimeSec),
            createTag('netapp-cs-slink', 'True')]);
    });

    test('createTags-HappyPath with FILE type and 2 tags', () => {
        const entry = {
            uid: 0, gid: 1, mode: 16822, type: TYPE_FILE, mtimeSec
        };
        const tags = [createTag('testKey', 'testValue'), createTag('testKey1', 'testValue1')];
        expect(createTagsArray(tags, entry)).toEqual(tags.concat([createTag('netapp-cs-uid', 0), createTag('netapp-cs-gid', 1), createTag('netapp-cs-mode', 16822), createTag('netapp-cs-mtimeSec', mtimeSec)]));
    });
});
