const AWS = require('aws-sdk-mock');
const {
    listSgwsObjects
} = require('../../../src/lib/sgws/sgws');
const { TYPE_FILE, TYPE_DIRECTORY, TYPE_SYMBOLIC } = require('../../../src/lib/consts');

const date = new Date('2018-01-30T13:09:52.000Z');
const mtimeSec = parseInt(date.getTime() / 1000);


describe('listSgwsObjects', () => {
    function sgwsFileObject(path, size, isSymbolic = false, uid = 0, gid = 0) {
        return {
            path,
            size,
            type: (isSymbolic) ? TYPE_SYMBOLIC : TYPE_FILE,
            mode: 33206,
            mtimeSec,
            uid,
            gid
        };
    }
    function sgwsDirectoryObject(path, uid = 0, gid = 0) {
        return {
            path,
            size: 0,
            type: TYPE_DIRECTORY,
            mode: 16822,
            mtimeSec,
            uid,
            gid
        };
    }
    afterEach(() => AWS.restore('S3', 'listObjects'));


    test('listObjects- with recursive calls', async () => {
        const listObjectsFunction = (params, callback) => {
            let result;
            if (params.Marker) {
                result = {
                    IsTruncated: false,
                    Contents: [{
                        Key: 'temp/lastObject.png',
                        LastModified: date,
                        ETag: '71e749519a5cdf0a437aa540d98d079a',
                        Size: 815104,
                        StorageClass: 'STANDARD'
                    },
                    {
                        Key: 'temp/',
                        LastModified: date,
                        ETag: '71e749519a5cdf0a437aa540d98d079a',
                        Size: 0,
                        StorageClass: 'STANDARD'
                    }],
                    Name: 'test',
                    Prefix: 'temp/',
                    Delimiter: '/',
                    MaxKeys: 1000,
                    CommonPrefixes: [{ Prefix: 'temp/dir1/dir2' }],
                    KeyCount: 3
                };
            } else {
                result = {
                    IsTruncated: true,
                    Contents: [{
                        Key: 'temp/testdoc.doc',
                        LastModified: date,
                        ETag: '71e749519a5cdf0a437aa540d98d079a',
                        Size: 815104,
                        StorageClass: 'STANDARD'
                    },
                    {
                        Key: 'temp/testdoc.html',
                        LastModified: date,
                        ETag: 'cab26c0e9c32031ee4d679b83eea13a8',
                        Size: 75121,
                        StorageClass: 'STANDARD'
                    }],
                    Name: 'test',
                    Prefix: 'temp/',
                    Delimiter: '/',
                    MaxKeys: 1000,
                    CommonPrefixes: [{ Prefix: 'temp/dir1/' }],
                    KeyCount: 3,
                    NextMarker: 'next'
                };
            }
            callback(null, result);
        };
        const getMetadata = (params, callback) => {
            const files = ['temp/testdoc.html', 'temp/lastObject.png'];
            let result;
            if (files.includes(params.Key)) {
                result = {
                    AcceptRanges: 'bytes',
                    LastModified: date,
                    ContentLength: 19223,
                    ETag: '"e938eed8ca783dc4bfab0946cad368a5"',
                    ContentType: 'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
                    Metadata:
                    {
                        'netapp-cs-gid': '0',
                        'netapp-cs-mode': '33206',
                        'netapp-cs-mtimesec': mtimeSec.toString(),
                        'netapp-cs-uid': '0'
                    }
                };
            } else if (params.Key === 'temp/testdoc.doc') {
                result = {
                    AcceptRanges: 'bytes',
                    LastModified: date,
                    ContentLength: 19223,
                    ETag: '"e938eed8ca783dc4bfab0946cad368a5"',
                    ContentType: 'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
                    Metadata:
                        {
                            'netapp-cs-gid': '0',
                            'netapp-cs-mode': '33206',
                            'netapp-cs-mtimesec': mtimeSec.toString(),
                            'netapp-cs-uid': '0',
                            'netapp-cs-slink': 'True'
                        }
                };
            } else {
                result = {
                    AcceptRanges: 'bytes',
                    LastModified: date,
                    ContentLength: 19223,
                    ETag: '"e938eed8ca783dc4bfab0946cad368a5"',
                    ContentType: 'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
                    Metadata:
                        {
                            'netapp-cs-gid': '0',
                            'netapp-cs-mode': '16822',
                            'netapp-cs-mtimesec': mtimeSec.toString(),
                            'netapp-cs-uid': '0'
                        }
                };
            }
            callback(null, result);
        };

        AWS.mock('S3', 'listObjects', listObjectsFunction);
        AWS.mock('S3', 'headObject', getMetadata);
        const listObjectResult = await listSgwsObjects('1.1.1.1', 1111, 'a', 'a', 'test', 'temp');

        expect(listObjectResult).toEqual(expect.arrayContaining([
            sgwsFileObject('testdoc.doc', 815104, true), sgwsFileObject('testdoc.html', 75121), sgwsFileObject('lastObject.png', 815104),
            sgwsDirectoryObject('dir1'), sgwsDirectoryObject('dir2')]));
    });
});
