const { cloneDeep } = require('lodash');
const { compareFiles, compareDirs } = require('../../src/workers/scanner');
const {
    TYPE_FILE, TYPE_OTHER, TYPE_DIRECTORY, TYPE_SYMBOLIC
} = require('../../src/lib/consts');

describe('compareFiles', () => {
    const file = {
        path: 'test1',
        size: 123,
        type: TYPE_FILE,
        mode: 33206,
        uid: 1,
        gid: 0,
        mtimeSec: 1507638051
    };

    test('compareFiles - tags are the same', () => {
        expect(compareFiles([file], [file])).toEqual({ TASK_TYPE_COPY: [], TASK_TYPE_DELETE: [], TASK_TYPE_IGNORE: [] });
    });

    test('compareFiles - source has different tag value', () => {
        const source = cloneDeep(file);
        source.uid = 0;
        expect(compareFiles([source], [file])).toEqual({ TASK_TYPE_COPY: [source], TASK_TYPE_DELETE: [], TASK_TYPE_IGNORE: [] });
    });

    test('compareFiles - target has different tag value', () => {
        const target = cloneDeep(file);
        target.uid = 0;
        expect(compareFiles([file], [target])).toEqual({ TASK_TYPE_COPY: [file], TASK_TYPE_DELETE: [], TASK_TYPE_IGNORE: [] });
    });

    test('compareFiles - target has different time value', () => {
        const target = cloneDeep(file);
        target.mtimeSec = 1507638042;
        expect(compareFiles([file], [target])).toEqual({ TASK_TYPE_COPY: [file], TASK_TYPE_DELETE: [], TASK_TYPE_IGNORE: [] });
    });

    test('compareFiles - source has different tags values', () => {
        const source = cloneDeep(file);
        source.uid = 0;
        source.gid = 1;
        expect(compareFiles([source], [file])).toEqual({ TASK_TYPE_COPY: [source], TASK_TYPE_DELETE: [], TASK_TYPE_IGNORE: [] });
    });

    test('compareFiles - target has different tag values', () => {
        const target = cloneDeep(file);
        target.uid = 0;
        target.gid = 1;
        expect(compareFiles([file], [target])).toEqual({ TASK_TYPE_COPY: [file], TASK_TYPE_DELETE: [], TASK_TYPE_IGNORE: [] });
    });

    test('compareFiles - source has ignored file tag value', () => {
        const source = cloneDeep(file);
        source.type = TYPE_OTHER;
        source.path = 'ignored';
        expect(compareFiles([source], [])).toEqual({ TASK_TYPE_COPY: [], TASK_TYPE_DELETE: [], TASK_TYPE_IGNORE: [source] });
    });

    test('compareFiles - source and target has ignored files with same path', () => {
        const source = cloneDeep(file);
        const target = cloneDeep(file);
        source.type = TYPE_OTHER;
        target.type = TYPE_OTHER;
        source.path = 'ignored';
        target.path = 'ignored';
        source.size = 1;
        expect(compareFiles([source], [target])).toEqual({ TASK_TYPE_COPY: [], TASK_TYPE_DELETE: [], TASK_TYPE_IGNORE: [source, target] });
    });

    test('compareFiles - target has ignored files that source doesn\'t have', () => {
        const target = cloneDeep(file);
        target.type = TYPE_OTHER;
        target.path = 'ignored';
        expect(compareFiles([], [target])).toEqual({ TASK_TYPE_COPY: [], TASK_TYPE_DELETE: [], TASK_TYPE_IGNORE: [target] });
    });

    test('compareFiles - source is a special file and target is a regular file - target file should be deleted', () => {
        const source = cloneDeep(file);
        const target = cloneDeep(file);
        source.type = TYPE_OTHER;
        source.path = 'ignored';
        expect(compareFiles([source], [target])).toEqual({ TASK_TYPE_COPY: [], TASK_TYPE_DELETE: [target], TASK_TYPE_IGNORE: [source] });
    });

    test('compareFiles - target is a special file but source is normal - should try to copy', () => {
        const source = cloneDeep(file);
        const target = cloneDeep(file);
        target.type = TYPE_OTHER;
        target.path = 'ignored';
        expect(compareFiles([source], [target])).toEqual({ TASK_TYPE_COPY: [source], TASK_TYPE_DELETE: [], TASK_TYPE_IGNORE: [target] });
    });

    test('compareFiles - symbolic file & tags are the same', () => {
        const symbolic = cloneDeep(file);
        symbolic.type = TYPE_SYMBOLIC;
        expect(compareFiles([symbolic], [symbolic])).toEqual({ TASK_TYPE_COPY: [], TASK_TYPE_DELETE: [], TASK_TYPE_IGNORE: [] });
    });

    test('compareFiles - symbolic file only in source', () => {
        const symbolic = cloneDeep(file);
        symbolic.type = TYPE_SYMBOLIC;
        expect(compareFiles([symbolic], [])).toEqual({ TASK_TYPE_COPY: [symbolic], TASK_TYPE_DELETE: [], TASK_TYPE_IGNORE: [] });
    });

    test('compareFiles - symbolic file only in target', () => {
        const symbolic = cloneDeep(file);
        symbolic.type = TYPE_SYMBOLIC;
        expect(compareFiles([], [symbolic])).toEqual({ TASK_TYPE_COPY: [], TASK_TYPE_DELETE: [symbolic], TASK_TYPE_IGNORE: [] });
    });

    test('compareFiles - source is regular file & target symbolic', () => {
        const symbolic = cloneDeep(file);
        symbolic.type = TYPE_SYMBOLIC;
        expect(compareFiles([file], [symbolic])).toEqual({ TASK_TYPE_COPY: [file], TASK_TYPE_DELETE: [], TASK_TYPE_IGNORE: [] });
    });

    test('compareFiles - source is symbolic file & target regular', () => {
        const symbolic = cloneDeep(file);
        symbolic.type = TYPE_SYMBOLIC;
        expect(compareFiles([symbolic], [file])).toEqual({ TASK_TYPE_COPY: [symbolic], TASK_TYPE_DELETE: [], TASK_TYPE_IGNORE: [] });
    });
});
describe('compareDirs', () => {
    const dir = {
        path: 'dir1',
        size: 0,
        type: TYPE_DIRECTORY,
        mode: 16877,
        uid: 1,
        gid: 0,
        mtimeSec: 1507638051
    };

    test('compareDirs - only source', () => {
        expect(compareDirs([dir], [])).toEqual({
            TASK_TYPE_EQUAL: [], TASK_TYPE_SOURCE_ONLY: [dir], TASK_TYPE_TARGET_ONLY: [], TASK_TYPE_IGNORE: []
        });
    });
    test('compareDirs - only target', () => {
        expect(compareDirs([], [dir])).toEqual({
            TASK_TYPE_EQUAL: [], TASK_TYPE_SOURCE_ONLY: [], TASK_TYPE_TARGET_ONLY: [dir], TASK_TYPE_IGNORE: []
        });
    });
    test('compareDirs- tags are the same', () => {
        expect(compareDirs([dir], [dir])).toEqual({
            TASK_TYPE_EQUAL: [dir], TASK_TYPE_SOURCE_ONLY: [], TASK_TYPE_TARGET_ONLY: [], TASK_TYPE_IGNORE: []
        });
    });

    test('compareDirs - source and target are different', () => {
        const target = cloneDeep(dir);
        target.path = 'differentDir';
        expect(compareDirs([dir], [target])).toEqual({
            TASK_TYPE_EQUAL: [], TASK_TYPE_SOURCE_ONLY: [dir], TASK_TYPE_TARGET_ONLY: [target], TASK_TYPE_IGNORE: []
        });
    });
    test('compareDirs - source has .snapshot', () => {
        const src = cloneDeep(dir);
        src.path = '.snapshot';
        expect(compareDirs([src], [])).toEqual({
            TASK_TYPE_EQUAL: [], TASK_TYPE_SOURCE_ONLY: [], TASK_TYPE_TARGET_ONLY: [], TASK_TYPE_IGNORE: [src]
        });
    });
    test('compareDirs - target has .snapshot', () => {
        const target = cloneDeep(dir);
        target.path = '.snapshot';
        expect(compareDirs([], [target])).toEqual({
            TASK_TYPE_EQUAL: [], TASK_TYPE_SOURCE_ONLY: [], TASK_TYPE_TARGET_ONLY: [], TASK_TYPE_IGNORE: [target]
        });
    });
    test('compareDirs - source & target has .snapshot', () => {
        const src = cloneDeep(dir);
        src.path = '.snapshot';
        const target = cloneDeep(dir);
        target.path = '.snapshot';
        expect(compareDirs([dir, src], [target])).toEqual({
            TASK_TYPE_EQUAL: [], TASK_TYPE_SOURCE_ONLY: [dir], TASK_TYPE_TARGET_ONLY: [], TASK_TYPE_IGNORE: [src, target]
        });
    });
});
